/* ============================================================================
 * TFTPProtocolCommandHandler(TFTPProtocolCommsManagement).java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 1999-2001 by Paregos Mediadesign, Erik Eklund,
                 Johan "Spocke" Sorlin, http://sourceforge.net/projects/danoftp
 * Copyright (C) 2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or
 *      its contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 *
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 *
 * ============================================================================
 *
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 *
 * ============================================================================
 */
package nl.jhhcvandermeijs.ftp.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.security.MessageDigest;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.text.SimpleDateFormat;
import java.util.LinkedList;
import java.util.List;
import nl.jhhcvandermeijs.tcpconns.client.TTCPConnection;
import nl.jhhcvandermeijs.tcpconns.server.TTCPServer;
import nl.jhhcvandermeijs.tcpconns.utils.Utils;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

/**
 * TFTPProtocolCommandHandler implements the command handler part of the FTPProtocol
 */
public class TFTPProtocolCommandHandler extends TFTPProtocolCommsManagement {

    /**
     * Constructor. Do not make objects of this class directly. Instances are created
     * by the TFTPServer class when a client connects through the method newSeverSideProtocol.
     *
     * @param aTTCPConnection handle to the client; the client can be accessed through this.conn()
     * @param aTTCPServer handle to the server
     */
    public TFTPProtocolCommandHandler(TTCPConnection aTTCPConnection, TTCPServer aTTCPServer) {super(aTTCPConnection, aTTCPServer);}

    /**
     * A temporary place holder that is set to true when a rename action was started. The
     * RNFR (rename from) sets this value to true and it should be followed immediately by
     * RNTO (rename to).
     */
    protected boolean fRenameFromInvoked = false;

    /**
     * A temporary place holder to store the file/folder that is to be renamed. Used in
     * conjunction with fRenameFromInvoked
     */
    protected String fRenameFromPath = null;

	/**
	 * This is the FTP command handler
	 */
    @Override
    public void processRequest() throws IOException {
        if ((this.getAllowTLS()) && (this.getImplicitTLS())) {
            this.establishTLS();
        }

        this.sendResponse(FTP_SERVICE_READY, "Welcome (tcpconns/danoFTP Server)");

        while ((this.conn().isConnected()) && (Thread.currentThread().isInterrupted() == false)) {

            this.fRequest = null;

            try{

                this.fRequest = this.in().readLine();

            } catch (InterruptedIOException iioe) {
                if (this.isDataChannelOpen()) {
                    continue;
                } else {
                    try{this.conn().disconnect();} catch (Exception e3) {} //socket time out has been reached: terminate connection
                    throw iioe;
                }
            }
            
            if (this.fRequest == null) {
                try{this.conn().disconnect();} catch (Exception e3) {}
                this.log("\"" + this.getUserName() + "\" disconnected from server");
                throw new IOException();
            }

			// divide the command into command name and argument
	   	    if (this.fRequest.indexOf(" ") > -1) { //with argument
                this.fCommand = this.fRequest.substring(0,this.fRequest.indexOf(" ")).toUpperCase();
				this.fArgument = this.fRequest.substring(this.fRequest.indexOf(" ")+1,this.fRequest.length());
            } else { //no argument given
				this.fCommand = this.fRequest.toUpperCase();
				this.fArgument = "";
			}

			//ENFORCE LOG IN
			if (!(this.isLoggedIn())) {
				if (       (!(this.fCommand.startsWith("USER")))
                        && (!(this.fCommand.startsWith("PASS")))
                        && (!(this.fCommand.startsWith("FEAT")))
                        && (!(this.fCommand.startsWith("QUIT")))
                        && (!(this.fCommand.startsWith("AUTH")))
                        && (!(this.fCommand.startsWith("PBSZ")))
                        && (!(this.fCommand.startsWith("PROT")))
                        ) {
					this.sendResponse(FTP_NOT_LOGGED_IN, "Not logged in.");
					continue;
				}
			}

            //ENFORCE EXLPICIT TLS
            if (   (this.conn().isSecure() == false)
                &&
                   (this.getAllowTLS() == true)
                &&
                   (this.getImplicitTLS() == false)
                &&
                   (this.getEnforceExplicitTLSWhenNotImplicit())
                ) {
                    if (   (!(this.fCommand.startsWith("AUTH")))
                        && (!(this.fCommand.startsWith("FEAT")))
                        && (!(this.fCommand.startsWith("QUIT")))
                        && (!(this.fCommand.startsWith("PBSZ")))) {
                        this.sendResponse(FTP_ACCESS_DENIED, "ESTABLISH TLS FIRST: TLS IS MANDATORY");
                }
            }

            //ENFORCE DATA CHANNEL SECURITY (PROT P)
            if (   (this.conn().isSecure() == true)
                &&
                   (this.getImplicitTLS() == false)
                &&
                   (this.getEnforcePROTPWhenExplicit() == true)
                &&
                   (this.getPROTP() == false)
                ) {
                    if (   (!(this.fCommand.startsWith("USER")))
                        && (!(this.fCommand.startsWith("PASS")))
                        && (!(this.fCommand.startsWith("FEAT")))
                        && (!(this.fCommand.startsWith("PROT")))
                        && (!(this.fCommand.startsWith("QUIT")))) {
                        this.sendResponse(FTP_ACCESS_DENIED, "ESTABLISH PROT P FIRST: PROT P IS MANDATORY");
                }
            }

            //check whether we are actively sending or receiving (ie check for ABOR)
            if (this.getDataChannelState() != DATA_CHANNEL_STATE_IDLE) {
                if (this.fCommand.startsWith("ABOR")) {
                    this.onABOR(this.fArgument);
                    continue;
                } else if (this.fCommand.startsWith("QUIT")) {
                    this.onQUIT(this.fArgument);
                    continue;
                } else {
                    /* 
                     * on windows frequently a PASV was sent while the data channel state
                     * wasn't idle yet. We will allow a waiting period for up to 1.5 seconds
                     * to allow the data channel to finish. If it isn't finished by then
                     * an error message is sent to the client.
                     *
                     */
                    long start = System.currentTimeMillis();
                    long stop = start;
                    while (     (this.conn().isConnected())
                            &&
                                (Thread.currentThread().isInterrupted() == false)
                            &&
                                (this.getDataChannelState() != DATA_CHANNEL_STATE_IDLE)
                            &&
                                ((stop - start) < 1500)
                           ) {

                        try {Thread.sleep(50);} catch (Exception e) {}
                        stop = System.currentTimeMillis();

                    }
                    if ((this.getDataChannelState() != DATA_CHANNEL_STATE_IDLE)
                        &&
                       (stop - start >= 1500)) {
                            this.sendResponse(FTP_ACCESS_DENIED, "NOT ready for \"" + this.fRequest + "\". "
                                    + "Waited " + (stop-start) + " ms for data channel to close, but it is still open");
                            continue;
                    } else {
                            this.log("Now READY for \"" + this.fRequest + "\". "
                                    + "Waited " + (stop-start) + " ms for data channel to close");
                    }
                }
            }

            //if renameto is not called immediately after rename from then discard the renaming action
            if (this.fRenameFromInvoked) {
                if (this.fCommand.startsWith("RNTO") == false) {
                    this.fRenameFromInvoked = false;
                    this.fRenameFromPath = null;
                }
            }

            //command handlers
                   if (this.fCommand.equals("")) {
                this.sendResponse(FTP_NOT_IMPLEMENTED, "Sorry, \"" + this.fCommand + "\" command not implemented.");
                continue;
            } else if (this.fCommand.startsWith("USER")) {
                this.onUSER(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("PASS")) {
                this.onPASS(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("FEAT")) {
                this.onFEAT(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("QUIT")) {
                this.onQUIT(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("LIST")) {
                this.onLIST(this.fArgument);
                continue;
            } else if (this.fCommand.equals("MLST")) {
                this.onMLST(this.fArgument);
                continue;
            } else if (this.fCommand.equals("MLSD")) {
                this.onMLSD(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("SYST")) {
                this.onSYST(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("PWD")) {
                this.onPWD(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("TYPE")) {
                this.onTYPE(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("PASV")) {
                this.onPASV(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("NOOP")) {
                this.onNOOP(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("PORT")) {
                this.onPORT(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("CDUP")) {
                this.onCDUP(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("CWD")) {
                this.onCWD(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("MDTM")) {
                this.onMDTM(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("DELE")) {
                this.onDELE(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("MKD")) {
                this.onMKD(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("RMD")) {
                this.onRMD(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("NLST")) {
                this.onNLST(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("RETR")) {
                this.onRETR(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("RNFR")) {
                this.onRNFR(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("RNTO")) {
                this.onRNTO(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("SIZE")) {
                this.onSIZE(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("STOR")) {
                this.onSTOR(this.fArgument, false);
                continue;
            } else if (this.fCommand.startsWith("APPE")) {
                this.onSTOR(this.fArgument, true);
                continue;
            } else if (this.fCommand.startsWith("REIN")) {
                this.onREIN(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("PBSZ")) {
                this.onPBSZ(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("AUTH")) {
                this.onAUTH(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("PROT")) {
                this.onPROT(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("CCC")) {
                this.onCCC(this.fArgument);
                continue;
            } else if (this.fCommand.startsWith("CDC")) {
                this.onCDC(this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("MD2")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("MD4")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("MD5")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("RIPEMD128")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("RIPEMD160")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("RIPEMD256")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("RIPEMD320")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("SHA1")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("SHA224")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("SHA256")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("SHA384")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("SHA512")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("TIGER")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("GOST")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("GOST3411")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.endsWith("WHIRLPOOL")) {
                this.onMESSAGEDIGEST(this.fCommand, this.fArgument);
                continue;
            } else if (this.fCommand.equals("")) {

            } else if (this.fCommand.equals("")) {

            } else if (this.fCommand.equals("")) {

            } else if (this.fCommand.equals("")) {

            } else if (this.fCommand.equals("")) {

            } else if (this.fCommand.equals("")) {

            } else if (this.fCommand.equals("")) {

            } else {
                sendResponse(FTP_NOT_IMPLEMENTED, "Sorry, \"" + this.fCommand + "\" command not implemented.");
                continue;
            }
        }
    }

    /**
     * event handler for ABOR
     */
    public void onABOR(String argument) throws IOException {
        this.closeDataChannel();
        this.sendResponse(FTP_PATH_OK, "Abort on user request");
    }

    /**
     * event handler for USER
     * @param argument the username as given by the client
     * @throws java.io.IOException
     */
    public void onUSER(String argument) throws IOException {
        this.setUserName(argument);
        if (this.getUserName() != null) {
            this.sendResponse(FTP_USER_OK, "User \"" + this.getUserName() + "\" set, need password.");
            return;
        }
        this.sendResponse(FTP_BAD_ARGUMENT, "Bad argument  \"" + argument + "\"  for USER");
    }

    /**
     * event handler for PASS
     * @param argument the passphrase as given by the client
     * @throws java.io.IOException
     */
    public void onPASS(String argument) throws IOException {
        this.setPassphrase(argument);
        if (this.logIn()) {
            try{this.conn().mySocket().setSoTimeout(10*60*1000);} catch (Exception e) {}
            this.sendResponse(FTP_LOGIN_OK, "User \"" + this.getUserName() + "\" logged in");
        } else {
            this.sendResponse(FTP_NOT_LOGGED_IN, "Failed login attempt for \"" + this.getUserName() + "\"");
        }
    }

    /**
     * event handler for QUIT
     * @throws java.io.IOException
     */
    public void onQUIT(String argument) throws IOException {
        this.closeDataChannel();
        this.sendResponse(FTP_CLOSING_DATA_TRANSFER_PORT, "Closing connection on user request.");
        this.conn().disconnect();
    }

    /**
     * event handler for LIST
     * @throws java.io.IOException
     */
    public void onLIST(String argument) throws IOException {      
        String directoryName = null; if ((argument == null) || argument.equals("")) {directoryName = this.getCurrentDir();} else {directoryName = getAbsolutePath(argument, this.getCurrentDir());}
        FTPProtocolData.Folder folder = this.getFolder(directoryName);
        if (((directoryName != null) && (folder != null)) && ((new File(directoryName)).exists())) {
            File directory = new File(directoryName);            
            java.text.SimpleDateFormat dateFormat = new java.text.SimpleDateFormat( "MM-dd-yy  hh:mma" );
            String result, date, spacer, fileSize;
            List<String> resultList = new LinkedList<String>();
            String[] resultArray = new String[0];
                    
            if (directory.isFile()) {

                String fileName = directoryName;
                File file = directory;
                if (   ((this.isHidden(fileName, folder) == true ) && folder.getIsAllowedToSeeHiddenFiles())
                    ||
                       ((this.isHidden(fileName, folder) == false) && folder.getIsAllowedToSeeFiles())
                    ) {
                            date = "" + dateFormat.format(new java.util.Date(file.lastModified()));
                            fileSize = "" + file.length();
                            spacer = "                    "; spacer = spacer.substring(0, spacer.length() - fileSize.length());
                            result = date + " " + spacer + fileSize + " " + file.getName();
                            resultList.add(result);
                }

            } else if (directory.isDirectory()) {
                
                            File[] fileArray = directory.listFiles();
                            int i = 1-1; int n = fileArray.length;
                            while (i <= (n-1)) {
                                try {
                                    
                                    if (fileArray[i].isFile()) {
                                        if (    ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == true ) && folder.getIsAllowedToSeeHiddenFiles())
                                             ||
                                                ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == false) && folder.getIsAllowedToSeeFiles())
                                            ) {
                                                    date = "" + dateFormat.format(new java.util.Date(fileArray[i].lastModified()));
                                                    fileSize = "" + fileArray[i].length();
                                                    spacer = "                    "; spacer = spacer.substring(0, spacer.length() - fileSize.length());
                                                    result = date + " " + spacer + fileSize + " " + fileArray[i].getName();
                                                    resultList.add(result);
                                        }
                                    } else if (fileArray[i].isDirectory()) {
                                        if (   ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == true ) && folder.getIsAllowedToSeeHiddenSubDirectories())
                                            ||
                                               ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == false) && folder.getIsAllowedToSeeSubDirectories())
                                            ) {
                                                    date = "" + dateFormat.format(new java.util.Date(fileArray[i].lastModified()));
                                                    spacer = "      <DIR>         ";
                                                    result = date + " " + spacer + " " + fileArray[i].getName();
                                                    resultList.add(result);
                                        }
                                    }
                                    
                                } catch(Exception e) {
                                    resultList.add("04-10-01  04:28PM                    0 " + e.getMessage());
                                }
                            i++;
                            }                
            }
            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Listing \"" + directoryName + "\"");
            resultArray = resultList.toArray(resultArray);
            if (this.openDataChannel()) {
                this.sendFileList(resultArray);
                return;
            }
        }
        if ((argument == null) || (argument.equals(""))) {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + this.getCurrentDir() + "\"");
        } else {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + argument + "\"");
        }
    }

    /**
     * event handler for MLST
     * @throws java.io.IOException
     */
    public void onMLST(String argument) throws IOException {
        String directoryName = null; if ((argument == null) || argument.equals("")) {directoryName = this.getCurrentDir();} else {directoryName = getAbsolutePath(argument, this.getCurrentDir());}
        FTPProtocolData.Folder folder = this.getFolder(directoryName);
        if (((directoryName != null) && (folder != null)) && ((new File(directoryName)).exists())) {
            File directory = new File(directoryName);
            java.text.SimpleDateFormat dateFormat = new java.text.SimpleDateFormat("yyyyMMddhhmmss");
            String result, size, modify, type, perm;
            List<String> resultList = new LinkedList<String>();
            String[] resultArray = new String[0];

            if (directory.isFile()) {

                String fileName = directoryName;
                File file = directory;
                boolean isHidden = this.isHidden(fileName, folder);
                if (   ((isHidden == true ) && folder.getIsAllowedToSeeHiddenFiles())
                    ||
                       ((isHidden == false) && folder.getIsAllowedToSeeFiles())
                    ) {
                              size   = "Size="     + file.length();
                              modify = "Modify="   + dateFormat.format(new java.util.Date(file.lastModified()));
                              type   = "Type=file" ;
                              perm   = "Perm="     ;
                              if (isHidden) {
                                  if (folder.getIsAllowedToAppendToHiddenFiles())  {perm = perm + "a";}
                                  if (folder.getIsAllowedToDeleteHiddenFiles())    {perm = perm + "d";}
                                  if (folder.getIsAllowedToRenameHiddenFiles())    {perm = perm + "f";}
                                  if (folder.getIsAllowedToReadHiddenFiles())      {perm = perm + "r";}
                                  if (folder.getIsAllowedToOverwriteHiddenFiles()) {perm = perm + "w";}
                              } else {
                                  if (folder.getIsAllowedToAppendToFiles())        {perm = perm + "a";}
                                  if (folder.getIsAllowedToDeleteFiles())          {perm = perm + "d";}
                                  if (folder.getIsAllowedToRenameFiles())          {perm = perm + "f";}
                                  if (folder.getIsAllowedToReadFiles())            {perm = perm + "r";}
                                  if (folder.getIsAllowedToOverwriteFiles())       {perm = perm + "w";}
                              }
                              result = size + ";" + modify + ";" + type + ";" + perm + "; " + file.getName();
                              resultList.add(result);
                }
            } else if (directory.isDirectory()) {

                boolean isHidden = this.isHidden(directoryName, folder);
                if (   ((isHidden == true ) && folder.getIsAllowedToSeeHiddenSubDirectories())
                    ||
                       ((isHidden == false) && folder.getIsAllowedToSeeSubDirectories())
                    ||
                       (folder.getFolderName().equals(directoryName))
                    ) {
                              size   = "Size="     + directory.length();
                              modify = "Modify="   + dateFormat.format(new java.util.Date(directory.lastModified()));
                              type   = "Type=dir" ;
                              perm   = "Perm="     ;
                              if (folder.getIsAllowedToCreateNewFiles() || (folder.getIsAllowedToCreateNewHiddenFiles())) {perm = perm + "c";}
                              if (folder.getFolderName().equals(directoryName) == false) {
                                  if (folder.getIsAllowedToRemoveHiddenSubDirectories() || folder.getIsAllowedToRemoveSubDirectories()) {perm = perm + "d";}
                              }
                              if (folder.getIsAllowedToTraverseHiddenSubDirectories() || folder.getIsAllowedToTraverseSubDirectories()) {perm = perm + "e";}
                              if (folder.getFolderName().equals(directoryName) == false) {
                                  if (folder.getIsAllowedToRenameHiddenSubDirectories() || folder.getIsAllowedToRenameSubDirectories()) {perm = perm + "f";}
                              }
                              if (folder.getIsAllowedToSeeHiddenFiles() || folder.getIsAllowedToSeeFiles() || folder.getIsAllowedToSeeHiddenSubDirectories() || folder.getIsAllowedToSeeSubDirectories()) {perm = perm + "l";}
                              if (folder.getIsAllowedToMakeHiddenSubDirectories() || folder.getIsAllowedToMakeSubDirectories()) {perm = perm + "m";}
                              result = size + ";" + modify + ";" + type + ";" + perm + "; " + directory.getName();
                              resultList.add(result);
                }
            }
            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Listing \"" + directoryName + "\"");
            resultArray = resultList.toArray(resultArray);
            if (this.openDataChannel()) {
                this.sendFileList(resultArray);
                return;
            }
        }
        if ((argument == null) || (argument.equals(""))) {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + this.getCurrentDir() + "\"");
        } else {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + argument + "\"");
        }
    }

    /**
     * event handler for MLST
     * @throws java.io.IOException
     */
    public void onMLSD(String argument) throws IOException {
        String directoryName = null; if ((argument == null) || argument.equals("")) {directoryName = this.getCurrentDir();} else {directoryName = getAbsolutePath(argument, this.getCurrentDir());}
        FTPProtocolData.Folder folder = this.getFolder(directoryName);
        if (((directoryName != null) && (folder != null)) && ((new File(directoryName)).exists())) {
            File directory = new File(directoryName);
            java.text.SimpleDateFormat dateFormat = new java.text.SimpleDateFormat("yyyyMMddhhmmss");
            String result, size, modify, type, perm;
            List<String> resultList = new LinkedList<String>();
            String[] resultArray = new String[0];

            if (directory.isDirectory()) {

                            File[] fileArray = directory.listFiles();
                            int i = 1-1; int n = fileArray.length;
                            while (i <= (n-1)) {
                                try {
                                    boolean isHidden = this.isHidden(fileArray[i].getCanonicalPath(), folder);
                                    if (fileArray[i].isFile()) {
                                        if (    ((isHidden == true ) && folder.getIsAllowedToSeeHiddenFiles())
                                             ||
                                                ((isHidden == false) && folder.getIsAllowedToSeeFiles())
                                            ) {
                                                  size   = "Size="     + fileArray[i].length();
                                                  modify = "Modify="   + dateFormat.format(new java.util.Date(fileArray[i].lastModified()));
                                                  type   = "Type=file" ;
                                                  perm   = "Perm="     ;
                                                  if (isHidden) {
                                                      if (folder.getIsAllowedToAppendToHiddenFiles())  {perm = perm + "a";}
                                                      if (folder.getIsAllowedToDeleteHiddenFiles())    {perm = perm + "d";}
                                                      if (folder.getIsAllowedToRenameHiddenFiles())    {perm = perm + "f";}
                                                      if (folder.getIsAllowedToReadHiddenFiles())      {perm = perm + "r";}
                                                      if (folder.getIsAllowedToOverwriteHiddenFiles()) {perm = perm + "w";}
                                                  } else {
                                                      if (folder.getIsAllowedToAppendToFiles())        {perm = perm + "a";}
                                                      if (folder.getIsAllowedToDeleteFiles())          {perm = perm + "d";}
                                                      if (folder.getIsAllowedToRenameFiles())          {perm = perm + "f";}
                                                      if (folder.getIsAllowedToReadFiles())            {perm = perm + "r";}
                                                      if (folder.getIsAllowedToOverwriteFiles())       {perm = perm + "w";}
                                                  }
                                                  result = size + ";" + modify + ";" + type + ";" + perm + "; " + fileArray[i].getName();
                                                  resultList.add(result);
                                        }
                                    } else if (fileArray[i].isDirectory()) {

                                        if (   ((isHidden == true ) && folder.getIsAllowedToSeeHiddenSubDirectories())
                                            ||
                                               ((isHidden == false) && folder.getIsAllowedToSeeSubDirectories())
                                            ) {
                                                  size   = "Size="     + fileArray[i].length();
                                                  modify = "Modify="   + dateFormat.format(new java.util.Date(fileArray[i].lastModified()));
                                                  type   = "Type=dir" ;
                                                  perm   = "Perm="     ;
                                                  if (folder.getIsAllowedToCreateNewFiles() || (folder.getIsAllowedToCreateNewHiddenFiles())) {perm = perm + "c";}
                                                  if (folder.getIsAllowedToRemoveHiddenSubDirectories() || folder.getIsAllowedToRemoveSubDirectories()) {perm = perm + "d";}
                                                  if (folder.getIsAllowedToTraverseHiddenSubDirectories() || folder.getIsAllowedToTraverseSubDirectories()) {perm = perm + "e";}
                                                  if (folder.getIsAllowedToRenameHiddenSubDirectories() || folder.getIsAllowedToRenameSubDirectories()) {perm = perm + "f";}
                                                  if (folder.getIsAllowedToSeeHiddenFiles() || folder.getIsAllowedToSeeFiles() || folder.getIsAllowedToSeeHiddenSubDirectories() || folder.getIsAllowedToSeeSubDirectories()) {perm = perm + "l";}
                                                  if (folder.getIsAllowedToMakeHiddenSubDirectories() || folder.getIsAllowedToMakeSubDirectories()) {perm = perm + "m";}
                                                  result = size + ";" + modify + ";" + type + ";" + perm + "; " + fileArray[i].getName();
                                                  resultList.add(result);
                                        }
                                    }

                                } catch(Exception e) {}
                            i++;
                            }
            }
            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Listing \"" + directoryName + "\"");
            resultArray = resultList.toArray(resultArray);
            if (this.openDataChannel()) {
                this.sendFileList(resultArray);
                return;
            }
        }
        if ((argument == null) || (argument.equals(""))) {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + this.getCurrentDir() + "\"");
        } else {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + argument + "\"");
        }
    }

    /**
     * event handler for NLST
     * @throws java.io.IOException
     */
    public void onNLST(String argument) throws IOException {
        String directoryName = null; if ((argument == null) || argument.equals("")) {directoryName = this.getCurrentDir();} else {directoryName = getAbsolutePath(argument, this.getCurrentDir());}
        FTPProtocolData.Folder folder = this.getFolder(directoryName);
        if (((directoryName != null) && (folder != null)) && ((new File(directoryName)).exists())) {
            File directory = new File(directoryName);
            java.text.SimpleDateFormat dateFormat = new java.text.SimpleDateFormat( "MM-dd-yy  hh:mma" );
            List<String> resultList = new LinkedList<String>();
            String[] resultArray = new String[0];

            if (directory.isFile()) {

                String fileName = directoryName;
                File file = directory;
                if (   ((this.isHidden(fileName, folder) == true ) && folder.getIsAllowedToSeeHiddenFiles())
                    ||
                       ((this.isHidden(fileName, folder) == false) && folder.getIsAllowedToSeeFiles())
                    ) {
                            resultList.add(file.getName());
                }

            } else if (directory.isDirectory()) {

                            File[] fileArray = directory.listFiles();
                            int i = 1-1; int n = fileArray.length;
                            while (i <= (n-1)) {
                                try {

                                    if (fileArray[i].isFile()) {
                                        if (    ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == true ) && folder.getIsAllowedToSeeHiddenFiles())
                                             ||
                                                ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == false) && folder.getIsAllowedToSeeFiles())
                                            ) {
                                                    resultList.add(fileArray[i].getName());
                                        }
                                    } else if (fileArray[i].isDirectory()) {
                                        if (   ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == true ) && folder.getIsAllowedToSeeHiddenSubDirectories())
                                            ||
                                               ((this.isHidden(fileArray[i].getCanonicalPath(), folder) == false) && folder.getIsAllowedToSeeSubDirectories())
                                            ) {
                                                    resultList.add(fileArray[i].getName());
                                        }
                                    }

                                } catch(Exception e) {
                                    resultList.add("" + e.getMessage());
                                }
                            i++;
                            }
            }
            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Listing \"" + directoryName + "\"");
            resultArray = resultList.toArray(resultArray);
            if (this.openDataChannel()) {
                this.sendFileList(resultArray);
                return;
            }
        }
        if ((argument == null) || (argument.equals(""))) {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + this.getCurrentDir() + "\"");
        } else {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot list \"" + argument + "\"");
        }
    }

    /**
     * event handler for SYST
     * @throws java.io.IOException
     */
    public void onSYST(String argument) throws IOException {
        String syst = null;
        syst =   ""
               + System.getProperty("os.name")
               + "/"
               + System.getProperty("os.version")
               + " ("
               + System.getProperty("os.arch")
               + ") "
               + " (onSYST)";
        this.sendResponse(FTP_SYSTEM_NAME, syst);
    }

    /**
     * event handler for PWD
     * @throws java.io.IOException
     */
    public void onPWD(String argument) throws IOException {
        this.sendResponse(FTP_PATH_OK, "\"" + this.getCurrentDir() + "\" is current directory");
    }

    /**
     * event handler for TYPE
     * @param argument the argument as given by the client
     * @throws java.io.IOException
     */
    public void onTYPE(String argument) throws IOException {
        argument = argument.toUpperCase();
        if (argument.equals("I")) {
            this.setDataChannelDataMode(DATA_CHANNEL_DATA_MODE_BINARY);
            this.sendResponse(FTP_COMMAND_OK, "Type set to BINARY");
        } else if (argument.equals("A")) {
            this.setDataChannelDataMode(DATA_CHANNEL_DATA_MODE_ASCII);
            this.sendResponse(FTP_COMMAND_OK, "Type set to ASCII");
        } else {
            this.sendResponse(FTP_BAD_ARGUMENT, "Unknown type \"" + argument + "\" (I for BINARY or A for ASCII)");
        }
    }

    /**
     * event handler for PASV
     * @throws java.io.IOException
     */
    public void onPASV(String argument) throws IOException {
        String conndata = this.enterPassiveMode();
        if (conndata != null) {
            this.sendResponse(FTP_PASSIVE_MODE, "Entering Passive Mode (" + conndata + ").");
        } else {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot enter passive mode");
        }
    }

    /**
     * event handler for NOOP
     * @throws java.io.IOException
     */
    public void onNOOP(String argument) throws IOException {
        this.sendResponse(FTP_COMMAND_OK, "NOOP successful");
    }

    /**
     * event handler for PORT
     * @param argument the FTP address and port description as given by the client
     * @throws java.io.IOException
     */
    public void onPORT(String argument) throws IOException { //IPV6?????
        String address = argument;
        this.enterActiveMode();

        int[] numbers = new int[6];
        int i,posb, posa=0;

        //use client ip behind firewall/router instead of reported ip, which we may not reach
        byte[] rawClientIP = null;
        rawClientIP = this.conn().mySocket().getInetAddress().getAddress();

		try {
			for (i=0;i<5;i++) {
				posb = address.indexOf(",", posa);
				numbers[i]= Integer.parseInt(address.substring(posa,posb));
                if ((rawClientIP.length >= 4) && (i <= (4-1))) {
                    numbers[i]= rawClientIP[i]; //use real client ip instead
                }
				posa=posb+1;
			}
			numbers[i]=Integer.parseInt(address.substring(posa,address.length()));
		} catch (Exception e) {
			this.sendResponse(FTP_BAD_ARGUMENT, "Not a valid PORT specification");
			return;
		}

        this.setRemoteIP(numbers[0]+"."+numbers[1]+"."+numbers[2]+"."+numbers[3]);
		this.setRemotePort((numbers[4]<<8)+numbers[5]);
		this.sendResponse(FTP_COMMAND_OK, "PORT Command successful");
    }

    /**
     * event handler for CDUP
     * @throws java.io.IOException
     */
    public void onCDUP(String argument) throws IOException {
        //change up to parent directory
        File parentDirectory = null;
        try {
            //get the parent directory
            parentDirectory = (new File(this.getCurrentDir())).getParentFile();
            if (
                    (parentDirectory != null)

                &&

                    (this.getFolder(parentDirectory.getCanonicalPath()) != null)

                ) {
                        //change to parent dir
                        this.setCurrentDir(parentDirectory.getCanonicalPath());
                        this.sendResponse(FTP_PATH_OK, "Directory changed to \"" + this.getCurrentDir() + "\"");
                        return;
            }
        } catch (Exception e) {}
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot change to parent directory of \"" + this.getCurrentDir() + "\"");
    }


    /**
     * event handler for CWD
     * @param argument a directory as given by the client
     * @throws java.io.IOException
     */
    public void onCWD(String argument) throws IOException {
        argument = getAbsolutePath(argument, this.getCurrentDir());
        if (this.getFolder(argument) != null) {
            File directory = new File(argument);
            if ((directory.exists()) && (directory.isDirectory())) {

                this.setCurrentDir(argument);
                this.sendResponse(FTP_PATH_OK, "Directory changed to \"" + this.getCurrentDir() + "\"");
                return;

            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot change current directory to " + this.fArgument + "");
    }

    /**
     * event handler for MDTM
     * @param argument path as given by the client
     */
    public void onMDTM(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {
                    File file = new File(argument);
                    if (file.exists()) {
                        if (file.isFile()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToSeeHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToSeeFiles()))
                                ) {

                                        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddhhmmss");
                                        String dtg = "" + dateFormat.format(new java.util.Date(file.lastModified()));
                                        this.sendResponse(FTP_FILE_STATUS, dtg);
                                        return;

                            }
                        } else if (file.isDirectory()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToSeeHiddenSubDirectories()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToSeeSubDirectories()))
                                ) {


                                        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddhhmmss");
                                        String dtg = "" + dateFormat.format(new java.util.Date(file.lastModified()));
                                        this.sendResponse(FTP_FILE_STATUS, dtg);
                                        return;

                            }
                        }
                    }
            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot list MDTM info for \"" + argument + "\"");
    }

    /**
     * event handler for DELE
     * @param argument a filename that is to be deleted
     * @throws java.io.IOException
     */
    public void onDELE(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {
                    File file = new File(argument);
                    if (file.exists()) {
                        if (file.isFile()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToDeleteHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToDeleteFiles()))
                                ) {

                                    if (file.delete()) {
                                        this.sendResponse(FTP_PATH_OK, "Deleted file \"" + argument + "\"");
                                        return;
                                    }

                            }
                        }
                    }
            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot delete \"" + argument + "\"");
    }

    /**
     * event handler for MKD
     * @param argument the directory that is to be created
     * @throws java.io.IOException
     */
    public void onMKD(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {

                    File directory = new File(argument);
                    if ((directory.exists()) == false) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToMakeHiddenSubDirectories()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToMakeSubDirectories()))
                                ) {

                                        if (directory.mkdir()) {
                                            this.sendResponse(FTP_PATH_OK, "Created directory \"" + argument + "\"");
                                            return;
                                        }

                            }
                    }
            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot create directory \"" + argument + "\"");
    }

    /**
     * event handler for RMD
     * @param argument the directory that is to be removed
     * @throws java.io.IOException
     */
    public void onRMD(String argument) throws IOException {
        argument = getAbsolutePath(argument, this.getCurrentDir());
        if (argument != null) {
            String parent = Utils.validateDirectory(((new File(argument)).getParent()));
            if (parent != null) {
                FTPProtocolData.Folder folder = this.getFolder(parent);                
                if ((folder != null)) {
                    
                        File directory = new File(argument);
                        if ((directory.exists()) && (directory.isDirectory())) {
                                boolean isHidden = this.isHidden(argument, folder);
                                if (
                                        ((isHidden == true ) && (folder.getIsAllowedToRemoveHiddenSubDirectories()))
                                    ||
                                        ((isHidden == false) && (folder.getIsAllowedToRemoveSubDirectories()))
                                    ) {
                                            if (argument.equals(this.getCurrentDir())) {
                                                this.setCurrentDir(parent);
                                            }

                                            if (directory.delete()) {
                                                this.sendResponse(FTP_PATH_OK, "Removed directory \"" + argument + "\"");
                                                return;
                                            }
                                }
                        }
                    }
            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot remove directory \"" + this.fArgument + "\"");
    }

    /**
     * event handler for RETR
     * @param argument a file that is to be retrieved
     * @throws java.io.IOException
     */
    public void onRETR(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {

                File file = new File(argument);
                if ((file.exists()) && (file.isFile())) {

                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToReadHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToReadFiles()))
                                ) {

                                        this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Retrieving file \"" + argument + "\"");
                                        if (this.openDataChannel()) {
                                            this.sendFile(argument);
                                            return;
                                        }

                            }
                }

            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot retrieve \"" + argument + "\"");
    }

    /**
     * event handler for RNFR
     * @param argument folder or filename that is to be renamed
     * @throws java.io.IOException
     */
    public void onRNFR(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {

                File rnfrFile = new File(argument);
                if (rnfrFile.exists()) {

                        if (rnfrFile.isFile()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToRenameHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToRenameFiles()))
                                ) {
                        
                                    this.fRenameFromInvoked = true;
                                    this.fRenameFromPath = rnfrFile.getCanonicalPath();
                                    this.sendResponse(FTP_MORE_FILE_INFO, "Renaming \"" + this.fRenameFromPath + "\". Ready for destination name");
                                    return;
                            }                        
                        } else if (rnfrFile.isDirectory()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToRenameHiddenSubDirectories()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToRenameSubDirectories()))
                                ) {
                            
                                    this.fRenameFromInvoked = true;
                                    this.fRenameFromPath = rnfrFile.getCanonicalPath();
                                    this.sendResponse(FTP_MORE_FILE_INFO, "Renaming \"" + this.fRenameFromPath + "\". Ready for destination name");
                                    return;                        
                            }                        
                        }

                }

            }
        }
        this.fRenameFromInvoked = false;
        this.fRenameFromPath = null;
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot rename \"" + argument + "\"");
    }

    /**
     * event hanlder for RNTO
     * @param argument the destination name of the file or folder that is in the process of being renamed
     * @throws java.io.IOException
     */
    public void onRNTO(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {

                String rnto = argument;
                String rnfr = this.fRenameFromPath;
                if ((rnto != null) && (rnfr != null) && (this.fRenameFromInvoked)) {
                    this.fRenameFromInvoked = false;
                    File rnfrFile = new File(rnfr);
                    File rntoFile = new File(rnto);

                    if ((rnfrFile.exists() == true) && (rntoFile.exists() == false)) {

                        if (rnfrFile.isFile()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToRenameHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToRenameFiles()))
                                ) {

                                        if (rnfrFile.renameTo(rntoFile)) {
                                            this.sendResponse(FTP_COMMAND_OK, "\"" + rnfr + "\" renamed to \"" + rnto + "\"");
                                            return;
                                        }

                            }
                        } else if (rnfrFile.isDirectory()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToRenameHiddenSubDirectories()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToRenameSubDirectories()))
                                ) {

                                        if (rnfrFile.renameTo(rntoFile)) {
                                            this.sendResponse(FTP_COMMAND_OK, "\"" + rnfr + "\" renamed to \"" + rnto + "\"");
                                            return;
                                        }

                            }
                        }
                    }
                }

            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot rename \"" + this.fRenameFromPath +"\" to \"" + argument + "\"");
        this.fRenameFromPath = null;
        this.fRenameFromInvoked = false;
    }

    /**
     * event handler for SIZE
     * @param argument a file of which the size is to be determined
     * @throws java.io.IOException
     */
    public void onSIZE(String argument) throws IOException {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {
                    File file = new File(argument);
                    if (file.exists()) {
                        if (file.isFile()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToSeeHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToSeeFiles()))
                                ) {

                                        this.sendResponse(FTP_FILE_STATUS, "" + file.length());
                                        return;

                            }
                        }
                    }
            }
        }
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot calculate size for \"" + argument +  "\"");
    }

    /**
     * event handler for STOR/APPE
     * @param argument a file that is to be stored
     * @param append false when STOR, true when APPE
     * @throws java.io.IOException
     */
    public void onSTOR(String argument, boolean append) throws IOException { //onAPPE
        if ((getAbsolutePath(argument, this.getCurrentDir())) != null) {
            argument = getAbsolutePath(argument, this.getCurrentDir());
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {
                boolean performAction = false;

                File file = new File(argument);
                if (file.exists()) {
                    if (append) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToAppendToHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToAppendToFiles()))
                                ) {

                                    performAction = true;

                            }
                    } else {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToOverwriteHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToOverwriteFiles()))
                                ) {

                                    performAction = true;

                            }
                    }
                } else {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToCreateNewHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToCreateNewFiles()))
                                ) {

                                    performAction = true;

                            }
                }
                if (performAction) {
                    if (append) {
                        if (file.exists()) {
                            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Appending to file \"" + argument + "\"");
                        } else {
                            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Storing file \"" + argument + "\"");
                        }
                    } else {
                        if (file.exists()) {
                            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Overwriting file \"" + argument + "\"");
                        } else {
                            this.sendResponse(FTP_DATA_TRANSFER_PORT_WILL_OPEN, "Storing file \"" + argument + "\"");
                        }
                    }
                    if (this.openDataChannel()) {
                        this.storeFile(argument, append);
                        return;
                    }
                }
            }
        }

        if (getAbsolutePath(argument, this.getCurrentDir()) == null) {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot store file \"" + argument +  "\" [folder == null]");
        } else {
            this.sendResponse(FTP_ACCESS_DENIED, "Cannot store file \"" + argument +  "\"");
        }
    }

    /**
     * event handler for REIN
     */
    public void onREIN(String argument) throws IOException {
        //reinitialize
        this.setUser(null);
        this.setGroup(null);
        this.setCurrentDir("");
        this.setUserName("");
        this.setPassphrase("");

        if (this.conn().isSecure()) {
            if (this.getImplicitTLS() == true) {
                this.conn().disconnect();
            } else {
                this.conn().makeInsecure();
                this.setAUTHTLS(false);
                this.setPROTP(false);
            }
        }
    }

    /**
     * event handler for PBSZ: returns OK always
     */
    public void onPBSZ(String argument) throws IOException {
        this.sendResponse(FTP_COMMAND_OK, "PBSZ=" + argument);
    }

    /**
     * event handler for AUTH
     * @param argument dummy = TLS
     */
    public void onAUTH(String argument) throws IOException {        
        if (this.conn().isSecure()) {
            this.sendResponse(FTP_COMMAND_OK, "TLS ALREADY IN PLACE");
        } else {
            if (this.getAllowTLS()) {
                if (       ((argument != null) && (argument.equals("") == false))
                      &&
                           (argument.toUpperCase().endsWith("C"))
                      &&
                           (this.getImplicitTLS() == false)
                      &&
                           (this.getEnforcePROTPWhenExplicit() == false)
                    ) {
                            this.setPROTP(false);
                            this.sendResponse(FTP_TLS_START_OK, "OK STARTING TLS, PROT P OPTIONAL");
                } else {
                            this.setPROTP(true);
                            this.sendResponse(FTP_TLS_START_OK, "OK STARTING TLS, PROT P REQUIRED");
                }
                this.setAUTHTLS(true);
                return;
            } 
            this.sendResponse(534, "TLS NOT ALLOWED");            
        }
    }

    /**
     * event handler for CCC
     * @param argument dummy
     */
    public void onCCC(String argument) throws IOException {
        if (this.conn().isSecure()) {
            if ((this.getImplicitTLS()) || (this.getEnforceExplicitTLSWhenNotImplicit())){
                this.sendResponse(534, "CANNOT REMOVE TLS SINCE TLS IS MANDATORY");
                return;
            }
            this.sendResponse(FTP_COMMAND_OK, "OK SHUTTING DOWN TLS");
            try {this.conn().makeInsecure(); this.invalidate();} catch (Exception e) {try{this.conn().disconnect();} catch (Exception e2) {}}
            this.setAUTHTLS(false);
            this.setPROTP(false);
            return;
        } 
        this.sendResponse(533, "?? TLS WAS NOT IN EFFECT");        
    }

    /**
     * event handler for PROT
     * @param argument C or P
     */
    public void onPROT(String argument) throws IOException {
        if (this.conn().isSecure()) {
            if ((this.getImplicitTLS()) || (this.getEnforcePROTPWhenExplicit())) {
                this.sendResponse(FTP_COMMAND_OK, "PROT P IS MANDATORY");
                return;
            } 
            if (argument.toUpperCase().endsWith("C")) {
                this.setPROTP(false);
                this.sendResponse(FTP_COMMAND_OK, "PROT C");
                return;
            }
            this.setPROTP(true);
            this.sendResponse(FTP_COMMAND_OK, "PROT P");
            return;
        }
        this.sendResponse(FTP_ACCESS_DENIED, "CALL AUTH BEFORE CALLING PROT");
    }

    /**
     * event handler for CDC
     * @param argument dummy
     */
    public void onCDC(String argument) throws IOException {
        this.onPROT("C");
    }

    /**
     * this method is to check a client X509 certificate for revocation; only necessary
     * when client authentication is required by the server
     * @return true if the certificate is VALID and has NOT been revoked;
     * false if the certificate is INVALID and/or has been REVOKED
     */
    @Override
    public boolean onCheckRevocationStatus(X509Certificate anX509) {
        return true;
    }

    public void onMESSAGEDIGEST(String type, String argument) {
        if (getAbsolutePath(argument, this.getCurrentDir()) != null) {
            argument = Utils.validateFileName(getAbsolutePath(argument, this.getCurrentDir()));
            FTPProtocolData.Folder folder = this.getFolder(argument);
            if ((folder != null)) {
                    File file = new File(argument);
                    if (file.exists()) {
                        if (file.isFile()) {
                            if (
                                    ((this.isHidden(argument, folder) == true ) && (folder.getIsAllowedToSeeHiddenFiles()))
                                ||
                                    ((this.isHidden(argument, folder) == false) && (folder.getIsAllowedToSeeFiles()))
                                ) {
                                    FileInputStream inStream = null;
                                    try {
                                        Security.addProvider(new BouncyCastleProvider());

                                        MessageDigest messageDigest = null;
                                               if (type.endsWith("MD2")) {
                                                        messageDigest = MessageDigest.getInstance("MD2", "BC");
                                        } else if (type.endsWith("MD4")) {
                                                        messageDigest = MessageDigest.getInstance("MD4", "BC");
                                        } else if (type.endsWith("MD5")) {
                                                        messageDigest = MessageDigest.getInstance("MD5", "BC");
                                        } else if (type.endsWith("RIPEMD128")) {
                                                        messageDigest = MessageDigest.getInstance("RIPEMD128", "BC");
                                        } else if (type.endsWith("RIPEMD160")) {
                                                        messageDigest = MessageDigest.getInstance("RIPEMD160", "BC");
                                        } else if (type.endsWith("RIPEMD256")) {
                                                        messageDigest = MessageDigest.getInstance("RIPEMD256", "BC");
                                        } else if (type.endsWith("RIPEMD320")) {
                                                        messageDigest = MessageDigest.getInstance("RIPEMD320", "BC");
                                        } else if (type.endsWith("SHA1")) {
                                                        messageDigest = MessageDigest.getInstance("SHA1", "BC");
                                        } else if (type.endsWith("SHA224")) {
                                                        messageDigest = MessageDigest.getInstance("SHA-224", "BC");
                                        } else if (type.endsWith("SHA256")) {
                                                        messageDigest = MessageDigest.getInstance("SHA-256", "BC");
                                        } else if (type.endsWith("SHA384")) {
                                                        messageDigest = MessageDigest.getInstance("SHA-384", "BC");
                                        } else if (type.endsWith("SHA512")) {
                                                        messageDigest = MessageDigest.getInstance("SHA-512", "BC");
                                        } else if (type.endsWith("TIGER")) {
                                                        messageDigest = MessageDigest.getInstance("TIGER", "BC");
                                        } else if (type.endsWith("GOST")) {
                                                        messageDigest = MessageDigest.getInstance("GOST3411", "BC");
                                        } else if (type.endsWith("GOST3411")) {
                                                        messageDigest = MessageDigest.getInstance("GOST3411", "BC");
                                        } else if (type.endsWith("WHIRLPOOL")) {
                                                        messageDigest = MessageDigest.getInstance("WHIRLPOOL", "BC");
                                        }
                                        if (messageDigest != null) {
                                            inStream = new FileInputStream(file);
                                            byte[] buffer = new byte[4096];
                                            long i = 1-1; long n = (long)(file.length() / 4096); int m = (int)(file.length() % 4096);
                                            while (i <= (n-1)) {
                                                inStream.read(buffer);
                                                messageDigest.update(buffer);
                                                i++;
                                            }
                                            if (m > 0) {
                                                inStream.read(buffer, 0, m);
                                                messageDigest.update(buffer, 0, m);
                                            }
                                            try {inStream.close();} catch (Exception e2) {}

                                            byte[] digest = messageDigest.digest();
                                            String digestString = Utils.toHex(digest);
                                            this.sendResponse(FTP_FILE_STATUS, "" + digestString);

                                            return;
                                        }
                                    } catch (Exception e) {} finally {
                                        try {inStream.close();} catch (Exception e2) {}
                                    }

                            }
                        }
                    }
            }
        }        
        this.sendResponse(FTP_ACCESS_DENIED, "Cannot calculate \"" + type + "\" value for \"" + argument +  "\"" + "  " + type + "  " + argument);
    }

    public void onFEAT(String argument) throws IOException {
        this.log("Listing Features");
        this.out().print(FTP_LIST_OF_FEATURES + "- Features:\r\n");
        this.out().print(" MDTM\r\n");
        this.out().print(" SIZE\r\n");
        this.out().print(" MLST Size*;Modify*;Type*;Perm*;\r\n");
        this.out().print(" MLSD Size*;Modify*;Type*;Perm*;\r\n");
        this.out().print(" AUTH SSL\r\n");
        this.out().print(" AUTH TLS\r\n");
        this.out().print(" MD5\r\n");
        this.out().print(" SHA1\r\n");
        this.out().print(" SHA256\r\n");
        this.out().print(" SHA384\r\n");
        this.out().print(" SHA512\r\n");
        this.out().print(" XMD5\r\n");
        this.out().print(" XSHA1\r\n");
        this.out().print(" XSHA256\r\n");
        this.out().print(" XSHA384\r\n");
        this.out().print(" XSHA512\r\n");
        this.out().print(FTP_LIST_OF_FEATURES + " End\r\n");
        this.out().flush();
    }
}
