/* ============================================================================
 * TFTPProtocolDataChannelManagement(TFTPProtocolUserManagement).java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 1999-2001 by Paregos Mediadesign, Erik Eklund,
                 Johan "Spocke" Sorlin, http://sourceforge.net/projects/danoftp
 * Copyright (C) 2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or
 *      its contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 *
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 *
 * ============================================================================
 *
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 *
 * ============================================================================
 */
package nl.jhhcvandermeijs.ftp.server;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.security.cert.X509Certificate;
import nl.jhhcvandermeijs.tcpconns.client.TTCPConnection;
import nl.jhhcvandermeijs.tcpconns.server.TTCPServer;
import nl.jhhcvandermeijs.tcpconns.utils.Utils;

/**
 * TFTPProtocolDataChannelManagement deals with setting up and closing down data channels
 */
public class TFTPProtocolDataChannelManagement extends TFTPProtocolUserManagement {

    /**
     * Constructor. Do not make objects of this class directly. Instances are created
     * by the TFTPServer class when a client connects through the method newSeverSideProtocol.
     *
     * @param aTTCPConnection handle to the client; the client can be accessed through this.conn()
     * @param aTTCPServer handle to the server
     */
    public TFTPProtocolDataChannelManagement(TTCPConnection aTTCPConnection, TTCPServer aTTCPServer) {super(aTTCPConnection, aTTCPServer);}

    /**
     * A handle to the data channel when there is a data channel open
     */
    private TTCPConnection fDataChannel = null;
    /**
     * A handle to the data channel when there is a data channel open
     * @return fDataChannel
     */
    public synchronized TTCPConnection getDataChannel() {return this.fDataChannel;}
    /**
     * method to set fDataChannel
     * @param dataChannel a dataChannel
     */
    public synchronized void setDataChannel(TTCPConnection dataChannel) {this.fDataChannel = dataChannel;}

    /**
     * Forms a handle to the thread that sends or receives data over the data channel
     */
    private Thread fDataChannelThread = null;
    /**
     * method to obtain fDataChannelThread
     * @return fDataChannelThread
     */
    public synchronized Thread getDataChannelThread() {
        return this.fDataChannelThread;
    }
    /**
     * method to set fDataChannelThread
     * @param dataChannelThread fDataChannelThread
     */
    public synchronized void setDataChannelThread(Thread dataChannelThread) {
        this.fDataChannelThread = dataChannelThread;
    }

    /**
     * A constant denoting ASCII transfer mode. Setting the mode to ASCII for now is irrelevant
     * since this protocol assumes BINARY always. This assumption is no big issue (at this moment), since
     * FTP clients such as FileZilla (http://filezilla-project.org/, which runs on Linux,
     * Windows XP/Vista, MacOSX, FreeBSD, Solaris => together > 99% of all systems!) request for
     * BINARY mode almost immediately and never request ASCII mode once BINARY is set.
     */
    protected static final int DATA_CHANNEL_DATA_MODE_ASCII = 0;
    /**
     * A constant denoting BINARY transfer mode. Setting the mode to ASCII for now is irrelevant
     * since this protocol assumes BINARY always. This assumption is no big issue (at this moment), since
     * FTP clients such as FileZilla (http://filezilla-project.org/, which runs on Linux,
     * Windows XP/Vista, MacOSX, FreeBSD, Solaris => together > 99% of all systems!) request for
     * BINARY mode almost immediately and never request ASCII mode once BINARY is set.
     */
    protected static final int DATA_CHANNEL_DATA_MODE_BINARY = 1;
    /**
     * Protected field that indicates either ASCII transfer mode or BINARY transfer mode. Setting the mode to ASCII for now is irrelevant
     * since this protocol assumes BINARY always. This assumption is no big issue (at this moment), since
     * FTP clients such as FileZilla (http://filezilla-project.org/, which runs on Linux,
     * Windows XP/Vista, MacOSX, FreeBSD, Solaris => together > 99% of all systems!) request for
     * BINARY mode almost immediately and never request ASCII mode once BINARY is set.
     */
    private int fDataChannelDataMode = DATA_CHANNEL_DATA_MODE_ASCII;
    /**
     * Public method to obtain the value for fDataChannelDataMode. Setting the mode to ASCII for now is irrelevant
     * since this protocol assumes BINARY always. This assumption is no big issue (at this moment), since
     * FTP clients such as FileZilla (http://filezilla-project.org/, which runs on Linux,
     * Windows XP/Vista, MacOSX, FreeBSD, Solaris => together > 99% of all systems!) request for
     * BINARY mode almost immediately and never request ASCII mode once BINARY is set.
     * @return either DATA_CHANNEL_DATA_MODE_ASCII or DATA_CHANNEL_DATA_MODE_BINARY
     */
    public synchronized int getDataChannelDataMode() {return this.fDataChannelDataMode;}
    /**
     * Public method to set the value for fDataChannelDataMode. Setting the mode to ASCII for now is irrelevant
     * since this protocol assumes BINARY always. This assumption is no big issue (at this moment), since
     * FTP clients such as FileZilla (http://filezilla-project.org/, which runs on Linux,
     * Windows XP/Vista, MacOSX, FreeBSD, Solaris => together > 99% of all systems!) request for
     * BINARY mode almost immediately and never request ASCII mode once BINARY is set.
     * @param aDataChannelDataMode
     */
    public synchronized void setDataChannelDataMode(int aDataChannelDataMode) {this.fDataChannelDataMode = aDataChannelDataMode;}

    /**
     * Protected value to indicate whether the client is active or passive in establishing
     * the datachannel connection. When the client is passive the server will open a channel, when
     * the client is active then the client will open a channel. Due to firewall settings, most of
     * the time the client requests for "PASSIVE" file transfers.
     */
    private boolean fPassiveMode = false;
   /**
     * use this method to test whether the client has requested for passive data transfers.
     * @return true if the client is passive, false if the client is active
     */
    public synchronized boolean getPassiveMode() {return this.fPassiveMode;}
    /**
     * method to set fPassiveMode
     * @param passiveMode true if client is passive
     */
    protected synchronized void setPassiveMode(boolean passiveMode) {
        this.fPassiveMode = passiveMode;
    }

    /**
     * Protected value that is to hold the remote port number when a client is in active mode and
     * the server has to connect to the client.
     */
    private int fRemotePortWhenNotPassive = -1;
    /**
     * Protected value that is to hold the remote IP string when a client is in active mode. When the
     * client is behind a router, the IP address reported by the client is not applicable; the IP
     * address should be obtained from the actual connection that was established, otherwise the
     * server cannot connect to the client.
     */
    private String fRemoteIPWhenNotPassive = null;
    /**
     * once the client has given the PORT signal, the remote port and remote IP are set. This method
     * is used to obtain these values once establishing the connection to the client. Only applicable
     * if the client is in active mode.
     * @return the set remote port number
     */
    public synchronized int getRemotePort() {return this.fRemotePortWhenNotPassive;}
    /**
     * this method is used to set the fRemotePortWhenNotPassive value
     * @param aRemotePort an integer value indicating the remote port for "active" data transfers
     */
    public synchronized void setRemotePort(int aRemotePort) {this.fRemotePortWhenNotPassive = aRemotePort;}
    /**
     * once the client has given the PORT signal, the remote port and remote IP are set. This method
     * is used to obtain these values once establishing the connection to the client. Only applicable
     * if the client is in active mode.
     * @return the set remote IP
     */
    public synchronized String getRemoteIP() {return this.fRemoteIPWhenNotPassive;}
    /**
     * this method is used to set the fRemoteIPWhenNotPassive string
     * @param aRemoteIP a string containing the remote IP address for "active" data transfers
     */
    public synchronized void setRemoteIP(String aRemoteIP) {this.fRemoteIPWhenNotPassive = aRemoteIP;}

    /**
     * Protected field that forms a handle to the ServerSocket that is used to open channels for
     * for data transfers when the client is in passive mode.
     */
    private ServerSocket fPassiveServerSocket = null;
    /**
     * method to obtain fPassiveServerSocket
     * @return fPassiveServerSocket
     */
    public synchronized ServerSocket getPassiveServerSocket() {
        return this.fPassiveServerSocket;
    }
    /**
     * method to set fPassiveServerSocket
     * @param passiveServerSocket fPassiveServerSocket
     */
    public synchronized void setPassiveServerSocket(ServerSocket passiveServerSocket) {
        this.fPassiveServerSocket = passiveServerSocket;
    }

    /**
     * Protected field that holds the current data channel IP and Port string that is to be send
     * to the client so that the client can connect. Note: the client must realize that the server
     * may be behind a router. If that is the case, then the reported IP from the server is probably
     * not correct, since it will report a local network IP instead of its proper router IP address.
     */
    private String fPassiveDataChannelIPAndPort = null;
    /**
     * method to obtain fPassiveDataChannelIPAndPort string
     * @return fPassiveDataChannelIPAndPort
     */
    public synchronized String getPassiveIPAndPort() {
        return this.fPassiveDataChannelIPAndPort;
    }
    /**
     * method to set fPassiveDataChannelIPAndPort
     * @param passiveDataChannelIPAndPortString fPassiveDataChannelIPAndPort
     */
    public synchronized void setPassiveIPAndPort(String passiveDataChannelIPAndPortString) {
        this.fPassiveDataChannelIPAndPort = passiveDataChannelIPAndPortString;
    }

    /**
     * Method that is called when the client requests "passive" data transfers; this means
     * that the server will open sockets for data transfers instead of the client (indicating
     * that the client is passive instead of active).
     * @return a string that contains the datachannel IP and Port. Note that the client must not
     * automatically assume the IP to be correct. The server can be placed in a local network and
     * be connected through a router to the outside world; if so, the reported IP is a local IP
     * instead of the correct router IP.
     */
    protected String enterPassiveMode() {
        if (this.getPassiveMode() == false) {
            String aDataChannelIPAndPort = null;

            this.setPassiveServerSocket(getServerSocket(this.getBindIPAddress())); //necessary for windows
            aDataChannelIPAndPort = getFTPIPPortString(Utils.getNetworkIp(), this.getPassiveServerSocket().getLocalPort());

            this.setPassiveMode(true);
            this.setPassiveIPAndPort(aDataChannelIPAndPort);
        }
      	return this.getPassiveIPAndPort();
    }

    /**
     * This method can be used to restore the protocol to active transfer mode if
     * applicable
     */
    protected void enterActiveMode() {
        if (this.getPassiveServerSocket() != null) {try {this.getPassiveServerSocket().close();} catch (Exception e) {}}
        this.setPassiveIPAndPort(null);
        this.setPassiveServerSocket(null);
        this.setPassiveMode(false);
    }

    /**
     * A constant indicating an IDLE state (no active data transfer)
     */
    public static final int DATA_CHANNEL_STATE_IDLE      = 100;
    /**
     * A constant indicating there is an active data transfer
     */
    public static final int DATA_CHANNEL_STATE_BUSY      = 200;
    /**
     * A protected field that indicates the state of the data channel (IDLE = closed data channel;
     * BUSY = open data channel)
     */
    private int fDataChannelState = DATA_CHANNEL_STATE_IDLE;
    /**
     * A method that can be used to obtain the value for fDataChannelState
     * @return fDataChannelState DATA_CHANNEL_STATE_IDLE or
     * DATA_CHANNEL_STATE_BUSY
     */
    public synchronized int getDataChannelState() {return this.fDataChannelState;}
    /**
     * A method that is used to set the value for fDataChannelState
     * @param state DATA_CHANNEL_STATE_IDLE or
     * DATA_CHANNEL_STATE_BUSY
     */
    public synchronized void setDataChannelState(int state) {this.fDataChannelState = state;}

    /**
     * a method to test whether the data channel should be secure
     * @return true if the connection should be secure, false if not
     */
    public synchronized boolean shouldDataChannelBeSecure() {
        if (
                (this.getAllowTLS())
             &&
                (     (this.getImplicitTLS())
                  ||
                      ((this.getEnforceExplicitTLSWhenNotImplicit()) && (this.getEnforcePROTPWhenExplicit()))
                  ||
                      ((this.getAUTHTLS()) && (this.getEnforcePROTPWhenExplicit())
                  ||
                      ((this.getAUTHTLS()) && (this.getPROTP()))
                ))) {

            return true;

        }
        return false;
    }

    protected void establishTLSOnDataChannel() throws IOException {
        if (this.getRequireClientAuthentication()) {
                
            try {this.getDataChannel().makeSecureWithClientCertAuthByServer(this.getSSLSocketFactory());} catch (Exception e) {this.getDataChannel().disconnect();}

            java.security.cert.Certificate[] peerCertChain = this.getDataChannel().getPeerCertificateChain();
            //check whether we have a peerCertChain
            if (peerCertChain == null) {this.getDataChannel().disconnect(); throw new IOException();}
            if (peerCertChain.length == 0) {this.getDataChannel().disconnect(); throw new IOException();}
            if (!(peerCertChain[0] instanceof X509Certificate)) {this.getDataChannel().disconnect(); new IOException();}

            //check X509
            X509Certificate anX509 = ((X509Certificate)(peerCertChain[0]));
            if (!(this.isValid(this.getPEMEncodedServerCertAndKeyFileName(), anX509))) {this.getDataChannel().disconnect(); throw new IOException();}

        } else {
            try {this.getDataChannel().makeSecureWithServerCertAuthByClient(this.getSSLSocketFactory());} catch (Exception e) {this.getDataChannel().disconnect();}
        }
    }

    /**
     * This method opens a data channel (when client is "passive") or connects to one (when client is "active")
     * @return whether the data transfer bridge has been succesfully established
     */
    protected synchronized boolean openDataChannel() {
        this.closeDataChannel();

        //connect
        Socket aClientSocket = null;
		if (this.getPassiveMode() == false) {
			try {   aClientSocket = new Socket(this.getRemoteIP(), this.getRemotePort());   } catch (Exception e) {return false;}
		} else {
			try {   aClientSocket = this.getPassiveServerSocket().accept();                 } catch (Exception e) {return false;}
            this.enterActiveMode(); //apparently obligatory for windows (at least for the vista that is running on my test machine)
		}

        try{aClientSocket.setSoTimeout(10*60*1000);}catch (Exception e) {}
        try{aClientSocket.setSoLinger(true, 10*1000);}catch (Exception e) {}

        //get a TTCPConnection
        this.setDataChannel(new TTCPConnection(false)); //false => we are server
        try{this.getDataChannel().connect(aClientSocket);} catch (Exception e) {this.closeDataChannel();}

        //establish TLS if indicated
        if (this.shouldDataChannelBeSecure()) {
            try {
                this.establishTLSOnDataChannel();
            } catch (Exception e) {
                this.closeDataChannel();
            }
        }

        //return openness of channel
        return this.isDataChannelOpen();
    }

    /**
     * "Tests" whether the data channel is open.
     * @return true if the data channel indeed is open, false if it's not
     */
    protected synchronized boolean isDataChannelOpen() {
        //test if we've recently opened a datachannel
        if (this.getDataChannel() == null) {return false;}

        //test if we're connected
        if (this.getDataChannel().isConnected() == false) {this.closeDataChannel(); return false;}

        //we are connected
        return true;                                    
    }

    /**
     * This method attempts to close the data channel
     */
    protected synchronized void closeDataChannel() {
        //close data channel
        if (this.getDataChannel() != null) {
            try{this.getDataChannel().makeInsecure();} catch (Exception e) {}
            try{this.getDataChannel().disconnect();} catch (Exception e) {}
                this.setDataChannel(null);
        }

        //terminate data channel thread
        if (this.getDataChannelThread() != null) {
            Thread aThread = this.getDataChannelThread();
            this.setDataChannelThread(null);
            try{aThread.interrupt();} catch (Exception e) {}
        }

        //reset data channel state
        this.setDataChannelState(DATA_CHANNEL_STATE_IDLE);
    }

}