/* ============================================================================
 * TFTPServer.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or
 *      its contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 *
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 *
 * ============================================================================
 *
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 *
 * ============================================================================
 */
package nl.jhhcvandermeijs.ftp.server;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import javax.net.ssl.SSLSocketFactory;
import nl.jhhcvandermeijs.tcpconns.client.TTCPConnection;
import nl.jhhcvandermeijs.tcpconns.keys.KeyStoreHelper;
import nl.jhhcvandermeijs.tcpconns.server.TServerSideProtocol;
import nl.jhhcvandermeijs.tcpconns.server.TTCPServer;
import nl.jhhcvandermeijs.tcpconns.tls.TTLSProvider;
import nl.jhhcvandermeijs.tcpconns.utils.Utils;

/**
 * TFTPServer implements a multi-threaded FTPServer
 */
public class TFTPServer extends TTCPServer {
    /**
     * the method to obtain new server side protocol instances
     * @param aTTCPConnection the connection that was made to the client
     * @param aTTCPServer the server that is to own the server side protocol instance
     * @return a new server side protocol instance
     * @throws java.io.IOException
     */
    @Override
    synchronized public TServerSideProtocol newServerSideProtocol(TTCPConnection aTTCPConnection, TTCPServer aTTCPServer ) throws IOException {
        TFTPProtocol ftpProtocol = new TFTPProtocol(aTTCPConnection, aTTCPServer);

        ftpProtocol.setMaxClients(this.data().getMaxClients());
        ftpProtocol.setServerPort(this.data().getServerPort());
        ftpProtocol.setBindIPAddress(this.data().getBindIPAddress());
        ftpProtocol.setExternalIPOfThisServer(this.data().getExternalIPOfThisServer());
        ftpProtocol.setTrackChanges(this.data().getTrackChanges());

        ftpProtocol.setLogDirectory(this.data().getLogDirectory());
        ftpProtocol.setMainLogFileName(this.data().getMainLogFileName());
        ftpProtocol.setLogEvents(this.data().getLogEvents());
        ftpProtocol.setLogAppend(this.data().getLogAppend());

        ftpProtocol.setAllowTLS(this.data().getAllowTLS());
        ftpProtocol.setImplicitTLS(this.data().getImplicitTLS());
        ftpProtocol.setEnforceExplicitTLSWhenNotImplicit(this.data().getEnforceExplicitTLSWhenNotImplicit());
        ftpProtocol.setEnforcePROTPWhenExplicit(this.data().getEnforcePROTPWhenExplicit());
        ftpProtocol.setRequireClientAuthentication(this.data().getRequireClientAuthentication());
        ftpProtocol.setPEMEncodedServerCertAndKeyFileName(this.data().getPEMEncodedServerCertAndKeyFileName());
        ftpProtocol.setTrustedCACertsDirectory(this.data().getTrustedCACertsDirectory());
        ftpProtocol.setKeyPassphrase(this.data().getKeyPassphrase());

        ftpProtocol.setSSLSocketFactory(this.fSSLSocketFactory);

        return ftpProtocol;
    }

    /**
     * Method to obtain a handle to the appdir inside the user folder. User here refers to
     * the user that was logged in to the system.
     * @return {SYSTEMUSERDIR}/.tcpconns/ftpserver/
     */
    public static String getAppdir() {
        String fileName = Utils.getUserHome() + ".tcpconns" + File.separator + "ftpserver" + File.separator;
        File file = new File(fileName);
        if (file.exists() == false) {file.mkdirs();}
        return fileName;
    }

    public static String getConfigurationFileName() {
        return (getAppdir() + "ftpserver.xml");
    }
    
    private TFTPProtocol.FTPProtocolData fData = new TFTPProtocol.FTPProtocolData();
    public synchronized TFTPProtocol.FTPProtocolData data() {return this.fData;}

    public synchronized void loadSettings(String xmlFileName) {
        this.data().loadFromXML(xmlFileName);
        if (this.data().getAllowTLS()) {
            try{this.fSSLSocketFactory = TTLSProvider.createSSLSocketFactory(KeyStoreHelper.createKeyManagers(this.data().getPEMEncodedServerCertAndKeyFileName(), this.data().getKeyPassphrase().toCharArray()), null, null);} catch (Exception e) {}
        }
    }

    public synchronized void loadSettings() {
        this.loadSettings(getConfigurationFileName());
    }

    public synchronized void saveSettings(String xmlFileName) {
        this.data().saveToXML(xmlFileName);
    }

    public synchronized void saveSettings() {
        this.saveSettings(getConfigurationFileName());
    }

    /**
     * A method to start logging
     * @param aLogFileName the file to which al events/message should be written
     */
    public synchronized void startLogging() {
        this.stopLogging();
        if (this.data().getMainLogFileName() != null) {
            try {
                this.fLogOutStream = new FileWriter(this.data().getMainLogFileName(), this.data().getLogAppend());
            } catch (Exception e) {
                this.stopLogging();
            }
        }
    }
    /**
     * Forms a handle to the log file
     */
    protected FileWriter fLogOutStream = null;
    /**
     * A method to be invoked to log a message/event to the logfile
     * @param aMessage a message or event that is to be written to disk
     */
    public synchronized void log(String aMessage) {
        if (this.fLogOutStream != null) {
            try{this.fLogOutStream.write(aMessage + "\r\n");} catch (Exception e) {}
            try{this.fLogOutStream.flush();} catch (Exception e) {}
        }
    }
    /**
     * A method to stop all logging
     */
    public synchronized void stopLogging() {
        try {if (this.fLogOutStream != null) {this.fLogOutStream.flush();}} catch (Exception e) {}
        try {if (this.fLogOutStream != null) {this.fLogOutStream.close();}} catch (Exception e) {}
        this.fLogOutStream = null;
    }

    protected SSLSocketFactory fSSLSocketFactory = null;

    /**
     * A method to close the stream to the log file
     * @throws java.lang.Throwable
     */
    @Override
    protected void finalize() throws Throwable {
        this.stopLogging();
    }
}
