/* ============================================================================
 * KeyStoreHelper.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or
 *      its contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 *
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 *
 * ============================================================================
 *
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 *
 * ============================================================================
 */
package nl.jhhcvandermeijs.tcpconns.keys;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.security.KeyPair;
import java.security.KeyStore;
import java.security.KeyStore.PrivateKeyEntry;
import java.security.KeyStore.SecretKeyEntry;
import java.security.KeyStore.TrustedCertificateEntry;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.security.auth.x500.X500Principal;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import nl.jhhcvandermeijs.tcpconns.utils.Utils;
import org.bouncycastle.asn1.ASN1Object;
import org.bouncycastle.asn1.ASN1OctetString;
import org.bouncycastle.asn1.ASN1Set;
import org.bouncycastle.asn1.DEREncodable;
import org.bouncycastle.asn1.DERObject;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.DERSet;
import org.bouncycastle.asn1.DERString;
import org.bouncycastle.asn1.pkcs.Attribute;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.x509.X509Extension;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.jce.PKCS10CertificationRequest;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openssl.PEMReader;
import org.bouncycastle.openssl.PEMWriter;
import org.bouncycastle.openssl.PasswordFinder;

/**
 * this class serves to make loading and saving keystores easier; with this class
 * you can read and write pkcs#12 keystores (.pfx, .p12), java key stores,
 * bouncycastle keystores, "uber" keystores, jceks keystores, and
 * openSSL's PEM keystores.
 */
public class KeyStoreHelper {

    /*
     * ***
     * *** CONSTANTS
     * ***
     */

        /**
         * constant denoting the "AES-128-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_128_CBC = "AES-128-CBC";
        /**
         * constant denoting the "AES-128-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_128_CFB = "AES-128-CFB";
        /**
         * constant denoting the "AES-128-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_128_ECB = "AES-128-ECB";
        /**
         * constant denoting the "AES-128-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_128_OFB = "AES-128-OFB";
        /**
         * constant denoting the "AES-192-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_192_CBC = "AES-192-CBC";
        /**
         * constant denoting the "AES-192-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_192_CFB = "AES-192-CFB";
        /**
         * constant denoting the "AES-192-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_192_ECB = "AES-192-ECB";
        /**
         * constant denoting the "AES-192-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_192_OFB = "AES-192-OFB";
        /**
         * constant denoting the "AES-256-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_256_CBC = "AES-256-CBC";
        /**
         * constant denoting the "AES-256-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_256_CFB = "AES-256-CFB";
        /**
         * constant denoting the "AES-256-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_256_ECB = "AES-256-ECB";
        /**
         * constant denoting the "AES-256-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_AES_256_OFB = "AES-256-OFB";
        /**
         * constant denoting the "BF-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_BF_CBC = "BF-CBC";
        /**
         * constant denoting the "BF-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_BF_CFB = "BF-CFB";
        /**
         * constant denoting the "BF-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_BF_ECB = "BF-ECB";
        /**
         * constant denoting the "BF-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_BF_OFB = "BF-OFB";
        /**
         * constant denoting the "DES-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_CBC = "DES-CBC";
        /**
         * constant denoting the "DES-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_CFB = "DES-CFB";
        /**
         * constant denoting the "DES-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_ECB = "DES-ECB";
        /**
         * constant denoting the "DES-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_OFB = "DES-OFB";
        /**
         * constant denoting the "DES-EDE" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE = "DES-EDE";
        /**
         * constant denoting the "DES-EDE-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE_CBC = "DES-EDE-CBC";
        /**
         * constant denoting the "DES-EDE-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE_CFB = "DES-EDE-CFB";
        /**
         * constant denoting the "DES-EDE-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE_ECB = "DES-EDE-ECB";
        /**
         * constant denoting the "DES-EDE-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE_OFB = "DES-EDE-OFB";
        /**
         * constant denoting the "DES-EDE3" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE3 = "DES-EDE3";
        /**
         * constant denoting the "DES-EDE3-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE3_CBC = "DES-EDE3-CBC";
        /**
         * constant denoting the "DES-EDE3-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE3_CFB = "DES-EDE3-CFB";
        /**
         * constant denoting the "DES-EDE3-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE3_ECB = "DES-EDE3-ECB";
        /**
         * constant denoting the "DES-EDE3-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_DES_EDE3_OFB = "DES-EDE3-OFB";
        /**
         * constant denoting the "RC2-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_RC2_CBC = "RC2-CBC";
        /**
         * constant denoting the "RC2-CFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_RC2_CFB = "RC2-CFB";
        /**
         * constant denoting the "RC2-ECB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_RC2_ECB = "RC2-ECB";
        /**
         * constant denoting the "RC2-OFB" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_RC2_OFB = "RC2-OFB";
        /**
         * constant denoting the "RC2-40-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_RC2_40_CBC = "RC2-40-CBC";
        /**
         * constant denoting the "RC2-64-CBC" algorithm for encryption when
         * storing private key information to PEM encoded files
         */
        public static final String PEM_ALGORITHM_RC2_64_CBC = "RC2-64-CBC";
        /**
         * constant denoting that no encryption is to be used when
         * storing private key information to PEM encoded files; private keys
         * will be stored unprotected
         */
        public static final String PEM_ALGORITHM_NO_ENCRYPTION = "NO_ENCRYPTION";

    /*
     * ***
     * *** PUBLIC OPEN METHODS
     * ***
     */

        /**
         * class method to open a keystore of type pkcs#12, jks, bks, uber,
         * jceks, pem/crt, or cer/der
         * @param fileName the file name of the keystore
         * @param passphrase the passphrase needed to unlock the data in the keystore
         * @return a KeyStore object when successful, otherwise null
         */
        public static KeyStore openKeyStore(String fileName, char[] passphrase) {
            File file = null;
            try {
                file = new File(fileName);
                return KeyStoreHelper.openKeyStore(file, passphrase);
            } catch (Exception e) {                
                return null;
            }
        }

        /**
         * class method to open a keystore of type pkcs#12, jks, bks, uber,
         * jceks, pem/crt, or cer/der
         * @param file the file object that points towards the keystore file
         * @param passphrase the passphrase needed to unlock the data in the keystore
         * @return a KeyStore object when successful, otherwise null
         */
        public static KeyStore openKeyStore(File file, char[] passphrase) {
            FileInputStream fileInputStream = null;
            try {
                try {
                    fileInputStream = new FileInputStream(file);
                    return KeyStoreHelper.openKeyStore(fileInputStream, passphrase);
                } finally {
                    try {fileInputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {                
                return null;
            }
        }

        /**
         * class method to open a keystore of type pkcs#12, jks, bks, uber,
         * jceks, pem/crt, or cer/der
         * @param inputStream stream that holds the series of bytes that form the keystore
         * @param passphrase the passphrase needed to unlock the data in the keystore
         * @return a KeyStore object when successful, otherwise null
         */
        public static KeyStore openKeyStore(InputStream inputStream, char[] passphrase) {
            ByteArrayOutputStream byteArrayOutputStream = null;
            ByteArrayInputStream byteArrayInputStream = null;
            try {
                byteArrayOutputStream = new ByteArrayOutputStream(inputStream.available());

                byte[] buffer = new byte[4096];
                int bytesRead = inputStream.read(buffer);
                while (bytesRead != -1) {
                    byteArrayOutputStream.write(buffer, 0, bytesRead);
                    bytesRead = inputStream.read(buffer);
                }
            } catch (Exception e) {                
                return null;
            }

            byteArrayInputStream = new ByteArrayInputStream(byteArrayOutputStream.toByteArray());
            byteArrayInputStream.mark(0);

            return KeyStoreHelper.loadKeyStore(byteArrayInputStream, passphrase);
        }

        /**
         * convenience method to load the keypair object from a pem encoded .key file
         * @param fileName file name of the .key file to be loaded
         * @param passphrase passphrase needed to unlock the .key data
         * @return a KeyPair object if successful, otherwise null
         */
        public static KeyPair loadKEYFile(String fileName, char[] passphrase) {
            File file = null;
            try {
                file = new File(fileName);
                return KeyStoreHelper.loadKEYFile(file, passphrase);
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method to load the keypair object from a pem encoded .key file
         * @param file the file that points to the .key file to be loaded
         * @param passphrase passphrase needed to unlock the .key data
         * @return a KeyPair object if successful, otherwise null
         */
        public static KeyPair loadKEYFile(File file, char[] passphrase) {
            FileInputStream fileInputStream = null;
            try {
                try {
                    fileInputStream = new FileInputStream(file);
                    return KeyStoreHelper.loadKEYFile(fileInputStream, passphrase);
                } finally {
                    try {fileInputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method to load the keypair object from a pem encoded .key file
         * @param inputStream the stream of bytes that contains the key object data
         * @param passphrase passphrase needed to unlock the .key data
         * @return a KeyPair object if successful, otherwise null
         */
        public static KeyPair loadKEYFile(InputStream inputStream, char[] passphrase) {
            ByteArrayOutputStream byteArrayOutputStream = null;
            ByteArrayInputStream byteArrayInputStream = null;
            try {
                byteArrayOutputStream = new ByteArrayOutputStream(inputStream.available());

                byte[] buffer = new byte[4096];
                int bytesRead = inputStream.read(buffer);
                while (bytesRead != -1) {
                    byteArrayOutputStream.write(buffer, 0, bytesRead);
                    bytesRead = inputStream.read(buffer);
                }
            } catch (Exception e) {
                return null;
            }

            byteArrayInputStream = new ByteArrayInputStream(byteArrayOutputStream.toByteArray());
            byteArrayInputStream.mark(0);

            return KeyStoreHelper.loadKEYFromStream(byteArrayInputStream, passphrase);
        }

        /**
         * convenience method to load an X509 certificate from a .der/.cer file or a
         * .pem/.crt file
         * @param fileName the file name of the certificate
         * @return an X509Certificate object if successful, otherwise null
         */
        public static X509Certificate loadX509(String fileName) {
            File file = null;
            try {
                file = new File(fileName);
                return KeyStoreHelper.loadX509(file);
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method to load an X509 certificate from a .der/.cer file or a
         * .pem/.crt file
         * @param file the file that points towards the certificate
         * @return an X509Certificate object if successful, otherwise null
         */
        public static X509Certificate loadX509(File file) {
            FileInputStream fileInputStream = null;
            try {
                try {
                    fileInputStream = new FileInputStream(file);
                    return KeyStoreHelper.loadX509(fileInputStream);
                } finally {
                    try {fileInputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method to load an X509 certificate from a .der/.cer or a
         * .pem/.crt encoded stream of bytes
         * @param inputStream the stream of bytes that contains the certificate
         * @return an X509Certificate object if successful, otherwise null
         */
        public static X509Certificate loadX509(InputStream inputStream) {
            ByteArrayOutputStream byteArrayOutputStream = null;
            ByteArrayInputStream byteArrayInputStream = null;
            try {
                byteArrayOutputStream = new ByteArrayOutputStream(inputStream.available());

                byte[] buffer = new byte[4096];
                int bytesRead = inputStream.read(buffer);
                while (bytesRead != -1) {
                    byteArrayOutputStream.write(buffer, 0, bytesRead);
                    bytesRead = inputStream.read(buffer);
                }
            } catch (Exception e) {
                return null;
            }

            byteArrayInputStream = new ByteArrayInputStream(byteArrayOutputStream.toByteArray());
            byteArrayInputStream.mark(0);

            return KeyStoreHelper.loadX509FromStream(byteArrayInputStream);
        }

    /*
     * ***
     * *** NON PUBLIC OPEN METHODS
     * ***
     */

        protected static X509Certificate loadX509FromStream(ByteArrayInputStream byteArrayInputStream) {
            KeyStoreHelper.assertProviders();
            X509Certificate x509 = null;
            CertificateFactory certificateFactory = null;
            try {
                certificateFactory = CertificateFactory.getInstance("X.509");
                x509 = ((X509Certificate)(certificateFactory.generateCertificate(byteArrayInputStream)));
                return x509;
            } catch (Exception e) {
                byteArrayInputStream.reset();
                PEMReader pemReader = null;
                Object object = null;
                x509 = null;
                try {
                    pemReader = new PEMReader(new BufferedReader(new InputStreamReader(byteArrayInputStream)));
                    object = pemReader.readObject();
                    while (object != null) {
                        if (object instanceof X509Certificate) {
                            x509 = ((X509Certificate)(object));
                            break;
                        } else {
                            object = pemReader.readObject();
                        }
                    }
                    return x509;
                } catch (Exception e2) {
                    return null;
                }
            }
        }

        protected static KeyPair loadKEYFromStream(ByteArrayInputStream byteArrayInputStream, char[] passphrase) {
            KeyStoreHelper.assertProviders();
            Reader reader = null;
            PEMReader pemReader = null;
            Object object = null;
            KeyPair keyPair = null;
            try {
                reader = new BufferedReader(new InputStreamReader(byteArrayInputStream));
                pemReader = new PEMReader(reader, new KeyStoreHelperPEMPasswordFinder(passphrase), "BC");

                object = pemReader.readObject();
                while (object != null) {
                    if (object instanceof KeyPair) {
                        keyPair = ((KeyPair)(object));
                        break;
                    } else {
                        object = pemReader.readObject();
                    }
                }
                return keyPair;
            } catch (Exception e2) {                
                return null;
            }
        }

        protected static KeyStore loadKeyStore(ByteArrayInputStream byteArrayInputStream, char[] passphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore keyStore = null;

            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadPFXNEWTYPE(byteArrayInputStream, passphrase)  ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadPFXBCPROV(byteArrayInputStream, passphrase)   ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadPFXSUNPROV(byteArrayInputStream, passphrase)  ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadJKS(byteArrayInputStream, passphrase)  ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadBKS(byteArrayInputStream, passphrase)  ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadUBER(byteArrayInputStream, passphrase) ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadJCEKS(byteArrayInputStream, passphrase); if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadPEM(byteArrayInputStream, passphrase)  ; if (keyStore != null) {return keyStore;}
            byteArrayInputStream.reset(); keyStore = KeyStoreHelper.loadDER(byteArrayInputStream)              ; if (keyStore != null) {return keyStore;}

            return null;
        }

        protected static KeyStore loadPFXNEWTYPE(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("PKCS12-3DES-3DES", "BC");
                keyStore.load(inputStream, passphrase);
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadPFXBCPROV(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("PKCS12", "BC");
                keyStore.load(inputStream, passphrase);
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadPFXSUNPROV(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("PKCS12");
                keyStore.load(inputStream, passphrase);
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadJKS(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("jks");
                keyStore.load(inputStream, passphrase);
                return keyStore;                
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadBKS(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("BKS", "BC");
                keyStore.load(inputStream, passphrase);
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadUBER(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("UBER", "BC");
                keyStore.load(inputStream, passphrase);
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadJCEKS(InputStream inputStream, char[] passphrase) {
            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("JCEKS");
                keyStore.load(inputStream, passphrase);
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        protected static KeyStore loadPEM(InputStream inputStream, char[] passphrase) {
            KeyStoreHelper.assertProviders();

            List<Object> pemEntries = new LinkedList<Object>();
            Object[] pemEntriesArray = new Object[0];
            Object[] markForRemoval = new Object[0];

            {
                Reader reader = null;
                PEMReader pemReader = null;
                Object anObject = null;
                try {
                    reader = new BufferedReader(new InputStreamReader(inputStream));
                    pemReader = new PEMReader(reader, new KeyStoreHelperPEMPasswordFinder(passphrase), "BC");

                    while (true) {
                        anObject = pemReader.readObject();
                        if (anObject == null) {break;}
                        if (anObject instanceof KeyPair) {
                            KeyPair keyPair = ((KeyPair)(anObject));
                            pemEntries.add(keyPair);
                        } else if (anObject instanceof X509Certificate) {
                            X509Certificate x509 = ((X509Certificate)(anObject));
                            pemEntries.add(x509);
                        }
                    }

                } catch (Exception e) {} finally {
                    try {pemReader.close();} catch (Exception e) {}
                    try {reader.close();} catch (Exception e) {}
                }
            }

            pemEntriesArray = pemEntries.toArray(pemEntriesArray);
            markForRemoval = new Object[pemEntriesArray.length];
            return processPEMEntries(pemEntriesArray, markForRemoval, passphrase);
        }

        protected static KeyStore loadDER(InputStream inputStream) {
            KeyStore keyStore = null;
            X509Certificate x509 = null;
            CertificateFactory certificateFactory = null;
            try {
                certificateFactory = CertificateFactory.getInstance("X.509");
                x509 = ((X509Certificate)(certificateFactory.generateCertificate(inputStream)));

                keyStore = KeyStore.getInstance("jks");
                keyStore.load(null, null);

                keyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
            } catch (Exception e) {return null;}
            return keyStore;
        }

    /*
     * ***
     * *** NON PUBLIC PEM UTILITY CLASSES AND ROUTINES
     * ***
     */

        protected static class KeyStoreHelperPEMPasswordFinder implements PasswordFinder {
            protected char[] fPassphrase = null;
            public KeyStoreHelperPEMPasswordFinder(char[] passphrase) {
                this.fPassphrase = passphrase;
            }
            public char[] getPassword() {
                return this.fPassphrase;
            }
        }

        protected static boolean keyPairsAvailable(Object[] pemEntriesArray) {
            int i = 1-1; int n = pemEntriesArray.length;
            while (i <= (n-1)) {
                Object object = pemEntriesArray[i];
                if (object instanceof KeyPair) {
                    return true;
                }
                i++;
            }
            return false;
        }

        public static boolean samePublicKey(PublicKey publicKey1, PublicKey publicKey2) {
            byte[] pk1 = publicKey1.getEncoded();
            byte[] pk2 = publicKey2.getEncoded();
            if (pk1.length != pk2.length) {return false;}
            int i = 1-1; int n = pk1.length;
            while (i <= (n-1)) {
                if (pk1[i] != pk2[i]) {
                    return false;
                }
                i++;
            }
            return true;
        }

        protected static X509Certificate getCertificate(Object[] pemEntriesArray, Object[] markForRemoval, KeyPair keyPair) {
            int i = 1-1; int n = pemEntriesArray.length;
            while (i <= (n-1)) {
                Object object = pemEntriesArray[i];
                if (object instanceof X509Certificate) {
                    X509Certificate x509 = ((X509Certificate)(object));
                    PublicKey publicKey = x509.getPublicKey();
                    if (KeyStoreHelper.samePublicKey(publicKey, keyPair.getPublic())) {
                        markForRemoval[i] = x509;
                        return x509;
                    }
                }
                i++;
            }
            return null;
        }

        protected static X509Certificate getSigner(Object[] pemEntriesArray, Object[] markForRemoval, X509Certificate x509) {
            int i = 1-1; int n = pemEntriesArray.length;
            while (i <= (n-1)) {
                Object object = pemEntriesArray[i];
                if (object instanceof X509Certificate) {
                    X509Certificate signer = ((X509Certificate)(object));
                    try {
                        x509.verify(signer.getPublicKey());
                        try {
                            x509.verify(x509.getPublicKey());
                            markForRemoval[i] = signer;
                            return null; //if self signed prevent eternal loop
                        } catch (Exception e2) {
                            markForRemoval[i] = signer;
                            return signer;
                        }
                    } catch (Exception e) {}
                }
                i++;
            }
            return null;
        }

        protected static X509Certificate[] getCertificateChain(Object[] pemEntriesArray, Object[] markForRemoval, KeyPair keyPair) {
            List<X509Certificate> list = new LinkedList<X509Certificate>();
            X509Certificate x509 = null;
            X509Certificate[] x509Array = new X509Certificate[0];

            x509 = KeyStoreHelper.getCertificate(pemEntriesArray, markForRemoval, keyPair);
            if (x509 == null) {return null;}
            list.add(x509);

            while (x509 != null) {
                x509 = KeyStoreHelper.getSigner(pemEntriesArray, markForRemoval, x509);
                if (x509 != null) {
                    list.add(x509);
                }
            }

            if (list.size() == 0) {
                return null;
            } else {
                x509Array = list.toArray(x509Array);
                return x509Array;
            }
        }

        protected static Object[] removeDuplicates(Object[] pemEntriesArray) {
            int i = 1-1; int n = pemEntriesArray.length;
            while (i <= (n-1)) {
                if (i != (n-1)) {

                    int m = pemEntriesArray.length; int j = m-1;
                    while (j >= (i + 1)) {

                        Object obj1 = pemEntriesArray[i];
                        Object obj2 = pemEntriesArray[j];

                        if (obj1 instanceof KeyPair) {
                            if (obj2 instanceof KeyPair) {
                                KeyPair keyPair1 = ((KeyPair)(obj1));
                                KeyPair keyPair2 = ((KeyPair)(obj2));
                                if (KeyStoreHelper.samePublicKey(keyPair1.getPublic(), keyPair2.getPublic())) {
                                    pemEntriesArray[j] = null; //identical entry found!
                                }
                            }
                        } else if (obj1 instanceof X509Certificate) {
                            if (obj2 instanceof X509Certificate) {
                                X509Certificate x509_1 = ((X509Certificate)(obj1));
                                X509Certificate x509_2 = ((X509Certificate)(obj2));
                                if (KeyStoreHelper.samePublicKey(x509_1.getPublicKey(), x509_2.getPublicKey())) {
                                    pemEntriesArray[j] = null; //identical entry found!
                                }
                            }
                        }

                        j = j - 1;
                    }

                }
                i = i + 1;
            }

            return pemEntriesArray;
        }

        protected static KeyStore processPEMEntries(Object[] pemEntriesArray, Object[] markForRemoval, char[] passphrase) {
            if (pemEntriesArray.length == 0) {return null;}
            int j = 1-1; int m = markForRemoval.length;
            while (j <= (m-1)) {
                Object object = pemEntriesArray[j]; //remove all objects that aren't keypairs nor certificates
                if (((object instanceof KeyPair) || (object instanceof X509Certificate)) != true) {
                    pemEntriesArray[j] = null;
                }
                markForRemoval[j] = null;
                j++;
            }
            pemEntriesArray = KeyStoreHelper.removeDuplicates(pemEntriesArray);

            KeyStore keyStore = null;
            try {
                keyStore = KeyStore.getInstance("JKS");
                keyStore.load(null, null);
            } catch (Exception e) {return null;}

            //first retrieve all private keys including certificate chains (must be present in .pem file!!)
            while (KeyStoreHelper.keyPairsAvailable(pemEntriesArray)) {
                int i = 1-1; int n = pemEntriesArray.length;
                while (i <= (n-1)) {
                    Object object = pemEntriesArray[i];
                    if (object instanceof KeyPair) {
                        X509Certificate[] chain = KeyStoreHelper.getCertificateChain(pemEntriesArray, markForRemoval, ((KeyPair)(object)));
                        if (chain != null) {
                            if (chain.length > 0) {
                                if (chain[0] != null) {
                                    try {
                                        keyStore.setKeyEntry(KeyStoreHelper.getAlias(chain[0]), ((KeyPair)(object)).getPrivate(), passphrase, chain);
                                    } catch (Exception e) {}
                                }
                            }
                        }
                        pemEntriesArray[i] = null;
                    }
                    i++;
                }
            }

            j = 1-1; m = markForRemoval.length;
            while (j <= (m-1)) {
                if (markForRemoval[j] != null) {
                    markForRemoval[j] = null;
                    pemEntriesArray[j] = null;
                }
                j++;
            }

            //then store all trusted certificates, since they are the only ones left
            int i = 1-1; int n = pemEntriesArray.length;
            while (i <= (n-1)) {
                Object object = pemEntriesArray[i];
                if (object instanceof X509Certificate) {
                    X509Certificate x509 = ((X509Certificate)(object));
                    try {
                        keyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                    } catch (Exception e) {}
                    pemEntriesArray[i] = null;
                }
                i++;
            }

            //return the keyStore if there are actually items in the keystore
            try {
                Enumeration aliases = keyStore.aliases();
                if (aliases.hasMoreElements() == false) {
                    return null;
                } else {
                    return keyStore;
                }
            } catch (Exception e) {return keyStore;}
        }

    /*
     * ***
     * *** PUBLIC UTILITY ROUTINES
     * ***
     */

        /**
         * convenience method that returns an empty KeyStore object
         * @return a KeyStore object if successful, otherwise null
         */
        public static KeyStore newEmptyKeyStore() {
            KeyStoreHelper.assertProviders();
            KeyStore pkcs12Store = null;
            try {
                pkcs12Store = KeyStore.getInstance("PKCS12-3DES-3DES", "BC");
                pkcs12Store.load(null, null);
                return pkcs12Store;
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method that can be used to import a keyStore into another keyStore; with duplicate aliases/names,
         * items in the toKeyStore are replaced/overwritten
         * @param keyStore the keyStore that is to be imported into the toKeyStore object
         * @param passphrase the passphrase to unlock private key items in the keyStore object
         * @param toKeyStore the KeyStore object to which items should be added
         * @param toKeyStorePassphrase the passphrase to lock private key items in the toKeyStore object
         * @return toKeyStore if successful, otherwise null
         */
        public static KeyStore importDoOverwrite(KeyStore keyStore, char[] passphrase, KeyStore toKeyStore, char[] toKeyStorePassphrase) {
            try {
                Enumeration<String> aliases = keyStore.aliases();
                String alias = null;
                while (aliases.hasMoreElements()) {
                    alias = aliases.nextElement();

                    if (keyStore.entryInstanceOf(alias, PrivateKeyEntry.class)) {
                        try {
                            PrivateKeyEntry privateKeyEntry = ((PrivateKeyEntry)(keyStore.getEntry(alias, new KeyStore.PasswordProtection(passphrase))));
                            KeyStoreHelper.importDoOverwrite(privateKeyEntry, toKeyStore, toKeyStorePassphrase);
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, SecretKeyEntry.class)) {
                        try {
                            //for now do nothing
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, TrustedCertificateEntry.class)) {
                        try {
                            TrustedCertificateEntry trustedCertificateEntry = ((TrustedCertificateEntry)(keyStore.getEntry(alias, null)));
                            KeyStoreHelper.importDoOverwrite(trustedCertificateEntry, toKeyStore);
                        } catch (Exception e2) {}

                    }
                }
                
                return toKeyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importDoOverwrite(X509Certificate x509, KeyStore toKeyStore) {
            try {
                toKeyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                return toKeyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importDoOverwrite(TrustedCertificateEntry trustedCertificateEntry, KeyStore toKeyStore) {
            try {
                if (trustedCertificateEntry.getTrustedCertificate() instanceof X509Certificate) {
                    X509Certificate x509 = ((X509Certificate)(trustedCertificateEntry.getTrustedCertificate()));
                    toKeyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                    return toKeyStore;
                } else {
                    return null;
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importDoOverwrite(PrivateKeyEntry privateKeyEntry, KeyStore toKeyStore, char[] toKeyStorePassphrase) {
            try {
                PrivateKey privateKey = privateKeyEntry.getPrivateKey();
                if (privateKeyEntry.getCertificateChain() instanceof X509Certificate[]) {
                    X509Certificate[] chain = ((X509Certificate[])(privateKeyEntry.getCertificateChain()));
                    if (chain.length > 0) {
                        X509Certificate x509 = chain[0];
                        toKeyStore.setKeyEntry(KeyStoreHelper.getAlias(x509), privateKey, toKeyStorePassphrase, chain);
                        return toKeyStore;
                    } else {
                        return null;
                    }
                } else {
                    return null;
                }
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method that can be used to import a keyStore into another keyStore; with duplicate aliases/names,
         * items in the toKeyStore are NOT replaced/overwritten
         * @param keyStore the keyStore that is to be imported into the toKeyStore object
         * @param passphrase the passphrase to unlock private key items in the keyStore object
         * @param toKeyStore the KeyStore object to which items should be added
         * @param toKeyStorePassphrase the passphrase to lock private key items in the toKeyStore object
         * @return toKeyStore if successful, otherwise null
         */
        public static KeyStore importNoOverwrite(KeyStore keyStore, char[] passphrase, KeyStore toKeyStore, char[] toKeyStorePassphrase) {
            try {
                Enumeration<String> aliases = keyStore.aliases();
                String alias = null;
                while (aliases.hasMoreElements()) {
                    alias = aliases.nextElement();

                    if (keyStore.entryInstanceOf(alias, PrivateKeyEntry.class)) {
                        try {
                            PrivateKeyEntry privateKeyEntry = ((PrivateKeyEntry)(keyStore.getEntry(alias, new KeyStore.PasswordProtection(passphrase))));
                            KeyStoreHelper.importNoOverwrite(privateKeyEntry, toKeyStore, toKeyStorePassphrase);
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, SecretKeyEntry.class)) {
                        try {
                            //for now do nothing
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, TrustedCertificateEntry.class)) {
                        try {
                            TrustedCertificateEntry trustedCertificateEntry = ((TrustedCertificateEntry)(keyStore.getEntry(alias, null)));
                            KeyStoreHelper.importNoOverwrite(trustedCertificateEntry, toKeyStore);
                        } catch (Exception e2) {}

                    }
                }

                return toKeyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importNoOverwrite(X509Certificate x509, KeyStore toKeyStore) {
            try {
                Enumeration<String> aliases = toKeyStore.aliases();
                String alias = null;
                while (aliases.hasMoreElements()) {
                    alias = aliases.nextElement();
                    if (alias.equals(KeyStoreHelper.getAlias(x509))) {
                        return null;
                    }
                }
                
                toKeyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                return toKeyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importNoOverwrite(TrustedCertificateEntry trustedCertificateEntry, KeyStore toKeyStore) {
            try {
                if (trustedCertificateEntry.getTrustedCertificate() instanceof X509Certificate) {
                    X509Certificate x509 = ((X509Certificate)(trustedCertificateEntry.getTrustedCertificate()));

                    Enumeration<String> aliases = toKeyStore.aliases();
                    String alias = null;
                    while (aliases.hasMoreElements()) {
                        alias = aliases.nextElement();
                        if (alias.equals(KeyStoreHelper.getAlias(x509))) {
                            return null;
                        }
                    }

                    toKeyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                    return toKeyStore;
                } else {
                    return null;
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importNoOverwrite(PrivateKeyEntry privateKeyEntry, KeyStore toKeyStore, char[] passphrase) {
            try {
                PrivateKey privateKey = privateKeyEntry.getPrivateKey();
                if (privateKeyEntry.getCertificateChain() instanceof X509Certificate[]) {
                    X509Certificate[] chain = ((X509Certificate[])(privateKeyEntry.getCertificateChain()));
                    if (chain.length > 0) {
                        X509Certificate x509 = chain[0];

                        Enumeration<String> aliases = toKeyStore.aliases();
                        String alias = null;
                        while (aliases.hasMoreElements()) {
                            alias = aliases.nextElement();
                            if (alias.equals(KeyStoreHelper.getAlias(x509))) {
                                return null;
                            }
                        }

                        toKeyStore.setKeyEntry(KeyStoreHelper.getAlias(x509), privateKey, passphrase, chain);
                        return toKeyStore;
                    } else {
                        return null;
                    }
                } else {
                    return null;
                }
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * convenience method that can be used to import a keyStore into another keyStore; with duplicate aliases/names,
         * items in the toKeyStore are replaced/overwritten only after the user accepts the action by choosing [YES] on
         * a confirmation dialog
         * @param keyStore the keyStore that is to be imported into the toKeyStore object
         * @param passphrase the passphrase to unlock private key items in the keyStore object
         * @param toKeyStore the KeyStore object to which items should be added
         * @param toKeyStorePassphrase the passphrase to lock private key items in the toKeyStore object
         * @return toKeyStore if successful, otherwise null
         */
        public static KeyStore importConfirmOverwrite(KeyStore keyStore, char[] passphrase, KeyStore toKeyStore, char[] toKeyStorePassphrase) {
            try {
                Enumeration<String> aliases = keyStore.aliases();
                String alias = null;
                while (aliases.hasMoreElements()) {
                    alias = aliases.nextElement();

                    if (keyStore.entryInstanceOf(alias, PrivateKeyEntry.class)) {
                        try {
                            PrivateKeyEntry privateKeyEntry = ((PrivateKeyEntry)(keyStore.getEntry(alias, new KeyStore.PasswordProtection(passphrase))));
                            KeyStoreHelper.importConfirmOverwrite(privateKeyEntry, toKeyStore, toKeyStorePassphrase);
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, SecretKeyEntry.class)) {
                        try {
                            //for now do nothing
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, TrustedCertificateEntry.class)) {
                        try {
                            TrustedCertificateEntry trustedCertificateEntry = ((TrustedCertificateEntry)(keyStore.getEntry(alias, null)));
                            KeyStoreHelper.importConfirmOverwrite(trustedCertificateEntry, toKeyStore);
                        } catch (Exception e2) {}

                    }
                }

                return toKeyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importConfirmOverwrite(X509Certificate x509, KeyStore toKeyStore) {
            try {
                Enumeration<String> aliases = toKeyStore.aliases();
                String alias = null;
                while (aliases.hasMoreElements()) {
                    alias = aliases.nextElement();
                    if (alias.equals(KeyStoreHelper.getAlias(x509))) {
                        int choice = JOptionPane.showConfirmDialog(null, "\"" + alias + "\" already exists. Press [YES] to replace the item. ", "Duplicate name entry", JOptionPane.YES_NO_CANCEL_OPTION);
                        if (choice != JOptionPane.YES_OPTION) {
                            return null;
                        }
                    }
                }

                toKeyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                return toKeyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importConfirmOverwrite(TrustedCertificateEntry trustedCertificateEntry, KeyStore toKeyStore) {
            try {
                if (trustedCertificateEntry.getTrustedCertificate() instanceof X509Certificate) {
                    X509Certificate x509 = ((X509Certificate)(trustedCertificateEntry.getTrustedCertificate()));

                    Enumeration<String> aliases = toKeyStore.aliases();
                    String alias = null;
                    while (aliases.hasMoreElements()) {
                        alias = aliases.nextElement();
                        if (alias.equals(KeyStoreHelper.getAlias(x509))) {
                            int choice = JOptionPane.showConfirmDialog(null, "\"" + alias + "\" already exists. Press [YES] to replace the item. ", "Duplicate name entry", JOptionPane.YES_NO_CANCEL_OPTION);
                            if (choice != JOptionPane.YES_OPTION) {
                                return null;
                            }
                        }
                    }
                    toKeyStore.setCertificateEntry(KeyStoreHelper.getAlias(x509), x509);
                    return toKeyStore;
                } else {
                    return null;
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore importConfirmOverwrite(PrivateKeyEntry privateKeyEntry, KeyStore toKeyStore, char[] passphrase) {
            try {
                PrivateKey privateKey = privateKeyEntry.getPrivateKey();
                if (privateKeyEntry.getCertificateChain() instanceof X509Certificate[]) {
                    X509Certificate[] chain = ((X509Certificate[])(privateKeyEntry.getCertificateChain()));
                    if (chain.length > 0) {
                        X509Certificate x509 = chain[0];

                        Enumeration<String> aliases = toKeyStore.aliases();
                        String alias = null;
                        while (aliases.hasMoreElements()) {
                            alias = aliases.nextElement();
                            if (alias.equals(KeyStoreHelper.getAlias(x509))) {
                                int choice = JOptionPane.showConfirmDialog(null, "\"" + alias + "\" already exists. Press [YES] to replace the item. ", "Duplicate name entry", JOptionPane.YES_NO_CANCEL_OPTION);
                                if (choice != JOptionPane.YES_OPTION) {
                                    return null;
                                }
                            }
                        }

                        toKeyStore.setKeyEntry(KeyStoreHelper.getAlias(x509), privateKey, passphrase, chain);
                        return toKeyStore;
                    } else {
                        return null;
                    }
                } else {
                    return null;
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static String openDialog(javax.swing.JFrame parent, String openingDir) {
            if (openingDir == null) {
                openingDir = (new File(System.getProperty("user.home"))).getAbsolutePath();
            }
            JFileChooser jFileChooser = new JFileChooser();
            if (openingDir != null) {jFileChooser.setCurrentDirectory(new File(openingDir));}

            int returnValue = jFileChooser.showOpenDialog(parent);
            if (returnValue == JFileChooser.APPROVE_OPTION) {
                try {
                    File file = jFileChooser.getSelectedFile();
                    if (file != null) {
                        if (file.exists() && file.isFile()) {
                            return file.getCanonicalPath();
                        }
                    }
                } catch (Exception e) {}
            }
            return null;
        }

        public static String saveDialog(javax.swing.JFrame parent, String openingDir) {
            if (openingDir == null) {
                openingDir = (new File(System.getProperty("user.home"))).getAbsolutePath();
            }
            JFileChooser jFileChooser = new JFileChooser();
            if (openingDir != null) {jFileChooser.setCurrentDirectory(new File(openingDir));}

            int returnValue = jFileChooser.showSaveDialog(parent);
            if (returnValue == JFileChooser.APPROVE_OPTION) {
                try {
                    File file = jFileChooser.getSelectedFile();
                    if (file != null) {
                        if (file.exists()) {
                            int choice = JOptionPane.showConfirmDialog(parent, "File \"" + file.getCanonicalPath() + "\" already exists! Pressing [YES] will overwrite it?", "File overwrite warning", JOptionPane.YES_NO_CANCEL_OPTION);
                            if (choice != JOptionPane.YES_OPTION) {
                                return null;
                            }
                        }
                        return file.getCanonicalPath();
                    }
                } catch (Exception e) {}
            }
            return null;
        }

        public static char[] enterPassphrase(java.awt.Frame parent) {
            return EnterPassphraseDialog.enterPassphrase(parent);
        }

        public static void assertProviders() {
            try{if (Security.getProvider("BC") == null) {Security.addProvider(new BouncyCastleProvider());}} catch (Exception e) {}
        }

        public static boolean isSelfSigned(X509Certificate x509) {
            try {
                x509.verify(x509.getPublicKey());
                return true;
            } catch (Exception e) {
                return false;
            }
        }

        /**
         * this method returns true if the x509 is signed by trustedCA and the
         * x509 is within the validity period determined by trustedCA
         *
         * @param x509
         * @param trustedCA
         * @return true if not invalid (i.e. you must still check certificate
         * revocation status!!!)
         */
         public static boolean isValid(X509Certificate x509, X509Certificate trustedCA) {
            return KeyStoreHelper.isSignedBy(x509, trustedCA);
         }

        /**
         * this method returns true if the x509 is signed by trustedCA and the
         * x509 is within the validity period determined by trustedCA
         *
         * @param x509
         * @param trustedCA
         * @return true if not invalid (i.e. you must still check certificate
         * revocation status!!!)
         */
        public static boolean isSignedBy(X509Certificate x509, X509Certificate trustedCA) {
            try {
                if (KeyStoreHelper.isSelfSigned(x509) == true) {
                    return false;
                }
                x509.checkValidity();
                x509.verify(trustedCA.getPublicKey());
                return true;
            } catch (Exception e) {}
            return false;
        }
        
        public static boolean isValid(X509Certificate x509, X509Certificate[] trustedCAs) {
            return KeyStoreHelper.isSignedBy(x509, trustedCAs);
        }

        public static boolean isSignedBy(X509Certificate x509, X509Certificate[] trustedCAs) {
            try {
                if (KeyStoreHelper.isSelfSigned(x509) == true) {
                    return false;
                }

                X509Certificate trustedCA = null;
                int i = 1-1; int n = trustedCAs.length;
                while (i <= (n-1)) {
                    trustedCA = trustedCAs[i];
                    if ((KeyStoreHelper.isSignedBy(x509, trustedCA)) == true) {
                        return true;
                    }
                    i = i + 1;
                }
            } catch (Exception e) {}
            return false;
        }

        public static boolean isValid(X509Certificate x509, KeyStore trustStore) {
            return KeyStoreHelper.isSignedBy(x509, trustStore);
        }

        public static boolean isSignedBy(X509Certificate x509, KeyStore trustStore) {
            try {
                if (KeyStoreHelper.isSelfSigned(x509) == true) {
                    return false;
                }
                X509Certificate trustedCA = null;
                Enumeration<String> aliases = trustStore.aliases();
                while (aliases.hasMoreElements()) {
                    String alias = aliases.nextElement();
                    if (trustStore.entryInstanceOf(alias, TrustedCertificateEntry.class)) {
                        try {
                            TrustedCertificateEntry tce = ((TrustedCertificateEntry)(trustStore.getEntry(alias, null)));
                            if (tce.getTrustedCertificate() instanceof X509Certificate) {
                                trustedCA = ((X509Certificate)(tce.getTrustedCertificate()));
                                if ((KeyStoreHelper.isSignedBy(x509, trustedCA)) == true) {
                                    return true;
                                }
                            }
                        } catch (Exception e2) {}
                    }
                }
            } catch (Exception e) {}
            return false;
        }

        /**
         * Use this class method to obtain a keyManager[] from a keystore, which is used to
         * authenticate this side of an SSL/TLS connection by the other side
         * @param fileName
         * @param passphrase
         * @return a KeyManager[] if successful, otherwise null
         */
        public static KeyManager[] createKeyManagers(String fileName, char[] passphrase) {
            KeyStore keyStore = KeyStoreHelper.openKeyStore(fileName, passphrase);
            if (keyStore == null) {return null;}
            return KeyStoreHelper.createKeyManagers(keyStore, passphrase);
        }

        /**
         * Use this class method to obtain a keyManager[] from a keystore, which is used to
         * authenticate this side of an SSL/TLS connection by the other side
         * @param keyStore
         * @param passphrase
         * @return a KeyManager[] if successful, otherwise null
         */
        public static KeyManager[] createKeyManagers(KeyStore keyStore, char[] passphrase) {
            KeyStoreHelper.assertProviders();

            KeyManagerFactory keyManagerFactory = null;
            try {
                keyManagerFactory = KeyManagerFactory.getInstance("SunX509");
                keyManagerFactory.init(keyStore, passphrase);
                return keyManagerFactory.getKeyManagers();
            } catch (Exception e) {
                return null;
            }
        }

    /*
     * ***
     * *** NON PUBLIC STORE METHODS
     * ***
     */

        protected static void addKeyStoreEntriesTo(KeyStore fromKeyStore, char[] fromKeyStorePassphrase, KeyStore toKeyStore, char[] toKeyStorePassphrase) {
            try {
                Enumeration<String> aliases = fromKeyStore.aliases();
                while (aliases.hasMoreElements()) {
                    String alias = aliases.nextElement();

                    if (fromKeyStore.entryInstanceOf(alias, PrivateKeyEntry.class)) {
                        try {
                            PrivateKeyEntry pke = ((PrivateKeyEntry)(fromKeyStore.getEntry(alias, new KeyStore.PasswordProtection(fromKeyStorePassphrase))));
                            X509Certificate[] chain = ((X509Certificate[])(pke.getCertificateChain()));
                            PrivateKey pk = pke.getPrivateKey();
                            toKeyStore.setKeyEntry(KeyStoreHelper.getAlias(chain[0]), pk, toKeyStorePassphrase, chain);
                        } catch (Exception e2) {}

                    } else if (fromKeyStore.entryInstanceOf(alias, SecretKeyEntry.class)) {
                        try {
                            //for now do nothing
                        } catch (Exception e2) {}

                    } else if (fromKeyStore.entryInstanceOf(alias, TrustedCertificateEntry.class)) {
                        try {
                            TrustedCertificateEntry tce = ((TrustedCertificateEntry)(fromKeyStore.getEntry(alias, null)));
                            toKeyStore.setCertificateEntry(alias, tce.getTrustedCertificate());
                        } catch (Exception e2) {}

                    }

                }
            } catch (Exception e) {
                System.out.println(e);
            }
        }

    /*
     * ***
     * *** PUBLIC STORE METHODS
     * ***
     */

        public static KeyStore saveAsPFXNewType(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsPFXNewType(keyStore, passphrase, destinationFile, destinationPassphrase);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPFXNewType(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsPFXNewType(keyStore, passphrase, fileOutputStream, destinationPassphrase);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPFXNewType(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore pfxStore = null;
            try {
                pfxStore = KeyStore.getInstance("PKCS12-3DES-3DES", "BC");
                pfxStore.load(null, null);
                KeyStoreHelper.addKeyStoreEntriesTo(keyStore, passphrase, pfxStore, destinationPassphrase);
                pfxStore.store(destinationOutputStream, destinationPassphrase);
                return pfxStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPFXOldType(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsPFXOldType(keyStore, passphrase, destinationFile, destinationPassphrase);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPFXOldType(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsPFXOldType(keyStore, passphrase, fileOutputStream, destinationPassphrase);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPFXOldType(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore pfxStore = null;
            try {
                pfxStore = KeyStore.getInstance("PKCS12");
                pfxStore.load(null, null);
                KeyStoreHelper.addKeyStoreEntriesTo(keyStore, passphrase, pfxStore, destinationPassphrase);
                pfxStore.store(destinationOutputStream, destinationPassphrase);
                return pfxStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsJKS(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsJKS(keyStore, passphrase, destinationFile, destinationPassphrase);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsJKS(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsJKS(keyStore, passphrase, fileOutputStream, destinationPassphrase);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsJKS(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore jksStore = null;
            try {
                jksStore = KeyStore.getInstance("jks");
                jksStore.load(null, null);
                KeyStoreHelper.addKeyStoreEntriesTo(keyStore, passphrase, jksStore, destinationPassphrase);
                jksStore.store(destinationOutputStream, destinationPassphrase);
                return jksStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsBKS(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsBKS(keyStore, passphrase, destinationFile, destinationPassphrase);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsBKS(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsBKS(keyStore, passphrase, fileOutputStream, destinationPassphrase);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsBKS(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore bksStore = null;
            try {
                bksStore = KeyStore.getInstance("BKS", "BC");
                bksStore.load(null, null);
                KeyStoreHelper.addKeyStoreEntriesTo(keyStore, passphrase, bksStore, destinationPassphrase);
                bksStore.store(destinationOutputStream, destinationPassphrase);
                return bksStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsUBER(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsUBER(keyStore, passphrase, destinationFile, destinationPassphrase);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsUBER(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsUBER(keyStore, passphrase, fileOutputStream, destinationPassphrase);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsUBER(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore uberStore = null;
            try {
                uberStore = KeyStore.getInstance("UBER", "BC");
                uberStore.load(null, null);
                KeyStoreHelper.addKeyStoreEntriesTo(keyStore, passphrase, uberStore, destinationPassphrase);
                uberStore.store(destinationOutputStream, destinationPassphrase);
                return uberStore;
            } catch (Exception e) {return null;}
        }

        public static KeyStore saveAsJCEKS(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsJCEKS(keyStore, passphrase, destinationFile, destinationPassphrase);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsJCEKS(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsJCEKS(keyStore, passphrase, fileOutputStream, destinationPassphrase);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsJCEKS(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            KeyStoreHelper.assertProviders();
            KeyStore jceksStore = null;
            try {
                jceksStore = KeyStore.getInstance("JCEKS");
                jceksStore.load(null, null);
                KeyStoreHelper.addKeyStoreEntriesTo(keyStore, passphrase, jceksStore, destinationPassphrase);
                jceksStore.store(destinationOutputStream, destinationPassphrase);
                return jceksStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPEM(KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            return KeyStoreHelper.saveAsPEM(PEM_ALGORITHM_AES_256_CFB, keyStore, passphrase, destinationFileName, destinationPassphrase);
        }

        public static KeyStore saveAsPEM(KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            return KeyStoreHelper.saveAsPEM(PEM_ALGORITHM_AES_256_CFB, keyStore, passphrase, destinationFile, destinationPassphrase);
        }

        public static KeyStore saveAsPEM(KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            return KeyStoreHelper.saveAsPEM(PEM_ALGORITHM_AES_256_CFB, keyStore, passphrase, destinationOutputStream, destinationPassphrase);
        }

        public static KeyStore saveAsPEM(String algorithm, KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase) {
            return KeyStoreHelper.saveAsPEM(algorithm, keyStore, passphrase, destinationFileName, destinationPassphrase, null);
        }

        public static KeyStore saveAsPEM(String algorithm, KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase) {
            return KeyStoreHelper.saveAsPEM(algorithm, keyStore, passphrase, destinationFile, destinationPassphrase, null);
        }

        public static KeyStore saveAsPEM(String algorithm, KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase) {
            return KeyStoreHelper.saveAsPEM(algorithm, keyStore, passphrase, destinationOutputStream, destinationPassphrase, null);
        }

        public static KeyStore saveAsPEM(String algorithm, KeyStore keyStore, char[] passphrase, String destinationFileName, char[] destinationPassphrase, SecureRandom secureRandom) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsPEM(algorithm, keyStore, passphrase, destinationFile, destinationPassphrase, secureRandom);
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPEM(String algorithm, KeyStore keyStore, char[] passphrase, File destinationFile, char[] destinationPassphrase, SecureRandom secureRandom) {
            FileOutputStream outStream = null;
            try {
                try {
                    outStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsPEM(algorithm, keyStore, passphrase, outStream, destinationPassphrase, secureRandom);
                } finally {
                    try {outStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static KeyStore saveAsPEM(String algorithm, KeyStore keyStore, char[] passphrase, OutputStream destinationOutputStream, char[] destinationPassphrase, SecureRandom secureRandom) {
            KeyStoreHelper.assertProviders();
            if (secureRandom == null) {secureRandom = new SecureRandom();}

            PEMWriter pemWriter = null;
            try {                
                pemWriter = new PEMWriter(new OutputStreamWriter(new BufferedOutputStream(destinationOutputStream)), "BC");

                Enumeration<String> aliases = keyStore.aliases();
                while (aliases.hasMoreElements()) {
                    String alias = aliases.nextElement();

                    if (keyStore.entryInstanceOf(alias, PrivateKeyEntry.class)) {
                        try {
                            PrivateKeyEntry privateKeyEntry = ((PrivateKeyEntry)(keyStore.getEntry(alias, new KeyStore.PasswordProtection(passphrase))));
                            X509Certificate[] chain = ((X509Certificate[])(privateKeyEntry.getCertificateChain()));
                            PrivateKey privateKey = privateKeyEntry.getPrivateKey();

                            if (algorithm.equals(KeyStoreHelper.PEM_ALGORITHM_NO_ENCRYPTION)) {
                                pemWriter.writeObject(privateKey);
                            } else {
                                pemWriter.writeObject(privateKey, algorithm, destinationPassphrase, secureRandom);
                            }

                            int i = 1-1; int n = chain.length;
                            while (i <= (n-1)) {
                                pemWriter.writeObject(chain[i]);
                                i = i + 1;
                            }
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, SecretKeyEntry.class)) {
                        try {
                            //for now do nothing
                        } catch (Exception e2) {}

                    } else if (keyStore.entryInstanceOf(alias, TrustedCertificateEntry.class)) {
                        try {
                            TrustedCertificateEntry tce = ((TrustedCertificateEntry)(keyStore.getEntry(alias, null)));
                            pemWriter.writeObject(tce.getTrustedCertificate());
                        } catch (Exception e2) {}

                    }

                }
                pemWriter.flush();
                return keyStore;
            } catch (Exception e) {
                return null;
            }
        }

        public static PrivateKey saveAsKEYFile(KeyPair keyPair, char[] destinationPassphrase, String destinationFileName) {
            return KeyStoreHelper.saveAsKEYFile(PEM_ALGORITHM_AES_256_CFB, keyPair.getPrivate(), destinationPassphrase, destinationFileName, null);
        }

        public static PrivateKey saveAsKEYFile(KeyPair keyPair, char[] destinationPassphrase, File destinationFile) {
            return KeyStoreHelper.saveAsKEYFile(PEM_ALGORITHM_AES_256_CFB, keyPair.getPrivate(), destinationPassphrase, destinationFile, null);
        }

        public static PrivateKey saveAsKEYFile(KeyPair keyPair, char[] destinationPassphrase, OutputStream destinationOutputStream) {
            return KeyStoreHelper.saveAsKEYFile(PEM_ALGORITHM_AES_256_CFB, keyPair.getPrivate(), destinationPassphrase, destinationOutputStream, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, KeyPair keyPair, char[] destinationPassphrase, String destinationFileName) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, keyPair.getPrivate(), destinationPassphrase, destinationFileName, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, KeyPair keyPair, char[] destinationPassphrase, File destinationFile) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, keyPair.getPrivate(), destinationPassphrase, destinationFile, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, KeyPair keyPair, char[] destinationPassphrase, OutputStream destinationOutputStream) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, keyPair.getPrivate(), destinationPassphrase, destinationOutputStream, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, KeyPair keyPair, char[] destinationPassphrase, String destinationFileName, SecureRandom secureRandom) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, keyPair.getPrivate(), destinationPassphrase, destinationFileName, secureRandom);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, KeyPair keyPair, char[] destinationPassphrase, File destinationFile, SecureRandom secureRandom) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, keyPair.getPrivate(), destinationPassphrase, destinationFile, secureRandom);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, KeyPair keyPair, char[] destinationPassphrase, OutputStream destinationOutputStream, SecureRandom secureRandom) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, keyPair.getPrivate(), destinationPassphrase, destinationOutputStream, secureRandom);
        }

        public static PrivateKey saveAsKEYFile(PrivateKey privateKey, char[] destinationPassphrase, String destinationFileName) {
            return KeyStoreHelper.saveAsKEYFile(PEM_ALGORITHM_AES_256_CFB, privateKey, destinationPassphrase, destinationFileName, null);
        }

        public static PrivateKey saveAsKEYFile(PrivateKey privateKey, char[] destinationPassphrase, File destinationFile) {
            return KeyStoreHelper.saveAsKEYFile(PEM_ALGORITHM_AES_256_CFB, privateKey, destinationPassphrase, destinationFile, null);
        }

        public static PrivateKey saveAsKEYFile(PrivateKey privateKey, char[] destinationPassphrase, OutputStream destinationOutputStream) {
            return KeyStoreHelper.saveAsKEYFile(PEM_ALGORITHM_AES_256_CFB, privateKey, destinationPassphrase, destinationOutputStream, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, PrivateKey privateKey, char[] destinationPassphrase, String destinationFileName) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, privateKey, destinationPassphrase, destinationFileName, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, PrivateKey privateKey, char[] destinationPassphrase, File destinationFile) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, privateKey, destinationPassphrase, destinationFile, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, PrivateKey privateKey, char[] destinationPassphrase, OutputStream destinationOutputStream) {
            return KeyStoreHelper.saveAsKEYFile(algorithm, privateKey, destinationPassphrase, destinationOutputStream, null);
        }

        public static PrivateKey saveAsKEYFile(String algorithm, PrivateKey privateKey, char[] destinationPassphrase, String destinationFileName, SecureRandom secureRandom) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsKEYFile(algorithm, privateKey, destinationPassphrase, destinationFile, secureRandom);
            } catch (Exception e) {
                return null;
            }
        }

        public static PrivateKey saveAsKEYFile(String algorithm, PrivateKey privateKey, char[] destinationPassphrase, File destinationFile, SecureRandom secureRandom) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsKEYFile(algorithm, privateKey, destinationPassphrase, fileOutputStream, secureRandom);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }

        }

        public static PrivateKey saveAsKEYFile(String algorithm, PrivateKey privateKey, char[] destinationPassphrase, OutputStream destinationOutputStream, SecureRandom secureRandom) {
            KeyStoreHelper.assertProviders();
            if (secureRandom == null) {secureRandom = new SecureRandom();}

            PEMWriter pemWriter = null;
            try {
                pemWriter = new PEMWriter(new OutputStreamWriter(new BufferedOutputStream(destinationOutputStream)), "BC");
                if (algorithm.equals(KeyStoreHelper.PEM_ALGORITHM_NO_ENCRYPTION)) {
                    pemWriter.writeObject(privateKey);
                } else {
                    pemWriter.writeObject(privateKey, algorithm, destinationPassphrase, secureRandom);
                }                
                pemWriter.flush();
                return privateKey;
            } catch (Exception e) {
                return null;
            }
        }

        public static X509Certificate saveAsCRT(X509Certificate x509, String destinationFileName) {
            File destinationFile = null;
            try {
                destinationFile = new File(destinationFileName);
                return KeyStoreHelper.saveAsCRT(x509, destinationFile);
            } catch (Exception e) {
                return null;
            }
        }

        public static X509Certificate saveAsCRT(X509Certificate x509, File destinationFile) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsCRT(x509, fileOutputStream);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static X509Certificate saveAsCRT(X509Certificate x509, OutputStream destinationOutputStream) {
            KeyStoreHelper.assertProviders();        
            PEMWriter pemWriter = null;
            try {                
                pemWriter = new PEMWriter(new OutputStreamWriter(new BufferedOutputStream(destinationOutputStream)), "BC");
                pemWriter.writeObject(x509);
                pemWriter.flush();                                
                return x509;
            } catch (Exception e) {
                return null;
            }
        }

        public static X509Certificate saveAsCER(X509Certificate x509, String destinationFileName) {
            return KeyStoreHelper.saveAsDER(x509, destinationFileName);
        }

        public static X509Certificate saveAsCER(X509Certificate x509, File destinationFile) {
            return KeyStoreHelper.saveAsDER(x509, destinationFile);
        }

        public static X509Certificate saveAsCER(X509Certificate x509, OutputStream destinationOutputStream) {
            return KeyStoreHelper.saveAsDER(x509, destinationOutputStream);
        }

        public static X509Certificate saveAsDER(X509Certificate x509, String destinationFileName) {
            File file = null;
            try {
                file = new File(destinationFileName);
                return KeyStoreHelper.saveAsDER(x509, file);
            } catch (Exception e) {
                return null;
            }
        }

        public static X509Certificate saveAsDER(X509Certificate x509, File destinationFile) {
            FileOutputStream fileOutputStream = null;
            try {
                try {
                    fileOutputStream = new FileOutputStream(destinationFile);
                    return KeyStoreHelper.saveAsDER(x509, fileOutputStream);
                } finally {
                    try {fileOutputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {
                return null;
            }
        }

        public static X509Certificate saveAsDER(X509Certificate x509, OutputStream destinationOutputStream) {
            KeyStoreHelper.assertProviders();           
            try {
                byte[] buffer = x509.getEncoded();
                destinationOutputStream.write(buffer);
                destinationOutputStream.flush();
                return x509;
            } catch (Exception e) {
                return null;
            }
        } 

    /*
     * ***
     * *** ALIASES
     * ***
     */

        public static String getAlias(X509Certificate x509) {
            String alias = null;
            try {
                alias = (new DNProcessor(x509.getSubjectDN().getName())).getCommonName() + "_" + x509.getNotBefore() + "_" + (new DNProcessor(x509.getIssuerDN().getName())).getCommonName() + "_" + Utils.toHex(x509.getSerialNumber().toByteArray());
                return alias;
            } catch (Exception e) {
                return null;
            }
        }

    /*
     * ***
     * *** DNProcessor TOOLS
     * ***
     */

        public static DNProcessor processDN(String aDN) {
            return new DNProcessor(aDN);
        }

        public static class DNProcessor {
            private static final String COMMONNAME = "CN";
            private static final String ORGANIZATIONALUNIT = "OU";
            private static final String ORGANIZATION = "O";
            private static final String LOCALITY = "L";            
            private static final String STATEORPROVINCE = "ST";
            private static final String COUNTRY = "C";
            private static final String UID = "UID";

            private static final String NOTAVAILABLE = "n/a";
            private static final String EMPTY = "";

            public DNProcessor() {}

            public DNProcessor(X509Certificate x509) {
                this();
                try {
                    this.processDNString(x509.getSubjectX500Principal().getName(X500Principal.RFC2253));
                } catch (Exception e) {
                    System.out.println(e);
                }
                this.processCertificateAlternativeNameExtension(x509);
            }

            public DNProcessor(PKCS10CertificationRequest request) {
                this();
                this.processDNString(request.getCertificationRequestInfo().getSubject().toString());
                this.processCertificationRequestAlternativeNameExtension(request);
            }

            public DNProcessor(X500Principal x500Principal) {
                this();
                try {
                    this.processDNString(x500Principal.getName(X500Principal.RFC2253));
                } catch (Exception e) {}
            }

            public DNProcessor(String aDN) {
                this();
                this.processDNString(aDN);
            }

            public void processDNString(String aDN) {
                while (aDN.contains("=")) {
                    int pos = aDN.indexOf("=", 0);
                    String identifier = aDN.substring(0, pos);
                                   aDN = aDN.substring(pos + 1);
                         String value = aDN.substring(0);

                    int end = aDN.indexOf(",", 0);
                    if (end != -1) {
                        value = aDN.substring(0, end);
                           aDN = aDN.substring(end + 1);
                    }
                    value = value.trim();

                    identifier = identifier.trim();
                    identifier = identifier.toUpperCase();
                           if (identifier.equals(COMMONNAME)) {
                        this.fCommonName = value;

                    } else if (identifier.equals(ORGANIZATIONALUNIT)) {
                        this.fOrganizationalUnit = value;

                    } else if (identifier.equals(ORGANIZATION)) {
                        this.fOrganization = value;

                    } else if (identifier.equals(LOCALITY)) {
                        this.fLocality = value;

                    } else if (identifier.equals(STATEORPROVINCE)) {
                        this.fStateOrProvince = value;

                    } else if (identifier.equals(COUNTRY)) {
                        this.fCountry = value;

                    } else if (identifier.equals(UID)) {
                        this.fUID = value;

                    }
                }
            }

            public GeneralNames getAlternativeNamesAsGeneralNames() {
                GeneralNames generalNames = null;
                try {
                    if (   (this.getDNSName() != null)
                        || (this.getEmailAddress() != null)
                        || (this.getIPAddress() != null)
                        || (this.getURI() != null)
                       ) {
                        GeneralName dnsName = null;
                        GeneralName emailAddress = null;
                        GeneralName ipAddress = null;
                        GeneralName uri = null;

                        if (this.getDNSName() != null) {
                            try {
                                dnsName = new GeneralName(GeneralName.dNSName, this.getDNSName());
                            } catch (Exception e2) {}
                        }

                        if (this.getEmailAddress() != null) {
                            try {
                                emailAddress = new GeneralName(GeneralName.rfc822Name, this.getEmailAddress());
                            } catch (Exception e2) {}
                        }

                        if (this.getIPAddress() != null) {
                            try {
                                ipAddress = new GeneralName(GeneralName.iPAddress, this.getIPAddress());
                            } catch (Exception e2) {}
                        }

                        if (this.getURI() != null) {
                            try {
                                uri = new GeneralName(GeneralName.uniformResourceIdentifier, this.getURI());
                            } catch (Exception e2) {}
                        }

                        DERObject[] derObjects = null;
                        int count = 0;
                        if (dnsName != null) {count = count + 1;}
                        if (emailAddress != null) {count = count + 1;}
                        if (ipAddress != null) {count = count + 1;}
                        if (uri != null) {count = count + 1;}

                        if (count > 0) {
                            derObjects = new DERObject[count];
                            count = 0;
                            if (dnsName != null) {derObjects[count] = dnsName.getDERObject(); count = count + 1;}
                            if (emailAddress != null) {derObjects[count] = emailAddress.getDERObject(); count = count + 1;}
                            if (ipAddress != null) {derObjects[count] = ipAddress.getDERObject(); count = count + 1;}
                            if (uri != null) {derObjects[count] = uri.getDERObject(); count = count + 1;}

                            DERSequence derSequence = new DERSequence(derObjects);
                            generalNames = new GeneralNames(derSequence);
                        }
                    }
                } catch (Exception e) {}
                return generalNames;
            }

            public DERSet getCertificationRequestAlternativeNameExtension() {
                DERSet derSet = null;
                try {
                    if (   (this.getDNSName() != null)
                        || (this.getEmailAddress() != null)
                        || (this.getIPAddress() != null)
                        || (this.getURI() != null)
                       ) {
                        GeneralName dnsName = null;
                        GeneralName emailAddress = null;
                        GeneralName ipAddress = null;
                        GeneralName uri = null;

                        if (this.getDNSName() != null) {
                            try {
                                dnsName = new GeneralName(GeneralName.dNSName, this.getDNSName());
                            } catch (Exception e) {}
                        }

                        if (this.getEmailAddress() != null) {
                            try {
                                emailAddress = new GeneralName(GeneralName.rfc822Name, this.getEmailAddress());
                            } catch (Exception e) {}
                        }

                        if (this.getIPAddress() != null) {
                            try {
                                ipAddress = new GeneralName(GeneralName.iPAddress, this.getIPAddress());
                            } catch (Exception e) {}
                        }

                        if (this.getURI() != null) {
                            try {
                                uri = new GeneralName(GeneralName.uniformResourceIdentifier, this.getURI());
                            } catch (Exception e) {}
                        }

                        DERObject[] derObjects = null;
                        int count = 0;
                        if (dnsName != null) {count = count + 1;}
                        if (emailAddress != null) {count = count + 1;}
                        if (ipAddress != null) {count = count + 1;}
                        if (uri != null) {count = count + 1;}

                        if (count > 0) {
                            derObjects = new DERObject[count];
                            count = 0;
                            if (dnsName != null) {derObjects[count] = dnsName.getDERObject(); count = count + 1;}
                            if (emailAddress != null) {derObjects[count] = emailAddress.getDERObject(); count = count + 1;}
                            if (ipAddress != null) {derObjects[count] = ipAddress.getDERObject(); count = count + 1;}
                            if (uri != null) {derObjects[count] = uri.getDERObject(); count = count + 1;}

                            DERSequence derSequence = new DERSequence(derObjects);
                            GeneralNames generalNames = new GeneralNames(derSequence);
                            Vector<DERObjectIdentifier> oids = new Vector<DERObjectIdentifier>(); oids.add(X509Extensions.SubjectAlternativeName);
                            Vector<X509Extension> values = new Vector<X509Extension>(); values.add(new X509Extension(false, new DEROctetString(generalNames)));
                            X509Extensions extensions = new X509Extensions(oids, values);
                            Attribute attribute = new Attribute(PKCSObjectIdentifiers.pkcs_9_at_extensionRequest, new DERSet(extensions));
                            derSet = new DERSet(attribute);
                        }
                    }
                } catch (Exception e) {}
                return derSet;
            }

           public void processCertificateAlternativeNameExtension(X509Certificate x509) {
               try {
                   byte[] extnValue = null;
                   try {
                       extnValue = (DEROctetString.getInstance(ASN1Object.fromByteArray(x509.getExtensionValue("2.5.29.17")))).getOctets();
                   } catch (Exception e2) {}
                   if (extnValue == null) {return;}

                   Enumeration e = null;
                   try {
                       e = DERSequence.getInstance(ASN1Object.fromByteArray(extnValue)).getObjects();
                   } catch (Exception e2) {}
                   if (e == null) {return;}
                   while (e.hasMoreElements()) {
                       GeneralName generalName = GeneralName.getInstance(e.nextElement());
                       if (generalName.getTagNo() == 1) {
                           try {
                               this.fRFC822Name = ((DERString)generalName.getName()).getString();
                           } catch (Exception e2) {}

                       } else if (generalName.getTagNo() == 2) {
                           try {
                               this.fDNSName = ((DERString)generalName.getName()).getString();
                           } catch (Exception e2) {}

                       } else if (generalName.getTagNo() == 6) {
                           try {
                               this.fURI = ((DERString)generalName.getName()).getString();
                           } catch (Exception e2) {}

                       } else if (generalName.getTagNo() == 7) {
                           try {
                               byte[] octets = DEROctetString.getInstance(generalName.getName()).getOctets();
                               String humanReadableIPAddress = null;
                               int k = 1-1; int o = octets.length;
                               int byteValue;
                               while (k <= (o-1)) {
                                   if (octets[k] >= 0) {
                                       byteValue = octets[k];
                                   } else {
                                       byteValue = 256 + octets[k];
                                   }

                                   if (humanReadableIPAddress == null) {
                                       humanReadableIPAddress = "" + byteValue;
                                   } else if (humanReadableIPAddress != null) {
                                       if (o == 4) {
                                           humanReadableIPAddress = humanReadableIPAddress + "." + byteValue;
                                       } else {
                                           humanReadableIPAddress = humanReadableIPAddress + ":" + byteValue;
                                       }
                                   }
                                   k = k + 1;
                               }
                               this.fIPAddress = humanReadableIPAddress;
                           } catch (Exception e2) {}

                       }
                   }
               } catch (Exception e) {}
            }

            public void processCertificationRequestAlternativeNameExtension(PKCS10CertificationRequest request) {
                try {
                    ASN1Set attributes = request.getCertificationRequestInfo().getAttributes();
                    Enumeration e = attributes.getObjects();
                    while (e.hasMoreElements()) {
                        Attribute attribute = Attribute.getInstance(e.nextElement());
                        if (attribute.getAttrType().equals(PKCSObjectIdentifiers.pkcs_9_at_extensionRequest)) {
                            X509Extensions extensions = X509Extensions.getInstance(attribute.getAttrValues().getObjectAt(0));
                            Enumeration i = extensions.oids();
                            while (i.hasMoreElements()) {
                                DERObjectIdentifier oid = (DERObjectIdentifier)i.nextElement();
                                X509Extension extension = extensions.getExtension(oid);
                                ASN1Object asn1Object = X509Extension.convertValueToObject(extension);
                                GeneralNames generalNames = GeneralNames.getInstance(asn1Object);
                                GeneralName[] generalNamesArray = generalNames.getNames();
                                int j = 1-1; int m = generalNamesArray.length;
                                while (j <= (m-1)) {
                                    if (generalNamesArray[j].getTagNo() == GeneralName.dNSName) {
                                        try {
                                            DEREncodable der = generalNamesArray[j].getName();
                                            this.fDNSName = der.toString();
                                        } catch (Exception e2) {}

                                    } else if (generalNamesArray[j].getTagNo() == GeneralName.rfc822Name) {
                                        try {
                                            DEREncodable der = generalNamesArray[j].getName();
                                            this.fRFC822Name = der.toString();
                                        } catch (Exception e2) {}

                                    } else if (generalNamesArray[j].getTagNo() == GeneralName.uniformResourceIdentifier) {
                                        try {
                                            DEREncodable der = generalNamesArray[j].getName();
                                            this.fURI = der.toString();
                                        } catch (Exception e2) {}

                                    } else if (generalNamesArray[j].getTagNo() == GeneralName.iPAddress) {
                                        try {
                                            DEREncodable der = generalNamesArray[j].getName();
                                            ASN1OctetString asn1OctetString = ASN1OctetString.getInstance(der);
                                            byte[] octets = asn1OctetString.getOctets();
                                            String humanReadableIPAddress = null;
                                            int k = 1-1; int o = octets.length;
                                            int byteValue;
                                            while (k <= (o-1)) {
                                                if (octets[k] >= 0) {
                                                    byteValue = octets[k];
                                                } else {
                                                    byteValue = 256 + octets[k];
                                                }

                                                if (humanReadableIPAddress == null) {
                                                    humanReadableIPAddress = "" + byteValue;
                                                } else if (humanReadableIPAddress != null) {
                                                    if (o == 4) {
                                                        humanReadableIPAddress = humanReadableIPAddress + "." + byteValue;
                                                    } else {
                                                        humanReadableIPAddress = humanReadableIPAddress + ":" + byteValue;
                                                    }
                                                }
                                                k = k + 1;
                                            }
                                            this.fIPAddress = humanReadableIPAddress;
                                        } catch (Exception e2) {}
                                    }
                                    j = j + 1;
                                }
                            }
                            break;
                        }
                    }
                } catch (Exception e) {
System.out.println("1\\" + e);
                }
            }

            String fCommonName = null;
            public String getCommonName() {
                if (this.fCommonName == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fCommonName;
                }
            }
            public void setCommonName(String commonName) {this.fCommonName = commonName;}

            String fOrganizationalUnit = null;
            public String getOrganizationalUnit() {
                if (this.fOrganizationalUnit == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fOrganizationalUnit;
                }
            }
            public void setOrganizationalUnit(String organizationalUnit) {this.fOrganizationalUnit = organizationalUnit;}

            String fOrganization = null;
            public String getOrganization() {
                if (this.fOrganization == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fOrganization;
                }
            }
            public void setOrganization(String organization) {this.fOrganization = organization;}

            String fLocality = null;
            public String getLocality() {
                if (this.fLocality == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fLocality;
                }
            }
            public void setLocality(String locality) {this.fLocality = locality;}

            String fStateOrProvince = null;
            public String getStateOrProvince() {
                if (this.fStateOrProvince == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fStateOrProvince;
                }
            }
            public void setStateOrProvince(String stateOrProvince) {this.fStateOrProvince = stateOrProvince;}

            String fCountry = null;
            public String getCountry() {
                if (this.fCountry == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fCountry;
                }
            }
            public void setCountry(String country) {this.fCountry = country;}

            String fUID = null;
            public String getUID() {
                if (this.fUID == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fUID;
                }
            }
            public void setUID(String uid) {this.fUID = uid;}

            String fDNSName = null;
            public String getDNSName() {
                if (this.fDNSName == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fDNSName;
                }
            }
            public void setDNSName(String dnsName) {this.fDNSName = dnsName;}

            String fURI = null;
            public String getURI() {
                if (this.fURI == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fURI;
                }
            }
            public void setURI(String uri) {this.fURI = uri;}

            String fRFC822Name = null;
            public String getEmailAddress() {
                if (this.fRFC822Name == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fRFC822Name;
                }
            }
            public void setEmailAddress(String emailAddress) {this.fRFC822Name = emailAddress;}

            String fIPAddress = null;
            public String getIPAddress() {
                if (this.fIPAddress == null) {
                    return NOTAVAILABLE;
                } else {
                    return this.fIPAddress;
                }
            }
            public void setIPAddress(String ipAddress) {this.fIPAddress = ipAddress;}

            public X500Principal getDNStringAsX500Principal() {
                return new X500Principal(this.getDNString());
            }

            public String getDNString() {
                String DNString = null;
                if (((this.fCommonName == null) || (this.fCommonName.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = COMMONNAME + "=" + this.fCommonName;
                    } else {
                        DNString = DNString + ", " + COMMONNAME + "=" + this.fCommonName;
                    }
                }

                if (((this.fOrganizationalUnit == null) || (this.fOrganizationalUnit.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = ORGANIZATIONALUNIT + "=" + this.fOrganizationalUnit;
                    } else {
                        DNString = DNString + ", " + ORGANIZATIONALUNIT + "=" + this.fOrganizationalUnit;
                    }
                }

                if (((this.fOrganization == null) || (this.fOrganization.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = ORGANIZATION + "=" + this.fOrganization;
                    } else {
                        DNString = DNString + ", " + ORGANIZATION + "=" + this.fOrganization;
                    }
                }

                if (((this.fLocality == null) || (this.fLocality.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = LOCALITY + "=" + this.fLocality;
                    } else {
                        DNString = DNString + ", " + LOCALITY + "=" + this.fLocality;
                    }
                }

                if (((this.fStateOrProvince == null) || (this.fStateOrProvince.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = STATEORPROVINCE + "=" + this.fStateOrProvince;
                    } else {
                        DNString = DNString + ", " + STATEORPROVINCE + "=" + this.fStateOrProvince;
                    }
                }

                if (((this.fCountry == null) || (this.fCountry.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = COUNTRY + "=" + this.fCountry;
                    } else {
                        DNString = DNString + ", " + COUNTRY + "=" + this.fCountry;
                    }
                }

                if (((this.fUID == null) || (this.fUID.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = UID + "=" + this.fUID;
                    } else {
                        DNString = DNString + ", " + UID + "=" + this.fUID;
                    }
                }

                if (DNString == null) {DNString = NOTAVAILABLE;}
                return DNString;
            }

            public X500Principal getDNStringReversedOrderAsX500Principal() {
                return new X500Principal(this.getDNStringReversedOrder());
            }

            public String getDNStringReversedOrder() {
                String DNString = null;
                if (((this.fUID == null) || (this.fUID.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = UID + "=" + this.fUID;
                    } else {
                        DNString = DNString + ", " + UID + "=" + this.fUID;
                    }
                }

                if (((this.fCountry == null) || (this.fCountry.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = COUNTRY + "=" + this.fCountry;
                    } else {
                        DNString = DNString + ", " + COUNTRY + "=" + this.fCountry;
                    }
                }

                if (((this.fStateOrProvince == null) || (this.fStateOrProvince.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = STATEORPROVINCE + "=" + this.fStateOrProvince;
                    } else {
                        DNString = DNString + ", " + STATEORPROVINCE + "=" + this.fStateOrProvince;
                    }
                }

                if (((this.fLocality == null) || (this.fLocality.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = LOCALITY + "=" + this.fLocality;
                    } else {
                        DNString = DNString + ", " + LOCALITY + "=" + this.fLocality;
                    }
                }

                if (((this.fOrganization == null) || (this.fOrganization.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = ORGANIZATION + "=" + this.fOrganization;
                    } else {
                        DNString = DNString + ", " + ORGANIZATION + "=" + this.fOrganization;
                    }
                }

                if (((this.fOrganizationalUnit == null) || (this.fOrganizationalUnit.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = ORGANIZATIONALUNIT + "=" + this.fOrganizationalUnit;
                    } else {
                        DNString = DNString + ", " + ORGANIZATIONALUNIT + "=" + this.fOrganizationalUnit;
                    }
                }

                if (((this.fCommonName == null) || (this.fCommonName.equals(EMPTY))) == false) {
                    if (DNString == null) {
                        DNString = COMMONNAME + "=" + this.fCommonName;
                    } else {
                        DNString = DNString + ", " + COMMONNAME + "=" + this.fCommonName;
                    }
                }

                if (DNString == null) {DNString = NOTAVAILABLE;}
                return DNString;
            }
        }


}