/* ============================================================================
 * TTLSHelper.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2007-2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 * 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or 
 *      its contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS 
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 * 
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON 
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES 
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE 
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE 
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON 
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 * 
 * ============================================================================
 * 
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 * 
 * ============================================================================
 */
package nl.jhhcvandermeijs.tcpconns.tls;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;

/**
 * provides some utilities for authentication with TTLSProvider
 */
public class TTLSHelper {

    /**
     * Use this class method to load a pkcs12 file as keypair (private and 
     * public key) which is used to authenticate this side of the SSL connection
     * by the other side
     * 
     * @param aPKCS12FileName pkcs12 filename that holds the keyset
     * @param aPKCS12FilePassphrase chararray that holds the appropriate
     *        passphrase for the pkcs12 file
     * @return array of KeyManager
     * @throws java.io.FileNotFoundException
     * @throws java.security.KeyStoreException
     * @throws java.io.IOException
     * @throws java.security.NoSuchAlgorithmException
     * @throws java.security.cert.CertificateException
     * @throws java.security.UnrecoverableKeyException
     */
    public synchronized static KeyManager[] createKeyManagers(String aPKCS12FileName,
                                                 char[] aPKCS12FilePassphrase)
            throws FileNotFoundException, 
                   KeyStoreException, 
                   IOException, 
                   NoSuchAlgorithmException, 
                   CertificateException, 
                   UnrecoverableKeyException {
        
        if (    (aPKCS12FileName == null)
             || (aPKCS12FilePassphrase == null)) {throw new UnrecoverableKeyException();}
        
        File aPKCS12File = new File(aPKCS12FileName);
        if (!(aPKCS12File.exists())) {throw new FileNotFoundException();}
        if (!(aPKCS12File.isFile())) {throw new IOException();}
                
        InputStream inStream = null;
        KeyStore aKeyStore = null;
        KeyManagerFactory aKeyManagerFactory = null;        
            try {
                inStream = new FileInputStream(aPKCS12File);
                aKeyStore = KeyStore.getInstance("PKCS12");
                aKeyStore.load(inStream, aPKCS12FilePassphrase);
                
                aKeyManagerFactory = KeyManagerFactory.getInstance("SunX509");
                aKeyManagerFactory.init(aKeyStore, aPKCS12FilePassphrase);
                return aKeyManagerFactory.getKeyManagers();
            } finally {
                try {inStream.close();} catch (Exception e2) {}
            }        
    }    
        
    /**
     * use this method to load a .cer certificate of e.g. a trusted CA
     * @param anX509CertificateFileName complete path and filename towards a X509 public
     *        certificate
     * @return an X509Certificate object 
     * @throws java.security.cert.CertificateException
     * @throws java.io.FileNotFoundException
     */
    public static X509Certificate loadX509Certificate(String anX509CertificateFileName) throws CertificateException, FileNotFoundException {
        File anX509File = null;
        InputStream inStream = null;
        X509Certificate anX509 = null;
        CertificateFactory aCertificateFactory = CertificateFactory.getInstance("X.509");
        try {
            anX509File = new File(anX509CertificateFileName);
            inStream = new FileInputStream(anX509File);
            anX509 = ((X509Certificate)(aCertificateFactory.generateCertificate(inStream)));
        } finally {
            try {inStream.close();} catch (Exception e) {}
        }
        return anX509;
    }
    
    /**
     * this method returns true if the anX509 is signed by aTrustedCA and the
     * anX509 is within the validity period determined by aTrustedCA
     * 
     * @param anX509
     * @param aTrustedCA
     * @return true if not invalid (i.e. you must still check certificate
     * revocation status!!!)
     */
    public static boolean isValid(X509Certificate anX509, X509Certificate aTrustedCA) {
        try {
            anX509.checkValidity();
            anX509.verify(aTrustedCA.getPublicKey());
            return true;
        } catch (Exception e) {
            return false;
        }                
    }
    
}
