/* ============================================================================
 * SRNGSeeder.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2008-2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 * 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or 
 *      its contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS 
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 * 
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON 
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES 
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE 
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE 
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON 
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 * 
 * ============================================================================
 * 
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 * 
 * ============================================================================
 */
package nl.jhhcvandermeijs.tcpconns.utils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.net.URL;
import java.security.MessageDigest;
import java.security.SecureRandom;
import java.security.Security;
import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import org.bouncycastle.jce.provider.BouncyCastleProvider;

/**
 * Class to establish more or less an adequate propagation of random uniqueness.
 * Obtain an initial true random state from e.g. 
 * https://www.fourmilab.ch/hotbits/secure_generate.html. And addRandom from
 * http://random.hd.org/getBits.jsp. Then regularly addRandom from both
 * hotbits and random.hd.org. The more sources, the more secure. This class now
 * includes a method to obtain random numbers from hotbits automatically, however,
 * for this to work, an active internet connection is required. The seed file is
 * stored in the user directory (/home/username/tcpconnsSRNGSeeder.rnd)
 * 
 */
public class SRNGSeeder {
    /**
     * this secure random is automatically refreshed when one calls
     * newSecureRandom(); its purpose is to provide initialization vectors
     * (discarded) and random keys (discarded) to shuffle the bytes in the
     * random pool
     */
    protected SecureRandom fSecureRandom = new SecureRandom();

    /**the byte array that holds the random seed state*/
    protected byte[] fSeed = new byte[4096];
    protected String fSeedFileName = null;

    public SRNGSeeder() {
        super();
        this.initialize(null);
    }

    public SRNGSeeder(String seedFileName) {
        super();
        this.initialize(seedFileName);
    }

    private void initialize(String seedFileName) {
        try {
        Security.addProvider(new BouncyCastleProvider());
        this.fSecureRandom.nextBytes(this.fSeed);
        if ((seedFileName == null)) {
            String userDirectoryName = Utils.getUserHome() + ".tcpconns" + File.separator + "random";
            File userDirectory = new File(userDirectoryName); if (!(userDirectory.exists())) {userDirectory.mkdirs();}
            this.fSeedFileName = userDirectoryName + File.separator + "tcpconns.rnd";            
        }
        byte[] buffer = new byte[4096];
        this.fSecureRandom.nextBytes(buffer);              

        
        InputStream inputStream = null;
        RandomAccessFile randomAccessFile = null;        
        File seedFile = new File(this.fSeedFileName);
       
        if (seedFile.exists() == false) {
            try {
                try {
                    seedFile.createNewFile();
                    randomAccessFile = new RandomAccessFile(seedFile, "rw");
                    randomAccessFile.setLength(4096);
                    randomAccessFile.seek(0);
                    randomAccessFile.write(buffer);

                    inputStream = (new URL("https://www.fourmilab.ch/cgi-bin/uncgi/Hotbits?nbytes=2048&fmt=bin")).openStream();
                    inputStream.read(buffer, 1024, 2048);

                    this.addRandom(buffer);

                    randomAccessFile.seek(0);
                    randomAccessFile.write(this.fSeed);

                } finally {
                    try {randomAccessFile.close();} catch (Exception e) {}
                }
            } catch (Exception e) {}

        } else {            
            this.fSeedFileName = seedFileName; File userSeedFile = new File(this.fSeedFileName);
            try {
                try {
                    inputStream = new FileInputStream(userSeedFile);
                    inputStream.read(buffer, 1024, 2048);

                    this.addRandom(buffer);

                    randomAccessFile = new RandomAccessFile(seedFile, "rw");
                    randomAccessFile.setLength(4096);
                    randomAccessFile.seek(0);
                    randomAccessFile.write(this.fSeed);
                } finally {
                    try {inputStream.close();} catch (Exception e2) {}
                }
            } catch (Exception e) {}
        }
        this.addRandom(); //keep different internal state

        } catch (Exception e) {}
    }

    /**shuffles all bytes in the random pool, and stores the in-between-result to disk*/
    protected void addRandom() {
        //try and collect some "random" data
        this.fSecureRandom = new SecureRandom();
        byte[] buffer = new byte[4096];
        this.fSecureRandom.nextBytes(buffer);

        InputStream inputStream = null;
        try {
            try {
                try {
                    File unixRandomFile = new File("/dev/random");
                    if ((unixRandomFile != null) && (unixRandomFile.exists())) {
                        inputStream = new FileInputStream(unixRandomFile);
                        inputStream.read(buffer, 3, 16); //=> finally
                    }
                } catch (Exception e) {}

                MessageDigest messageDigest = MessageDigest.getInstance("SHA-512", "BC");
                messageDigest.update(this.fSeed);
                byte[] digest = messageDigest.digest();
                System.arraycopy(digest, 0, buffer, 34, digest.length);

                messageDigest = MessageDigest.getInstance("Whirlpool", "BC");
                messageDigest.update(buffer);
                digest = messageDigest.digest();
                System.arraycopy(digest, 0, this.fSeed, 128, digest.length);

            } finally {
                try {inputStream.close();} catch (Exception e2) {}
            }
        } catch (Exception e) {}

        this.addRandom(buffer);
    }
        
    /**
     * adds new random data to the pool; please do this regularly with as much
     * true random data as you can find
     * @param randomData
     * @throws java.io.IOException
     */
    synchronized public void addRandom(byte[] randomData) {
        Security.addProvider(new BouncyCastleProvider());                
        try {
            //cipher
            Cipher AESCipher = null;
            
            //clear text
            byte[] clearText = new byte[this.fSeed.length];
            System.arraycopy(this.fSeed, 0, clearText, 0, this.fSeed.length);

            //cipher text
            byte[] cipherText = null;
        
            //key
            MessageDigest messageDigest = MessageDigest.getInstance("SHA-256", "BC");
            messageDigest.update(randomData);
            byte[] rawkey = messageDigest.digest();
            SecretKeySpec key = new SecretKeySpec(rawkey, "AES");
                    
            //initialization vector
            byte[] anIV = new byte[((int)(128/8))];
            this.fSecureRandom.nextBytes(anIV);
            IvParameterSpec anIvParameterSpec = new IvParameterSpec(anIV);            
                                 
            //prepare cipher
            AESCipher = Cipher.getInstance("AES/CFB/NoPadding", "BC");                                
            AESCipher.init(Cipher.ENCRYPT_MODE, key, anIvParameterSpec);
            
            //prepare target
            ByteArrayOutputStream bOut = new ByteArrayOutputStream();
            //discard iv   bOut.write(anIV, 0, anIV.length);
            
            //prepare source
            ByteArrayInputStream  bIn = new ByteArrayInputStream(clearText);
            CipherInputStream cIn = new CipherInputStream(bIn, AESCipher);

            //IO
            int ch;
            while ((ch = cIn.read()) >= 0) {
                bOut.write(ch);
            }
            cIn.close();
            
            //get result            
            cipherText = bOut.toByteArray();            
            
            //store interresult
                        
            //clear text 2
            clearText = cipherText;
        
            //cipher text 2
            cipherText = null;
        
            //key 2            
            this.fSecureRandom.nextBytes(rawkey);
            key = new SecretKeySpec(rawkey, "AES");
                    
            //initialization vector 2
            this.fSecureRandom.nextBytes(anIV);
            anIvParameterSpec = new IvParameterSpec(anIV);            
                                 
            //prepare cipher            
            AESCipher.init(Cipher.ENCRYPT_MODE, key, anIvParameterSpec);
            
            //prepare target
            bOut = new ByteArrayOutputStream();
            //discard iv   bOut.write(anIV, 0, anIV.length);
            
            //prepare source
            bIn = new ByteArrayInputStream(clearText);
            cIn = new CipherInputStream(bIn, AESCipher);

            //IO
            while ((ch = cIn.read()) >= 0) {
                bOut.write(ch);
            }
            cIn.close();
            
            //get result            
            cipherText = bOut.toByteArray();                                               

            //"shuffle" complete
            System.arraycopy(cipherText, 0, this.fSeed, 0, this.fSeed.length);            
        } catch (Exception e) {}                                        
    }
    
    /**
     * use this method to obtain a new secure random
     * @return a SecureRandom
     */
    synchronized public SecureRandom newSecureRandom() {
        this.addRandom();
        return new SecureRandom(this.fSeed);
    }
    
    /**
     * shortcut for obtaining a random byte sequence
     */
    synchronized public void nextBytes(byte[] buffer) throws IOException {
        (this.newSecureRandom()).nextBytes(buffer);
    }
    
}
