/* ============================================================================
 * XMLFileStore.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2007-2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or
 *      its contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 *
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 *
 * ============================================================================
 *
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 *
 * ============================================================================
 */
package nl.jhhcvandermeijs.tcpconns.utils;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.RandomAccessFile;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 * use this class to create a simple XML settings/configuration file
 */
public class XMLFileStore {

    /**
     * constructor
     */
    public XMLFileStore() {
        super();
    }
    /**
     * constructor (e.g. when creating a document instead of loading one)
     * @param rootTag the tag of the root element
     */
    public XMLFileStore(String rootTag) {
        super();
        this.getRootElement().setTag(rootTag);
    }
    /**
     * constructor (e.g. when creating a document instead of loading one)
     * @param rootTag the tag of the root element
     * @param comment a comment line e.g. to indicate the purpose of the .xml file
     */
    public XMLFileStore(String rootTag, String comment) {
        super();
        this.getRootElement().setTag(rootTag);
        this.getRootElement().setComment(comment);
    }

    /**
     * the Element class is used to store element data
     */
    public static class Element {
        /**
         * the Attribute class is used to store attribute data
         */
        public static class Attribute {
            /**
             * field to store attribute name
             */
            private String fDescription = "";
            /**
             * method to obtain fDescription
             * @return fDescription
             */
            public String getDescription() {return this.fDescription;}
            /**
             * method to set fDescription
             * @param description fDescription
             */
            public void setDescription(String description) {this.fDescription = description;}
            /**
             * field to store attribute value
             */
            private String fValue = "";
            /**
             * method to obtain fValue
             * @return fValue
             */
            public String getValue() {return this.fValue;}
            /**
             * method to set fValue
             * @param value fValue
             */
            public void setValue(String value) {this.fValue = value;}
        }

        /**
         * field to indicate indent of <TAG></TAG>, i.e. the root element
         * has an fLevel of 0, the immediate child elements have an fLevel
         * of 1, and the child elements of the immediate child elements to the
         * root element have an fLevel of 2, and so on....
         */
        private int fLevel = 0;
        /**
         * method to obtain fLevel
         * @return fLevel
         */
        public int getLevel() {return this.fLevel;}
        /**
         * method to set fLevel
         * @param level fLevel
         */
        public void setLevel(int level) {this.fLevel = level;}

        /**
         * handle to the parent of this element; the root element is the
         * only element without a parent
         */
        private Element fParent = null;
        /**
         * method to obtain fParent
         * @return fParent
         */
        public Element getParent() {return this.fParent;}
        /**
         * method to set fParent
         * @param parent fParent
         */
        public void setParent(Element parent) {this.fParent = parent;}

        /**
         * field to store a comment that immediately precedes an element
         */
        private String fComment = null;
        /**
         * method to obtain fComment
         * @return fComment
         */
        public String getComment() {return this.fComment;}
        /**
         * method to set fComment
         * @param comment fComment
         */
        public void setComment(String comment) {this.fComment = comment;}
        /**
         * method to indicate whether this element owns a fComment
         * @return true if this element has a comment
         */
        public boolean hasComment() {
            if (this.fComment == null) {return false;}
            if (this.fComment.equals("")) {return false;}
            return true;
        }

        /**
         * field to store the TAG name
         */
        private String fTag = null;
        /**
         * method to obtain fTag
         * @return fTag
         */
        public String getTag() {return this.fTag;}
        /**
         * method to set fTag
         * @param tag fTag
         */
        public void setTag(String tag) {this.fTag = tag;}

        /**
         * field to store a value, NOTE(!): if this element has child elements
         * the value field is IGNORED!!
         */
        private String fValue = null;
        /**
         * method to obtain fValue
         * @return fValue
         */
        public String getValue() {return this.fValue;}
        /**
         * method to obtain fValue
         * @return fValue
         */
        public String getValueAsString() {
            return this.fValue;
        }
        /**
         * method to obtain fValue as short (fValue must contain a short value)
         * @return fValue converted into short data type
         */
        public short getValueAsShortInteger() {
            return Short.parseShort(this.fValue);
        }

        /**
         * method to obtain fValue as integer (fValue must contain an integer value)
         * @return fValue converted into integer data type
         */
        public int getValueAsInteger() {
            return Integer.parseInt(this.fValue);
        }
        /**
         * method to obtain fValue as long (fValue must contain a long value)
         * @return fValue converted into a long data type
         */
        public long getValueAsLongInteger() {
            return Long.parseLong(this.fValue);
        }
        /**
         * method to obtain fValue as float (fValue must contain a float value)
         * @return fValue converted into a float data type
         */
        public float getValueAsFloat() {
            return Float.parseFloat(this.fValue);
        }
        /**
         * method to obtain fValue as float (fValue must contain a float value)
         * @return fValue converted into a float data type
         */
        public double getValueAsDouble() {
            return Double.parseDouble(this.fValue);
        }

        /**
         * method to obtain fValue as boolean (fValue must be TRUE or FALSE, in
         * lower and/or uppercase or mixed lower and uppercase)
         * @return true if fValue == "TRUE", false if fValue != "TRUE"
         */
        public boolean getValueAsBoolean() {
            if (this.fValue.toUpperCase().startsWith(DATA_TYPE_BOOLEAN_VALUE_TRUE)) {
                return true;
            } else {
                return false;
            }
        }
        /**
         * method to obtain fValue as byte (fValue must contain a byte value)
         * @return fValue converted into byte data type
         */
        public byte getValueAsByte() {
            int value = this.getValueAsInteger();
            if (value >= 128) {
                value = value - 256;
            }
            return ((byte)(value));
        }
        /**
         * method to obtain fValue if it was encoded into base 64 representation
         * @return decoded fValue
         */
        public byte[] getDecodedBase64ValueAsByteArray() {
            try {
                byte[] encoded = this.fValue.getBytes("UTF-8");
                byte[] decoded = Base64.decode(encoded);
                return decoded;
            } catch (Exception e) {} //UTF-8 will pose no problems (I reckon....)
            return null;
        }
        /**
         * method to obtain fValue if it was encoded into base 64 representation
         * @return decoded fValue
         */
        public String getDecodedBase64ValueAsString() {
            try {
                byte[] decoded = this.getDecodedBase64ValueAsByteArray();
                String value = new String(decoded, "UTF-8");
                return value;
            } catch (Exception e) {} //UTF-8 will pose no problems (I reckon....)
            return null;
        }
        /**
         * method to obtain fValue if it was encoded into hex representation
         * @return decoded fValue
         */
        public byte[] getDecodedHexValue() {
            return Utils.decodeHex(this.fValue);
        }

        public static final String DATA_TYPE = "DATA_TYPE";
        public static final String DATA_TYPE_STRING = "STRING";
        public static final String DATA_TYPE_SHORTINTEGER = "SHORT [-32768 .. +32767]";
        public static final String DATA_TYPE_INTEGER = "INTEGER [-2147483648 .. +2147483647]";
        public static final String DATA_TYPE_LONGINTEGER = "LONG [-9223372036854775808 .. +9223372036854775807]";
        public static final String DATA_TYPE_BYTE = "BYTE [0..255]";
        public static final String DATA_TYPE_BOOLEAN = "BOOLEAN [TRUE || FALSE]";
        public static final String DATA_TYPE_BOOLEAN_VALUE_TRUE = "TRUE";
        public static final String DATA_TYPE_BOOLEAN_VALUE_FALSE = "FALSE";
        public static final String DATA_TYPE_BASE64 = "BASE64";
        public static final String DATA_TYPE_HEXADECIMAL = "HEXADECIMAL";
        public static final String DATA_TYPE_FLOAT = "FLOAT [32 BITS IEEE 754]";
        public static final String DATA_TYPE_DOUBLE = "DOUBLE [64 BITS IEEE 754]";
        
        /**
         * method to set fValue
         * @param value fValue
         */
        public void setValue(String value) {
            this.fValue = value;
            if (this.getAttribute(DATA_TYPE) != null) {
                Element.Attribute attribute = this.getAttribute(DATA_TYPE);
                this.getAttributes().remove(attribute);
            }
        }
        /**
         * method to set fValue
         * @param value fValue
         */
        public void setValueAsString(String value) {
            this.fValue = value;
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_STRING);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_STRING;
            }
        }
        /**
         * method set fValue as short
         * @param aShortInteger a short
         */
        public void setValueAsShortInteger(short aShortInteger) {
            this.fValue = "" + aShortInteger;
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_SHORTINTEGER);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_SHORTINTEGER;
            }
        }
        /**
         * method set fValue as integer
         * @param anInteger an integer
         */
        public void setValueAsInteger(int anInteger) {
            this.fValue = "" + anInteger;
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_INTEGER);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_INTEGER;
            }
        }
        /**
         * method to set fValue as long
         * @param aLong a long
         */
        public void setValueAsLongInteger(long aLong) {
            this.fValue = "" + aLong;
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_LONGINTEGER);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_LONGINTEGER;
            }
        }
        /**
         * method to set fValue as float
         * @param aFloat a float
         */
        public void setValueAsFloat(float aFloat) {
            this.fValue = Float.toString(aFloat);
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_FLOAT);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_FLOAT;
            }
        }
        /**
         * method to set fValue as double
         * @param aDouble a double
         */
        public void setValueAsDouble(double aDouble) {
            this.fValue = Double.toString(aDouble);
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_FLOAT);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_FLOAT;
            }
        }
        /**
         * method to set fValue as boolean
         * @param aBoolean a boolean
         */
        public void setValueAsBoolean(boolean aBoolean) {
            if (aBoolean == true) {
                this.fValue = DATA_TYPE_BOOLEAN_VALUE_TRUE;
            } else {
                this.fValue = DATA_TYPE_BOOLEAN_VALUE_FALSE;
            }
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_BOOLEAN);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_BOOLEAN;
            }
        }
        /**
         * method to set fValue as byte
         * @param aByte a byte
         */
        public void setValueAsByte(byte aByte) {
            int value = ((int)(aByte));
            if (aByte < 0) {
                value = 256 + value;
            } 
            this.fValue = "" + value;
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_BYTE);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_BYTE;
            }
        }
        /**
         * method to set fValue as base64 encoded string
         * @param value a string
         */
        public void setValueAsBase64(String value) {
            try {
                byte[] buffer = value.getBytes("UTF-8");
                this.setValueAsBase64(buffer);
            } catch (Exception e) {} //UTF-8 will pose no problems (I reckon....)
        }
        /**
         * method to set fValue as base64 encoded string
         * @param value a byte array
         */
        public void setValueAsBase64(byte[] value) {
            try {
                byte[] encoded = Base64.encodeToByte(value, false);
                this.fValue = new String(encoded, "UTF-8");
                if (this.getAttribute(DATA_TYPE) == null) {
                    this.addAttribute(DATA_TYPE, DATA_TYPE_BASE64);
                } else {
                    this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_BASE64;
                }
            } catch (Exception e) {} //UTF-8 will pose no problems (I reckon....)
        }
        /**
         * method to set fValue as hex encoded string
         * @param value a byte array
         */
        public void setValueAsHex(byte[] value) {
            String encoded = Utils.toHex(value);
            this.fValue = encoded;
            if (this.getAttribute(DATA_TYPE) == null) {
                this.addAttribute(DATA_TYPE, DATA_TYPE_HEXADECIMAL);
            } else {
                this.getAttribute(DATA_TYPE).fValue = DATA_TYPE_HEXADECIMAL;
            }
        }
        /**
         * method to ascertain if this element has a value or not
         * @return true if this element has a value
         */
        public boolean hasValue() {
            if (this.fValue == null) {return false;}
            if (this.fValue.equals("")) {return false;}
            return true;
        }

        /**
         * list to store attributes associated with this element
         */
        private List<Attribute> fAttributes = new LinkedList<Attribute>();
        /**
         * method to obtain fAttributes
         * @return fAttributes
         */
        public List<Attribute> getAttributes() {return this.fAttributes;}
        /**
         * method to ascertain whether or not this element has attributes
         * @return true if this element has attributes
         */
        public boolean hasAttributes() {
            if (this.fAttributes == null) {return false;}
            if (this.fAttributes.size() == 0) {return false;} 
            return true;
        }

        /**
         * list to store child elements
         */
        private List<Element> fChildren = new LinkedList<Element>();
        /**
         * method to obtain fChildren
         * @return fChildren
         */
        public List<Element> getElements() {return this.fChildren;}
        /**
         * method that returns true if the element has children, NOTE(!): the
         * fValue field is ignored if this element does have children (it's either
         * a value OR children, but neither both)
         * @return true if this element has children
         */
        public boolean hasElements() {
            if (this.fChildren == null) {return false;}            
            if (this.fChildren.size() == 0) {return false;}
            return true;
        }

        /**
         * convenience method that creates a new element and adds it to this
         * element as a child element
         * @param tag element name, may be null
         * @param value element value, may be null
         * @param comment comment value, may be null
         * @return the newly created element object
         */
        public Element addElement(String tag, String value, String comment) {
            Element element = new Element();
            element.setTag(tag);
            element.setLevel(this.getLevel()+1);
            element.setValue(value);
            element.setComment(comment);
            this.getElements().add(element);
            return element;
        }
        /**
         * convenience method that creates a new element and adds it to this
         * element as a child element
         * @param tag element name, may be null
         * @param value element value, may be null
         * @return the newly created element object
         */
        public Element addElement(String tag, String value) {
            return this.addElement(tag, value, null);
        }
        /**
         * convenience method that creates a new element and adds it to this
         * element as a child element
         * @param tag element name, may be null
         * @return the newly created element object
         */
        public Element addElement(String tag) {
            return this.addElement(tag, null, null);
        }

        /**
         * convenience method to add a section (=> addElement)
         */
        public Element addSection(String tag, String comment) {
            return this.addElement(tag, null, comment);
        }
        /**
         * convenience method to add a section (=> addElement)
         */
        public Element addSection(String tag) {
            return this.addElement(tag);
        }
        /**
         * convenience method to get a section (=> getElement)
         */
        public Element getSection(String tag) {
            return this.getElement(tag);
        }

        /**
         * convenience method that creates a new attribute and adds it to this
         * element
         * @param attributeName attribute name, may be null
         * @param value attribute value, may be null
         * @return the newly created attribute object
         */
        public Element.Attribute addAttribute(String attributeName, String value) {
            Element.Attribute attribute = new Element.Attribute();
            attribute.setDescription(attributeName);
            attribute.setValue(value);
            this.getAttributes().add(attribute);
            return attribute;
        }

        /**
         * convenience method to find the child element with the given tag name
         * NOTE(!): ALWAYS finds the FIRST element with the given tag name and does not test
         * for more elements with the same name.
         *
         * @param tag the name of the element to be found
         * @return the (first) element object found OR null
         */
        public Element getElement(String tag) {
            Element child;
            int i = 1-1; int n = this.getElements().size();
            while (i <= (n-1)) {
                child = this.getElements().get(i);
                if (child.getTag().equals(tag)) {
                    return child;
                }
                i++;
            }
            return null;
        }

        public boolean isElement(String tag) {
            if (this.getElement(tag) == null) {
                return false;
            }
            return true;
        }

        public boolean isSection(String tag) {
            return this.isElement(tag);
        }

        /**
         * convenience method to find all child elements with the same tag name
         * @param tag the name of the elements to be found
         * @return an array of Element objects
         */
        public Element[] getElements(String tag) {
            Element child;
            Element[] childArray = new Element[0];
            List<Element> children = new LinkedList<Element>();

            int i = 1-1; int n = this.getElements().size();
            while (i <= (n-1)) {
                child = this.getElements().get(i);
                if (child.getTag().equals(tag)) {
                    children.add(child);
                }
                i++;
            }
            return children.toArray(childArray);
        }

        /**
         * convenience method to read the data associated with an attribute name
         * @param description the name of the attribute         
         * @return the attribute object that was found OR null
         */
        public Element.Attribute getAttribute(String description) {
            Element.Attribute attribute;
            int i = 1-1; int n = this.getAttributes().size();
            while (i <= (n-1)) {
                attribute = this.getAttributes().get(i);
                if (attribute.getDescription().equals(description)) {
                    return attribute;
                }
                i++;
            }
            return null;
        }

    }

    /**
     * convenience method to find the child element of the root element with the given tag name
     * NOTE(!): ALWAYS finds the FIRST element with the given tag name and does not test
     * for more elements with the same name.
     *
     * @param tag the name of the element to be found
     * @return the (first) element object found OR null
     */
    public Element getElement(String tag) {
        return this.getRootElement().getElement(tag);
    }
    /**
     * convenience method to add an element
     */
    public Element addElement(String tag) {
        return this.getRootElement().addElement(tag);
    }
    /**
     * convenience method to get a section (=> getElement)
     */
    public Element getSection(String tag) {
        return this.getElement(tag);
    }
    /**
     * convenience method to add a section (=> addElement)
     */
    public Element addSection(String tag) {
        return this.addElement(tag);
    }


    /**
     * the root element of the xml tree
     */
    private Element fRootElement = new Element(); {this.fRootElement.setTag("SETTINGS");}
    /**
     * method to obtain fRootElement
     * @return fRootElement
     */
    public Element getRootElement() {return this.fRootElement;}

    /**
     * use this method to parse an xml file
     * @param fileName the filename of an xml file that is to be read
     * @return true if parsing was done without errors
     */
    public boolean loadFromFile(String fileName) {
        this.fRootElement = new Element(); //get rid of old tree
        RandomAccessFile raf = null;
        try {            
            raf = new RandomAccessFile(fileName, "r");
            this.loadHeader(raf);
            this.parseDocument(raf);
            return true;
        } catch (Exception e) {
            this.fRootElement = new Element();
            return false;
        } finally {
            try {raf.close();} catch (Exception e2) {}
        }

    }

    /**
     * dummy string to indicate xml version == 1.0
     */
    private String fVersion = "1.0";
    /**
     * dummy string to indicate xml encoding == UTF-8
     */
    private String fEncoding = "UTF-8";

    /**
     * method to obtain <?xml version="1.0" encoding="UTF-8">
     * @param raf the xml file as RandomAccessFile
     * @return dummy string
     * @throws java.io.IOException
     */
    private String loadHeader(RandomAccessFile raf) throws IOException {
        long start = -1;
        long stop = -1;
        if (pos("<?xml ", raf, 2048) != -1) {
            start = pos("<?xml", raf) + 5 - 1;
            if (pos("?>", raf, 2048) != -1) {
                stop = pos("?>", raf) - 2;

                //read header
                byte[] headerBytes = new byte[((int)(stop-start))+1];
                raf.seek(start + 1);
                raf.read(headerBytes);
                raf.seek(stop + 1);

                try {
                    String header = new String(headerBytes, "UTF-8");
                    header = decode(header);
                    header = header.trim();

                    String versionAttribute = header.substring(0, header.indexOf("="));
                    header = header.substring(header.indexOf("=") + 2);
                    String version = header.substring(0, header.indexOf("\""));
                    header = header.substring(header.indexOf("\"") + 2);

                    String encodingAttribute = header.substring(0, header.indexOf("="));
                    header = header.substring(header.indexOf("=") + 2);
                    String encoding = header.substring(0, header.indexOf("\""));
                
                    if(   (!(versionAttribute.toUpperCase().equals("VERSION")))
                       ||
                          (!(version.equals("1.0")))
                       ||
                          (!(encodingAttribute.toUpperCase().equals("ENCODING")))
                       ||
                          (!(encoding.toUpperCase().equals("UTF-8")))
                       ) {
                        this.fVersion = null;
                        this.fEncoding = null;
                        return "UNRECOGNIZED HEADER - TRYING ANYWAYS";                
                    } else {
                        this.fVersion = version;
                        this.fEncoding = encoding;
                        return decode(new String(headerBytes, "UTF-8")).trim();
                    }
                } catch (Exception e) {
                    this.fVersion = null;
                    this.fEncoding = null;
                    return "UNRECOGNIZED HEADER";
                }
            }
        }
        this.fVersion = null;
        this.fEncoding = null;
        return "NO HEADER PRESENT";
    }

    /**
     * strips a string from unwanted characteropenings and endings, e.g. <!-- and -->
     * @param str the string that could contain the unwanted characteropenings and endings
     * @return the string with the unwanted characteropenings and endings
     */
    private static String strip(String str) {
        if (str.startsWith("<!--")) {
            if (str.length() > 4) {
                str = str.substring(4);                
            } else {
                return "";
            }
        }

        if (str.endsWith("-->")) {
            if (str.length() > 3) {
                str = str.substring(0, str.length()-3);
            } else {
                return "";
            }
        }

        if (str.startsWith("<")) {
            if (str.length() > 1) {
                str = str.substring(1);
            } else {
                return "";
            }
        }

        if (str.startsWith(">")) {
            if (str.length() > 1) {
                str = str.substring(1);
            } else {
                return "";
            }
        }

        if (str.endsWith("/>")) {
            if (str.length() > 2) {
                str = str.substring(0, str.length()-2);
            } else {
                return "";
            }
        }

        if (str.endsWith(">")) {
            if (str.length() > 1) {
                str = str.substring(0, str.length()-1);
            } else {
                return "";
            }
        }

        return str;
    }

    /**
     * method to load a part of an xml file
     * @param partStop the position at which to stop, the method reads (partStop - partStart + 1)
     * bytes from the raf stream
     * @param partStart the position at which to start, partStart is positioned
     * just before the first element to be read
     * @param raf the xml file as RandomAccessFile
     * @return the read bytes as an UTF-8 string
     * @throws java.io.IOException
     */
    private String loadPart(long partStop, long partStart, RandomAccessFile raf) throws IOException {
        raf.seek(partStart);
        int partLength = ((int)(partStop - partStart + 1));
        byte[] partBytes = new byte[partLength];
        raf.read(partBytes);
        String partString = new String(partBytes, "UTF-8");
        partString = strip(partString); //get rid of <!--, <, > and -->, />, > (prn)
        partString = decode(partString);
        partString = partString.trim();
        return partString;
    }

    /**
     * this method parses the document contents, from root opening tag to root end tag,
     * to all the nested child tags; this method is called recursively
     * @param raf the xml file as RandomAccessFile
     * @param positionStop a marker that indicates the end of the search block
     * @param positionStart a marker that indicates the start of a search block
     * @param parent the root element (if this method is called recursively, it is no
     * longer the root element, but the parent element, hence "parent" for name)
     * @param isRoot should be true when invoked (if this method is called recursively,
     * isRoot is always false, but the first time this method is invoked, it is presented
     * with the root element, so therefore it should be true)
     * @throws java.io.IOException
     */
    private void parseChildElements(RandomAccessFile raf, long positionStop, long positionStart, Element parent, boolean isRoot) throws IOException {
try {
            Element element = null;
            raf.seek(positionStart);

            long partStart = -1;
            long partStop = -1;

            long contentStart = -1;
            long contentStop = -1;

            //traverse: determine number of child elements, ignoring comments
            int numberOfElements = 0;
            while (raf.getFilePointer() < positionStop) {
                //get root element or get new element
                if (isRoot) {
                    element = this.getRootElement();
                } else {
                    element = new Element();
                    element.setLevel(parent.getLevel()+1);
                }

                //either list of child elements XOR value
                partStart = pos("<", raf, (positionStop - raf.getFilePointer() + 1));
                if (partStart == -1) {
                    if (isRoot) {
                        throw new IOException("NO DEFINITION OF ROOT ELEMENT WAS FOUND");
                    } else {
                        //if not root element: no child elements ==> set value!
                        //String contents = this.loadPart(positionStop, raf.getFilePointer(), raf);
                        String contents = this.loadPart(positionStop, positionStart, raf);
                        parent.setValueAsString(contents);
                        return;                    
                    }
                }
                raf.seek(partStart);

                //comment
                String comment = null;
                if (pos("<!--", raf, 4) == partStart) {
                    partStop = pos("-->", raf, (positionStop - raf.getFilePointer() + 1));
                    if (partStop == -1) {throw new IOException("UNTERMINATED COMMENT STRING");}
                    comment = this.loadPart(partStop+2, partStart, raf);
                    raf.seek(partStop+3);
                    element.setComment(comment);
                } else {
                    raf.seek(partStart);
                }

                //tag
                String tag = null;
                partStart = pos("<", raf, (positionStop - raf.getFilePointer() + 1));
                if (pos("</", raf, 3) == partStart) {
                    throw new IOException("UNEXPECTED END TAG");
                } else if (partStart != -1) {
                    raf.seek(partStart);
                }
                if (partStart == -1) {throw new IOException("UNEXPECTED EMPTY CHILD ELEMENT LIST OR NOT ROOT ELEMENT");}                
                partStop = pos(">", "/>", " ", raf, (positionStop - raf.getFilePointer() + 1));
                if (partStop == -1) {throw new IOException("MALFORMED XML");}
                tag = this.loadPart(partStop, partStart, raf);
                if (tag.endsWith("/")) {                    
                    tag = tag.substring(0, tag.length()-1);
                }
                element.setTag(tag);
                if (!(isRoot)) {
                    parent.getElements().add(element);
                }
                raf.seek(partStart);
                partStop = pos(">", "/>", " ", raf, (positionStop - raf.getFilePointer() + 1));
                if (pos(" ", raf, 2) == partStop) {
                    partStart = partStop;
                    partStop = pos (">", "/>", raf, (positionStop - raf.getFilePointer() + 1));
                    //get attributeString for root element
                    String attributeString = this.loadPart(partStop, partStart, raf);
                    processAttributeString(attributeString, element);
                }
                raf.seek(partStop);
                partStop = pos(">", "/>", raf, (positionStop - raf.getFilePointer() + 1));                
                if (pos("/>", raf, 2) == partStop) {
                    //no value, so return! VALID RETURN!
                    raf.seek(partStop + 1);
                    continue;
                }
                raf.seek(partStop);
    contentStart = pos(">", raf, 2) + 1;

                String endtag = "</" + tag + ">";
                int count = 1;
                while (count > 0) {
                    partStop = pos(endtag, ("<" + tag), raf, (positionStop - raf.getFilePointer() + 1));
                    if (partStop == -1) {throw new IOException("MALFORMED XML PARTSTOP");}
                    if (pos(endtag, raf, 3) == partStop) {
                        partStop = partStop + 1;
                        count = count - 1;
                    } else {
                        raf.seek(partStop);
                        if (pos("<" + tag + ">", raf, 5) == partStop) {
                            partStop = partStop + 1;
                            count = count + 1;
                        } else {
                            raf.seek(partStop);
                            long q = pos(">", "/>", raf, (positionStop - raf.getFilePointer() + 1));
                            if (pos("/>", raf, 2) == q) {
                                //nothing: count + 1 - 1
                                partStop = partStop + 1;
                            } else {
                                partStop = partStop + 1;
                            }
                        }
                    }
                    raf.seek(partStop);
                }
    contentStop = raf.getFilePointer()-1-1;

                //positioned at start of </TAG>, now reposition just after </TAG>
                partStart = pos(">", raf, (positionStop - raf.getFilePointer() + 1)) + 1;
                this.parseChildElements(raf, contentStop, contentStart, element, false);
                raf.seek(partStart);

                if ((raf.getFilePointer() + endtag.length() + 2) >= positionStop) {
                    break; //no more elements
                }

                if (isRoot) {break;}
                //there are more elements!
                numberOfElements++;
            }

            int i = 1-1; int n = numberOfElements;
            while (i <= (n-1)) {

                i++;
            }
} catch (Exception e)             {
    System.out.println(e);
}
    }

    /**
     * the method that invokes the parseChildElements method
     * @param raf the xml file as RandomAccessFile
     * @throws java.io.IOException
     */
    private void parseDocument(RandomAccessFile raf) throws IOException {
        this.fRootElement = new Element();
        long positionStart = raf.getFilePointer();
        long positionStop = raf.length()-1;
        this.parseChildElements(raf, positionStop, positionStart, null, true);
    }

    /**
     * if there are attributes to an element, this method parses them and adds them
     * to the element
     * @param attributeString the string that contains all attributes in the form of
     * name="value" name="value" etc.
     * @param element the element that should possess the attributes in question
     */
    private static void processAttributeString(String attributeString, Element element) {
        //process attributes
        attributeString = attributeString.trim(); 
        if (!(attributeString.equals(""))) {
            while (attributeString.contains("=")) {
                String attributeName = attributeString.substring(0, attributeString.indexOf("="));
                if (attributeString.length() > attributeName.length() + 1) {
                    attributeString = attributeString.substring(attributeString.indexOf("=") + 1);
                } else {
                    attributeString = "";
                }
                String value = null;
                if ((attributeString.indexOf("'")) == 0) {
                    //value is enclosed by single quotes
                    value = attributeString.substring(1);
                    attributeString = attributeString.substring(1);
                    value = value.substring(0, value.indexOf("'"));
                    if (attributeString.length() > value.length() + 1) {
                        attributeString = attributeString.substring(attributeString.indexOf("'") + 1);
                        attributeString = attributeString.trim();
                    } else {
                        attributeString = "";
                    }
                } else if (attributeString.indexOf("\"") == 0) {
                    //value is enclosed between double quotes
                    value = attributeString.substring(1);
                    attributeString = attributeString.substring(1);
                    value = value.substring(0, value.indexOf("\""));
                    if (attributeString.length() > value.length() + 1) {
                        attributeString = attributeString.substring(attributeString.indexOf("\"") + 1);
                        attributeString = attributeString.trim();
                    } else {
                        attributeString = "";
                    }
                }
                if (!(attributeName.equals(""))) {
                    Element.Attribute attribute = new Element.Attribute();
                    attribute.setDescription(attributeName);
                    attribute.setValue(value);
                    element.getAttributes().add(attribute);
                }
            }
        }
    }

    /**
     * see private static long pos(String searchString1, String searchString2, String searchString3, RandomAccessFile raf, long maxBytes)
     */
    private static long pos(String ch, RandomAccessFile raf) {
        return XMLFileStore.pos(ch, ch, ch, raf, -1);
    }
    /**
     * see private static long pos(String searchString1, String searchString2, String searchString3, RandomAccessFile raf, long maxBytes)
     */
    private static long pos(String ch1, String ch2, RandomAccessFile raf) {
        return XMLFileStore.pos(ch1, ch2, ch2, raf, -1);
    }
    /**
     * see private static long pos(String searchString1, String searchString2, String searchString3, RandomAccessFile raf, long maxBytes)
     */
    private static long pos(String ch1, String ch2, String ch3, RandomAccessFile raf) {
        return XMLFileStore.pos(ch1, ch2, ch3, raf, -1);
    }
    /**
     * see private static long pos(String searchString1, String searchString2, String searchString3, RandomAccessFile raf, long maxBytes)
     */
    private static long pos(String ch, RandomAccessFile raf, long maxBytes) {
        return XMLFileStore.pos(ch, ch, ch, raf, maxBytes);
    }
    /**
     * see private static long pos(String searchString1, String searchString2, String searchString3, RandomAccessFile raf, long maxBytes)
     */
    private static long pos(String ch1, String ch2, RandomAccessFile raf, long maxBytes) {
        return XMLFileStore.pos(ch1, ch2, ch2, raf, maxBytes);
    }
    /**
     * method to position the stream raf at the beginning of EITHER searchString1 OR searchString2 OR SearchString3, whichever is found first
     * @param raf the xml file as RandomAccessFile
     * @param maxBytes the maximum number of bytes the filepointer should move [if pos("/>", ">", " ", raf) returns, we can invoke
     * pos once more to find out which of the three searchstrings was found first, e.g. partStop = pos("/>", ">", " ", raf),
     * if (pos(" ", raf, 3) == partStop) {then this} else {reposition at partStop}]
     * @return the position at which one of the searchstrings was found OR -1 if this did not occur
     */
    private static long pos(String searchString1, String searchString2, String searchString3, RandomAccessFile raf, long maxBytes) {
        if (searchString1.length() < searchString3.length()) {String help = searchString1; searchString1 = searchString3; searchString3 = help;}
        if (searchString1.length() < searchString2.length()) {String help = searchString1; searchString1 = searchString2; searchString2 = help;}

        byte[] actualBytesRead = new byte[16*1024];
        byte[] searchContent1 = searchString1.getBytes(); byte[] actualContent1 = new byte[searchContent1.length]; //xxxxxxxxxxxxxxxx
        byte[] searchContent2 = searchString2.getBytes(); byte[] actualContent2 = new byte[searchContent2.length]; //xxxxxxxxxxxx
        byte[] searchContent3 = searchString3.getBytes(); byte[] actualContent3 = new byte[searchContent3.length]; //xxxxxx

        try {
            long originalPos = raf.getFilePointer();
            long currentPos;

            do {
                //read some bytes
                currentPos = raf.getFilePointer();
                int numberOfBytesRead = raf.read(actualBytesRead);
                if (numberOfBytesRead == -1) {return -1;} //end of file, no more bytes to read

                int i = 1-1; int n = numberOfBytesRead;
                while (i <= (n-1)) {
                    //i = (numberOfBytesRead-1) - j;
                    if (maxBytes > 0) {
                        if (((long)((currentPos + i) - originalPos)) >= ((long)(maxBytes))) {
                            return -1;
                        }
                    }
                    if ((actualBytesRead[i] == searchContent1[0]) && ((i + searchContent1.length - 1) <= (numberOfBytesRead - 1))) {
                       System.arraycopy(actualBytesRead, i, actualContent1, 0, searchContent1.length);
                       if (Arrays.equals(searchContent1,actualContent1)) {
                            currentPos = currentPos + i;
                            raf.seek(currentPos);
                            return currentPos;
                       }
                    }
                    if ((actualBytesRead[i] == searchContent2[0]) && ((i + searchContent2.length - 1) <= (numberOfBytesRead - 1))) {
                        System.arraycopy(actualBytesRead, i, actualContent2, 0, searchContent2.length);
                        if (Arrays.equals(searchContent2,actualContent2)) {
                            currentPos = currentPos + i;
                            raf.seek(currentPos);
                            return currentPos;
                        }
                    }
                    if ((actualBytesRead[i] == searchContent3[0]) && ((i + searchContent3.length - 1) <= (numberOfBytesRead - 1))) {
                        System.arraycopy(actualBytesRead, i, actualContent3, 0, searchContent3.length);
                        if (Arrays.equals(searchContent3,actualContent3)) {
                            currentPos = currentPos + i;
                            raf.seek(currentPos);
                            return currentPos;
                        }
                    }                    
                    i++;
                }                
            } while (true);           
        } catch (Exception e) {
            return -1;
        }
    }

    /**
     * decodes a string that was encoded with encode
     */
    private static String decode(String str) {
        try {
            //these entities all consist of just one byte, so can use to compare
            byte amp = (new String("&").getBytes("UTF-8"))[0]; int amp_size = 5;
            byte lt = (new String("<").getBytes("UTF-8"))[0]; int lt_size = 4;
            byte gt = (new String(">").getBytes("UTF-8"))[0]; int gt_size = 4;
            byte apos = (new String("'").getBytes("UTF-8"))[0]; int apos_size = 6;
            byte quot = (new String("\"").getBytes("UTF-8"))[0]; int quot_size = 6;

            byte[] buffer = str.getBytes("UTF-8");
            int sizerequirement = buffer.length;
            int i = 1-1; int n = buffer.length;
            while (i <= (n-1)) {
                if (buffer[i] == amp) {
                    if (buffer[i+1] == (new String("a").getBytes("UTF-8"))[0]) {
                        if (buffer[i+2] == (new String("m").getBytes("UTF-8"))[0]) {
                            sizerequirement = sizerequirement - amp_size + 1;
                            i = i + amp_size - 1;
                        } else { //buffer[i+2] == "p"
                            sizerequirement = sizerequirement - apos_size + 1;
                            i = i + apos_size - 1;
                        }
                    } else if (buffer[i+1] == (new String("l").getBytes("UTF-8"))[0]) {
                            sizerequirement = sizerequirement - lt_size + 1;
                            i = i + lt_size - 1;
                    } else if (buffer[i+1] == (new String("g").getBytes("UTF-8"))[0]) {
                            sizerequirement = sizerequirement - gt_size + 1;
                            i = i + gt_size - 1;
                    } else if (buffer[i+1] == (new String("q").getBytes("UTF-8"))[0]) {
                            sizerequirement = sizerequirement - quot_size + 1;
                            i = i + quot_size - 1;
                    }
                } 
                i++;
            }

            byte[] output = new byte[sizerequirement];
            i = 1-1; n = buffer.length; int position = 1-1;
            while (i <= (n-1)) {
                if (buffer[i] == amp) {
                    if (buffer[i+1] == (new String("a").getBytes("UTF-8"))[0]) {
                        if (buffer[i+2] == (new String("m").getBytes("UTF-8"))[0]) {
                            output[position] = amp;
                            i = i + amp_size - 1;
                            position = position + 1;
                        } else { //buffer[i+2] == "p"
                            output[position] = apos;
                            i = i + apos_size - 1;
                            position = position + 1;
                        }
                    } else if (buffer[i+1] == (new String("l").getBytes("UTF-8"))[0]) {
                            output[position] = lt;
                            i = i + lt_size - 1;
                            position = position + 1;
                    } else if (buffer[i+1] == (new String("g").getBytes("UTF-8"))[0]) {
                            output[position] = gt;
                            i = i + gt_size - 1;
                            position = position + 1;
                    } else if (buffer[i+1] == (new String("q").getBytes("UTF-8"))[0]) {
                            output[position] = quot;
                            i = i + quot_size - 1;
                            position = position + 1;
                    } else {
                        output[position] = buffer[i];
                        position = position + 1;
                    }
                } else {
                    output[position] = buffer[i];
                    position = position + 1;
                }
                i++;
            }
            str = new String(output, "UTF-8");

            return str;
        } catch (Exception e) {
            return str; //hope for the best: well, ....should pose no problems since UTF-8 encoding IS very likely to be present
        }
    }

    /**
     * encodes a string: replaces &, <, >, ', and " for safe descriptions
     */
    private static String encode(String str) {
        if (str.equals("")) {return str;}
        try {
            //these entities all consist of just one byte, so can use to compare
            byte amp = (new String("&").getBytes("UTF-8"))[0]; int amp_size = 5; byte [] amp_encoded = new String("&amp;").getBytes("UTF-8");
            byte lt = (new String("<").getBytes("UTF-8"))[0]; int lt_size = 4; byte [] lt_encoded = new String("&lt;").getBytes("UTF-8");
            byte gt = (new String(">").getBytes("UTF-8"))[0]; int gt_size = 4; byte [] gt_encoded = new String("&gt;").getBytes("UTF-8");
            byte apos = (new String("'").getBytes("UTF-8"))[0]; int apos_size = 6; byte [] apos_encoded = new String("&apos;").getBytes("UTF-8");
            byte quot = (new String("\"").getBytes("UTF-8"))[0]; int quot_size = 6; byte [] quot_encoded = new String("&quot;").getBytes("UTF-8");
                        
            byte[] buffer = str.getBytes("UTF-8");
            int sizerequirement = buffer.length;
            int i = 1-1; int n = buffer.length;
            while (i <= (n-1)) {
                       if (buffer[i] == amp) {
                    sizerequirement = sizerequirement + amp_size - 1;
                } else if (buffer[i] == lt) {
                    sizerequirement = sizerequirement + lt_size - 1;
                } else if (buffer[i] == gt) {
                    sizerequirement = sizerequirement + gt_size - 1;
                } else if (buffer[i] == apos) {
                    sizerequirement = sizerequirement + apos_size - 1;
                } else if (buffer[i] == quot) {
                    sizerequirement = sizerequirement + quot_size - 1;
                }
                i++;
            }

            byte[] output = new byte[sizerequirement];
            i = 1-1; n = buffer.length; int position = 1-1;
            while (i <= (n-1)) {

                       if (buffer[i] == amp) {
                    System.arraycopy(amp_encoded, 0, output, position, amp_size);
                    position = position + amp_size;
                } else if (buffer[i] == lt) {
                    System.arraycopy(lt_encoded, 0, output, position, lt_size);
                    position = position + lt_size;
                } else if (buffer[i] == gt) {
                    System.arraycopy(gt_encoded, 0, output, position, gt_size);
                    position = position + gt_size;
                } else if (buffer[i] == apos) {
                    System.arraycopy(apos_encoded, 0, output, position, apos_size);
                    position = position + apos_size;
                } else if (buffer[i] == quot) {
                    System.arraycopy(quot_encoded, 0, output, position, quot_size);
                    position = position + quot_size;
                } else {
                    output[position] = buffer[i];
                    position = position + 1;
                }
                i++;
            }

            str = new String(output, "UTF-8");
            return str;
        } catch (Exception e) {
            return str; //hope for the best: well, ....should pose no problems since UTF-8 encoding IS very likely to be present
        }
    }

    /**
     * use this method to save to an xml file
     * @param fileName the filename of an xml file
     * @return true if the write was done without errors
     */
    public boolean saveToFile(String fileName) {
        return this.saveToFile(fileName, true);
    }

    /**
     * use this method to save to an xml file
     * @param fileName the filename of an xml file
     * @param nicerOutput if true this method adds CRLF control characters to the
     * .xml file
     * @return true if the write was done without errors
     */
    public boolean saveToFile(String fileName, boolean nicerOutput) {
        OutputStreamWriter osw = null;
        try {
            //open document for saving
            osw = new OutputStreamWriter(new FileOutputStream(fileName), "UTF-8");

            //save version info and comment prn
            osw.write("<?xml version=\"1.0\" encoding=\"UTF-8\"" + " ?>"); if (nicerOutput) {osw.write("\r\n");}
            if (this.fRootElement.hasComment()) {
                osw.write("<!-- " + encode(this.fRootElement.getComment()) + " -->"); if (nicerOutput) {osw.write("\r\n");}
            }

            //start of document
            osw.write("<" + encode(this.fRootElement.getTag()));
            if (this.fRootElement.hasAttributes()) {
                Element.Attribute attribute = null;
                int k = 1-1; int o = this.fRootElement.getAttributes().size();
                while (k <= (o-1)) {
                    attribute = this.fRootElement.getAttributes().get(k);
                    if (!(attribute.getDescription().equals(""))) {
                        osw.write(" " + encode(attribute.getDescription()) + "=\"" + encode(attribute.getValue()) + "\"");
                    }
                    k++;
                }
            }
            osw.write(">");
            
            //middle part of document
            if     (this.fRootElement.hasElements() == false) {
                if (this.fRootElement.hasValue()) {
                    osw.write(encode(this.fRootElement.getValueAsString()));
                }
            } else {
                 if (nicerOutput) {osw.write("\r\n");}
                XMLFileStore.listChildren(this.fRootElement, osw, nicerOutput);
            }

            //end of document            
            osw.write("</" + encode(this.fRootElement.getTag()) + ">");
            return true;
        } catch (Exception e) {
            return false;
        } finally {
            try {osw.close();} catch (Exception e2) {}
        }
    }

    /**
     * method that is called recursively to write out the whole element tree
     * to disk as .xml
     */
    private static void listChildren(Element element, OutputStreamWriter osw, boolean nicerOutput) throws IOException { //called recursively
        if (element.hasElements()) {
            Element child;
            int i = 1-1; int n = element.getElements().size();
            while (i <= (n-1)) {
                //get the next child
                child = element.getElements().get(i);
                
                //get indent
                String indent = "";
                if (nicerOutput) {
                    int l = 1-1; int p = child.getLevel();
                    while (l <= (p-1)) {
                        indent = indent + "    ";
                        l++;
                    }
                }

                //if child has a comment then print that
                if (child.hasComment()) {
                    osw.write(indent + "<!-- " + encode(child.getComment()) + " -->"); if (nicerOutput) {osw.write("\r\n");}
                }
                
                //print tag and attributes prn
                osw.write(indent + "<" + encode(child.getTag()));
                if (child.hasAttributes()) {
                    int j = 1-1; int m = child.getAttributes().size();
                    while (j <= (m-1)) {
                        if (child.getAttributes().get(j).getDescription() != null) {
                            if (!(child.getAttributes().get(j).getDescription().equals(""))) {
                                osw.write(" " + encode(child.getAttributes().get(j).getDescription()));
                                osw.write("=\"");
                                osw.write(encode(child.getAttributes().get(j).getValue()));
                                osw.write("\"");
                            }
                        }                        
                        j++;
                    }
                }
                osw.write(">");

                if (child.hasElements() == false) {
                    if (child.hasValue()) {
                        osw.write(encode(child.getValueAsString()));
                    }
                    osw.write("</" + encode(child.getTag()) + ">");
                     if (nicerOutput) {osw.write("\r\n");}
                } else {
                     if (nicerOutput) {osw.write("\r\n");}

                    //if child has children then perform all these actions on the children as well
                    XMLFileStore.listChildren(child, osw, nicerOutput);

                    //then print end tag
                    osw.write(indent + "</" + encode(child.getTag()) + ">");
                     if (nicerOutput) {osw.write("\r\n");}
                }
                i++;
            }
        }
        
    }

    public void clear() {
        this.fRootElement = null;
    }

    public static void main(String[] args) throws IOException {
        String  valueOne    = "A field to store some kind of information"; String  readOne;
        String  valueTwo    = "A string encoded as base 64";               String  readTwo;
        boolean valueThree  = true;                                        boolean readThree;
        byte    valueFour   = ((byte)(-123));                              byte    readFour;
        double  valueFive   = 4.2323;                                      double  readFive;
        float   valueSix    = ((float)(-4.2323));                          float   readSix;
        byte[]  valueSeven  = null;                                        byte[]  readSeven;
        int     valueEight  = 434343;                                      int     readEight;
        long    valueNine   = Long.parseLong("34234343434423");            long    readNine;
        short   valueTen    = ((short)(34323));                            short   readTen;
        String  valueEleven = "A string";                                  String  readEleven;
        try {
            MessageDigest hash = MessageDigest.getInstance("SHA-1");
            hash.update("get the message digest of this message\n\r".getBytes("UTF-8"));
            hash.update("which is composed of these two lines\n\r".getBytes("UTF-8"));
            valueSeven = hash.digest();
        } catch (Exception e) {
            valueSeven = new byte[256];
            int i = 1-1; int n = valueSeven.length;
            while (i <= (n-1)) {
                valueSeven[i] = ((byte)(i));
                i++;
            }
        }

        XMLFileStore xml = new XMLFileStore("TEST", "This is an XMLFileStore .xml file!");
        final String FIRST_SECTION = "FIRST_SECTION";
        final String SECOND_SECTION = "SECOND_SECTION";
        final String THIRD_SECTION = "THIRD_SECTION";

        final String FIRST_VALUE = "FIRST_VALUE";
        final String SECOND_VALUE = "SECOND_VALUE";
        final String THIRD_VALUE = "THIRD_VALUE";
        final String FOURTH_VALUE = "FOURTH_VALUE";
        final String FIFTH_VALUE = "FIFTH_VALUE";
        final String SIXTH_VALUE = "SIXTH_VALUE";
        final String SEVENTH_VALUE = "SEVENTH_VALUE";
        final String EIGHTH_VALUE = "EIGHTH_VALUE";
        final String NINTH_VALUE = "NINTH_VALUE";
        final String TENTH_VALUE = "TENTH_VALUE";
        final String ELEVENTH_VALUE = "ELEVENTH_VALUE";
        
        xml.getRootElement().getSection(FIRST_SECTION).getElement(FIRST_VALUE).setValue(valueOne);
        xml.getRootElement().getSection(FIRST_SECTION).getElement(SECOND_VALUE).setValueAsBase64(valueTwo);
        xml.getRootElement().getSection(SECOND_SECTION).getElement(THIRD_VALUE).setValueAsBoolean(valueThree);
        xml.getRootElement().getSection(FIRST_SECTION).getElement(FOURTH_VALUE).setValueAsByte(valueFour);
        xml.getRootElement().getSection(SECOND_SECTION).getElement(FIFTH_VALUE).setValueAsDouble(valueFive);
        xml.getRootElement().getSection(THIRD_SECTION).getElement(SIXTH_VALUE).setValueAsFloat(valueSix);
        xml.getRootElement().getSection(FIRST_SECTION).getElement(SEVENTH_VALUE).setValueAsHex(valueSeven);
        xml.getRootElement().getSection(FIRST_SECTION).getElement(EIGHTH_VALUE).setValueAsInteger(valueEight);
        xml.getRootElement().getSection(SECOND_SECTION).getElement(NINTH_VALUE).setValueAsLongInteger(valueNine);
        xml.getRootElement().getSection(THIRD_SECTION).getElement(TENTH_VALUE).setValueAsShortInteger(valueTen);
        xml.getRootElement().getSection(FIRST_SECTION).getElement(ELEVENTH_VALUE).setValueAsString(valueEleven);
        
        //File tempFile = TTempFile.newTempFile();
        //xml.saveToFile(tempFile.getCanonicalPath());
        //xml.clear();
        //xml.loadFromFile(tempFile.getCanonicalPath());

        //now try to read
        readOne = xml.getRootElement().getSection(FIRST_SECTION).getElement(FIRST_VALUE).getValue();
        readTwo = xml.getRootElement().getSection(FIRST_SECTION).getElement(SECOND_VALUE).getDecodedBase64ValueAsString();
        readThree = xml.getRootElement().getSection(SECOND_SECTION).getElement(THIRD_VALUE).getValueAsBoolean();
        readFour = xml.getRootElement().getSection(FIRST_SECTION).getElement(FOURTH_VALUE).getValueAsByte();
        readFive = xml.getRootElement().getSection(SECOND_SECTION).getElement(FIFTH_VALUE).getValueAsDouble();
        readSix = xml.getRootElement().getSection(THIRD_SECTION).getElement(SIXTH_VALUE).getValueAsFloat();
        readSeven = xml.getRootElement().getSection(FIRST_SECTION).getElement(SEVENTH_VALUE).getDecodedHexValue();
        readEight = xml.getRootElement().getSection(FIRST_SECTION).getElement(EIGHTH_VALUE).getValueAsInteger();
        readNine = xml.getRootElement().getSection(SECOND_SECTION).getElement(NINTH_VALUE).getValueAsLongInteger();
        readTen = xml.getRootElement().getSection(THIRD_SECTION).getElement(TENTH_VALUE).getValueAsShortInteger();
        readEleven = xml.getRootElement().getSection(FIRST_SECTION).getElement(ELEVENTH_VALUE).getValueAsString();

        System.out.println("value  1 " + valueOne);
        System.out.println("read   1 " + readOne);

        System.out.println("value  2 " + valueTwo);
        System.out.println("read   2 " + readTwo);

        System.out.println("value  3 " + valueThree);
        System.out.println("read   3 " + readThree);

        System.out.println("value  4 " + valueFour);
        System.out.println("read   4 " + readFour);

        System.out.println("value  5 " + valueFive);
        System.out.println("read   5 " + readFive);

        System.out.println("value  6 " + valueSix);
        System.out.println("read   6 " + readSix);

        System.out.println("value  7 " + Utils.toHex(valueSeven));
        System.out.println("read   7 " + Utils.toHex(readSeven));

        System.out.println("value  8 " + valueEight);
        System.out.println("read   8 " + readEight);

        System.out.println("value  9 " + valueNine);
        System.out.println("read   9 " + readNine);

        System.out.println("value 10 " + valueTen);
        System.out.println("read  10 " + readTen);

        System.out.println("value 11 " + valueEleven);
        System.out.println("read  11 " + readEleven);
        
    }
}
